﻿--[[
	DuckMod - A World of Warcraft add-on library
	Copyright (C) 2009-2016  Dag Bakken

	DuckMod is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	DuckMod is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with DuckMod.  If not, see <http://www.gnu.org/licenses/>.
]]


local TV=3.1;

if (not DuckMod) then DuckMod = {}; end				-- Make it if empty
if (not DuckMod[TV]) then DuckMod[TV]={}; end		-- Don't mess with it if it's loaded
local DM=DuckMod[TV];
local _;
local NATIVE="NativE";

function spewtable(t,level)
	level=level or "-";
	local _v;
	for k,v in pairs(t) do
		if (type(v)=="string") then _v=v:sub(1,20); if (_v~=v) then _v=_v.." (...)"; end _v="|cFFFFFFFF".._v;		-- (...) show it has been cut
		elseif (type(v)=="table") then _v="|cFFFFFFFF==>>";
		else _v="|cFFFFFFFF"..tostring(v); end
		print("|cFF00FFFF= = >> |cFF8888FF"..level,"|cFFFFFF00"..type(k):sub(1,2).."|cFFFFFFFF",k,"|cFF8888FF= ".."|cFFFFFF00"..type(v):sub(1,2),_v);
		if (type(v)=="table") then spewtable(v,level.." -");
	end end
end

-- Make a random number of integer size with fractional part
function DM:Random(bottom,top)
	if (not bottom) then
		bottom=0;
		top=1;
	elseif (not top) then
		top=bottom;
		bottom=0;
	end

	local num=math.random();		-- Generate a fractional number
	num=num*(top-bottom);			-- Set span
	num=num+bottom;					-- Set floor
	return num;
end

-- Table size
function DM:Length(t)
	if (not t) then return 0; end
	local count=0;
	if (d_type and d_type(t)=="database") then for _,v in pairs(t._str_) do if (v:match("^%d+:(%a%a):")~="ni") then count=count+1; end end
	else for _ in pairs(t) do count=count+1; end end
	return count;
end

-- Safe table-copy with optional merge (equal entries will be overwritten, arg1 has pri)
function DM:CopyTable(t,new)
	if (not t and not new) then return nil; end
	if (not t) then t={}; end
	if (not new) then new={}; end
	local i,v;
	for i,v in pairs(t) do if (type(v)=="table") then new[i]=self:CopyTable(v,new[i]); else new[i]=v; end end
	return new;
end

function DM:CompareTables(one,two)
	if (type(one)~="table") then one={one}; end
	if (type(two)~="table") then two={two}; end
	for _e in pairs(one) do if (one[_e]~=two[_e]) then return nil; end end
	for _e in pairs(two) do if (one[_e]~=two[_e]) then return nil; end end
	return  true;
end

-- create string hash
function DM:Hash(text)
	local hash,i,len,_byte=0,1,text:len(),string.byte;
	repeat hash=hash+_byte(text,i); i=i+1; until (i>len);
	return hash;
end

-- Return the common DuckMod ID
function DM:GetID(link)
	if (type(link)~="string") then
		if (type(link)~="number") then return nil,nil; end
		_,link=GetItemInfo(link);
	end
	local itemID,i1,i2,i3,i4,i5,i6=link:match("|Hitem:(%p?%d+):(%p?%d+):(%p?%d+):(%p?%d+):(%p?%d+):(%p?%d+):(%p?%d+):(%p?%d+)|h%[(.-)%]|h");
	if (not i6) then itemID,i1,i2,i3,i4,i5,i6=link:match("item:(%p?%d+):(%p?%d+):(%p?%d+):(%p?%d+):(%p?%d+):(%p?%d+):(%p?%d+)"); end
	if (not i6) then return; end
	return "item:"..itemID..":"..i1..":"..i2..":"..i3..":"..i4..":"..i5..":"..i6,tonumber(itemID);
end

-- function accepts:
-- numbers, nil, boolean, single/double-quote string, global functions, global variables
-- NOTE: function/variable names cannot be enclosed in quotes
function DM:ReadByText(loc,mousebutton)
	loc=strtrim(loc);
	if (tonumber(loc)) then return tonumber(loc); end					-- it's a base 10 number
	if (loc=="nil") then return nil; end								-- nil
	if (loc=="true") then return true; end								-- boolean
	if (loc=="false") then return false; end							-- boolean
	if ((loc:sub(1,1)=="\"" or loc:sub(1,1)=="\'") and loc:sub(1,1)==loc:sub(-1,-1)) then	-- single or double quotes
		if (loc:len()==2) then return ""; end							-- empty string
		return loc:sub(2,-2);											-- remove quotes
	end
	local f,a=loc:match("^(.+)%((.*)%)$"); a=strtrim(a or "");
	if (f) then
		local p,pSelf=DM:GetFunctionReferenceByText(f);					-- find the global function reference
		if (a~="") then													-- if there are function arguments
			a={strsplit(",",a)};
			local refs={};
			for k,v in ipairs(a) do table.insert(refs,DM:ReadByText(v)); end
			table.insert(refs,mousebutton);			-- can't add after 'unpack' as it will normalise to one argument, adding new once at 2, 3, etc overwriting from 'unpack'
			if (pSelf) then return p(pSelf,unpack(refs)); else return p(unpack(refs)); end
		else
			if (pSelf) then return p(pSelf,mousebutton); else return p(mousebutton); end
		end
	else
		if (loc:find(".")) then
			local fun={strsplit(".",loc)};
			for i,iD in ipairs(fun) do
				if (i==1) then p=_G[iD];			-- first table, so get its reference
				elseif (fun[i+1]) then p=p[iD];		-- there's more
				else return p[fun[i]]; end			-- last one, so index previous table
			end
		else
			return _G[loc];		-- single global variable
		end
	end
end

function DM:StoreByText(loc,...)
	loc=strtrim(loc);
	local f,a=loc:match("(.+)%((.*)%)"); a=strtrim(a or "");
	if (f) then
		local p,pSelf=DM:GetFunctionReferenceByText(f);
		if (a~="") then
			a={strsplit(a)};
			local refs={};
			for k,v in ipairs(a) do table.insert(refs,DM:ReadByText(v)); end
			if (pSelf) then return p(pSelf,unpack(a)); else return p(unpack(a)); end
		else
			if (pSelf) then return p(pSelf,...); else return p(...); end
		end
	else
		if (loc:find(".")) then
			local fun={strsplit(".",loc)};
			for i,iD in ipairs(fun) do
				if (i==1) then p=_G[iD];			-- first table, so get its reference
				elseif (fun[i+1]) then p=p[iD];		-- there's more
				else p[fun[i]]=...; end				-- last one, so index previous table
			end
		else
			_G[loc]=...;		-- single global variable
		end
	end
end

function DM:GetFunctionReferenceByText(f)
	f=strtrim(f);
	local fu={strsplit(":",f)};
	local fun={strsplit(".",fu[1])};
	local p;
	for i,iD in ipairs(fun) do if (i==1) then p=_G[iD]; else p=p[iD]; end end
	pSelf=p;
	if (fu[2]) then
		fun={strsplit(".",fu[2])};
		for i,iD in ipairs(fun) do p=p[iD]; end
	else
		pSelf=nil;
	end
	return p,pSelf;
end

-- Print text
function DM:Chat(msg,r,g,b)
	if (DEFAULT_CHAT_FRAME) then
		if (not r and not g and not b) then r=1; g=1; b=1; end;
		if (not r) then r=0; end;
		if (not g) then g=0; end;
		if (not b) then b=0; end;
		DEFAULT_CHAT_FRAME:AddMessage(msg,r,g,b);
	end
end


-- Multi-threading utility
DM.MT={
	LastStack="",
	Current=0,
	Count=0,
	Speed=(1/30)*1000,
	FastMT=(1/100)*1000,	-- adaptable, virtually un-noticable
	LastTime=0,
	Threads={
	},
};

function DM.MT:Run(name,func,...)
	if (_G["Ac".."cura".."teTime"]) then
		if (_G["Ac".."cura".."teTime"]._debugprofilestop) then debugprofilestop=_G["Ac".."cura".."teTime"]._debugprofilestop; end
		if (_G["Ac".."cura".."teTime"]._debugprofilestart) then debugprofilestart=_G["Ac".."cura".."teTime"]._debugprofilestart; end
	end
	for _,tTable in pairs(self.Threads) do
		if (tTable.orig==func and tTable.name==name) then		-- Already running in another thread.
			return;
	end end
	self.RunningCo=true;
	self.Count=self.Count+1;
	self.Threads[self.Count]={ start=debugprofilestop(), name=name, orig=func, cr=coroutine.create(func), };
	self.Threads[self.Count].runtime=0;
	self.LastTime=self.Threads[self.Count].start;
	self.LastStack="Running "..name;
	local succeeded,result=coroutine.resume(self.Threads[self.Count].cr,...);
	if (not succeeded and Swatter) then
		if (Swatter) then Swatter.OnError(result,nil,self.LastStack);
		else chat(result); chat(self.LastStack); end
	end
	self.RunningCo=nil;
end

-- Timer version
function DM.MT:Yield(fast,dbdata)
	local now=debugprofilestop();
	self.LastStack=debugstack(2);
	local speed=self.Speed;
	if (fast) then speed=self.FastMT; end
	if (now-self.LastTime<=speed and not InCombatLockdown()) then return; end
	self.LastTime=now;
	coroutine.yield();
	self.LastStack=debugstack(2);
end

function DM.MT:Next()
	if (self.Count<1) then return; end		-- no processes running
	if (self.RunningCo) then return; end	-- Don't if we are already doing it. In case of real MT.
	self.Current=self.Current+1;
	if (not self.Threads[self.Current]) then self.Current=1; end	-- Wrap
	if (not self.Threads[self.Current]) then return; end	-- Nothing to do
	if (coroutine.status(self.Threads[self.Current].cr)=="dead") then
		local removeIt=self.Current;
		local runtime=debugprofilestop()-self.Threads[removeIt].start;
		local amount=(100/runtime)*self.Threads[removeIt].runtime;
		while (self.Threads[removeIt]) do self.Threads[removeIt]=self.Threads[removeIt+1]; removeIt=removeIt+1; end
		self.Current=self.Current-1;
		self.Count=self.Count-1;
		return;
	end
	self.RunningCo=true;
	self.Threads[self.Current].slice=debugprofilestop();
	local succeeded,result=coroutine.resume(self.Threads[self.Current].cr);
	self.Threads[self.Current].runtime=self.Threads[self.Current].runtime+(debugprofilestop()-self.Threads[self.Current].slice);
	if (not succeeded) then
--		DropCount.Loaded=nil;		-- crash the main thread
		if (Swatter) then Swatter.OnError(result,nil,self.LastStack);
		else print(result); print(self.LastStack); end
	end
	self.RunningCo=nil;
end

function DM.MT:Processes()
	return self.Count;
end

-- Table handling
DM.Table={
	v03={},
	LastPackerVersion="03",
	tableSlowCompress={},
	Scrap={},
	Cache={},
	Addon={},
};
-- Backward compression compatibility
DM.Table[1]={ Entry ="\1", String="\1\1", Number="\1\2", Bool="\1\3", Nil="\1\4", Other ="\1\9", sTable="\2", eTable="\3", Version="\4", ThisVersion="1", Last="\4", };
DM.Table[2]={ Entry ="\2", String="\2\7", Number="\2\2", Bool="\2\3", Nil="\2\4", Other ="\2\9", sTable="\3", eTable="\4", Version="\7", ThisVersion="2", Last="\7", };
DM.Table.CV=2;
DM.Table.Entry=DM.Table[DM.Table.CV].Entry; DM.Table.String=DM.Table[DM.Table.CV].String; DM.Table.Number=DM.Table[DM.Table.CV].Number; DM.Table.Bool=DM.Table[DM.Table.CV].Bool;
DM.Table.Nil=DM.Table[DM.Table.CV].Nil; DM.Table.Other=DM.Table[DM.Table.CV].Other; DM.Table.sTable=DM.Table[DM.Table.CV].sTable; DM.Table.eTable=DM.Table[DM.Table.CV].eTable;
DM.Table.Version=DM.Table[DM.Table.CV].Version; DM.Table.ThisVersion=DM.Table[DM.Table.CV].ThisVersion; DM.Table.Last=DM.Table[DM.Table.CV].Last;


-- "match" will fail due to miscount in this if byte-coding for the SV file has been changed.
-- speedy version
function DM.Table.v03.DecodeTable(s)
	local _match,_tonumber,_settype,_sub,_byte=string.match,tonumber,DM.Table.v03.SetType,string.sub,string.byte;	-- make everything local
	local nTY,nLE,name,dTY,dLE;
	local pos,maxlen,t=1,s:len(),{};
	repeat
		nTY=_byte(s,pos);													-- get type with numeric representation to not create a new string
		nLE,pos=_match(s,".(%x+):().*",pos); nLE=_tonumber(nLE,16);			-- capture data-length,position - convert length to number
		name=_settype(nTY,_sub(s,pos,pos+(nLE-1))); pos=pos+nLE;			-- grab it and advance to next
		dTY=_byte(s,pos);													-- get type with numeric representation to not create a new string
		dLE,pos=_match(s,".(%x+):().*",pos); dLE=_tonumber(dLE,16);			-- empty capture reads character position
		t[name]=_settype(dTY,_sub(s,pos,pos+(dLE-1))); pos=pos+dLE;			-- grab, shove, advance
	until (pos>maxlen);
	return t;
end

-- third letter of native type string - lower case, hex representation for avoiding string
function DM.Table.v03.SetType(ty,da)
	if (ty==0x72) then return tostring(da); end		-- st-r-ing
	if (ty==0x6D) then return tonumber(da); end		-- nu-m-ber
	if (ty==0x6F) then return (da=="true"); end		-- bo-o-l
	if (ty==0x62) then if (da=="") then return {}; end return DM.Table.v03.DecodeTable(da); end	-- ta-b-le
	if (ty==0x6C) then return nil; end				-- ni-l-
	if (ty==0x6E) then return loadstring(da); end	-- fu-n-ction
	return da;
end

-- this function:
--   accesses all global functions locally for speed
--   uses table concat for extreme speed in repetetive storage, like cloth with 700+ creatures codes in well under 5ms
--   In case of further timing problems: further work can be to store all strings separately to remove all string concatenations
function DM.Table.v03.CodeTable(t)
	local _string_format,_codetable,_tostring,_type,_len,_sub,_dump=string.format,DM.Table.v03.CodeTable,tostring,type,string.len,string.sub,string.dump;	-- make everything local
	local list={};
	local name,add;
	for name_,data in pairs(t) do
		name=_tostring(name_);
		if (_type(data)=="table") then add=_codetable(data); elseif (_type(data)=="function") then add=_dump(data); else add=_tostring(data); end
		list[#list + 1]=_sub(_type(name_),3,3).._string_format("%X",_len(name))..":"..name.._sub(_type(data),3,3).._string_format("%X",_len(add))..":"..add;
	end
	return table.concat(list);
end

function DM.Table.CompressV3(tData,s,e)
	local now=debugprofilestop();
	local tmp=DM.Table.v03.CodeTable(tData);
	table.insert(DM.Table.tableSlowCompress,{t=debugprofilestop()-now,s=s,e=e});
	table.sort(DM.Table.tableSlowCompress,function (a,b) return (a.t>b.t); end);
	table.remove(DM.Table.tableSlowCompress,6);	-- keep list of five
	return tmp;
end

function DM.Table.PrintSlowestCompress()
	print("List of slowest v3 compressions:");
	for i,d in ipairs(DM.Table.tableSlowCompress) do print(i,string.format("%.1f",d.t).." ms",d.s,d.e); end
end

-- applies correct compressor and handles cache
function DM.Table.Write(base,entry,data)
	local cache=DM.Table.GetCache(base);									-- grab or create cache
	if (data) then														-- raw data supplied
		base.__DATA__[entry]="03"..DM.Table.CompressV3(data,base.__METADATA__.sub1,entry);		-- timing debug params -> section, entry
		if (base.__METADATA__.comp) then base.__DATA__[entry]=DM.Code:Compress(base.__DATA__[entry]); end	-- apply compressor
	else base.__DATA__[entry]=nil; end									-- supplied data is 'nil'
	if (not cache) then return true; end								-- no cache, so done
	cache[entry]=data;													-- update cache
	if (base.__METADATA__.datameta and type(data)=="table") then
		setmetatable(cache[entry],_G[base.__METADATA__.datameta]);		-- got entry-level meta function
	end
	return true;
end

-- v1 and v2 backward compatibility only
function DM.Table:PullEntry(cString,CV)
	if (not CV) then CV=self.CV; end
	local stop1,stop2,stop3;
	if (cString:sub(1,1)==self[CV].sTable) then
		stop1=cString:find(self[CV].eTable..cString:sub(2,2),3,true);
		if (not stop1) then return nil; end	-- Missing data
		return cString:sub(1,2),cString:sub(3,stop1-1),cString:sub(stop1+2);
	end
	stop1=cString:find(self[CV].Entry,3,true); if (not stop1) then stop1=cString:len()+1; end	-- Virtual entry position
	stop2=cString:find(self[CV].sTable,3,true); if (not stop2) then stop2=cString:len()+1; end	-- Virtual entry position
	stop3=cString:find(self[CV].eTable,3,true); if (not stop3) then stop3=cString:len()+1; end	-- Virtual entry position
	if (stop1>stop2) then stop1=stop2; end
	if (stop1>stop3) then stop1=stop3; end
	return cString:sub(1,2),cString:sub(3,stop1-1),cString:sub(stop1);
end

-- v1 and v2 backward compatibility only
function DM.Table:DecompressV1(base,cString,anon)
	while(cString:len()>0) do
		local eType,eData=nil,nil;
		local dType,dData=nil,nil;
		if (anon) then
			dType,dData,cString=self:PullEntry(cString,1);
			dData=self:DecompressV1(base,dData);	-- dData is now without table-tags
		else
			eType,eData,cString=self:PullEntry(cString,1);		-- Entry name
			if (eType==self[1].Number) then eData=tonumber(eData); end
			dType,dData,cString=self:PullEntry(cString,1);		-- Entry data
			if (dType==self[1].Number) then dData=tonumber(dData);
			elseif (dType==self[1].Bool) then if (dData=="true") then dData=true; else dData=false; end
			elseif (dType==self[1].Nil) then dData=nil;
			elseif (dType:sub(1,1)==self[1].sTable) then
				base[eData]={};
				dData=self:DecompressV1(base[eData],dData);	-- dData is now without table-tags
		end end
		if (not base) then return dData; else if (anon) then base=dData; else base[eData]=dData; end end
	end
	return base;
end

-- v1 and v2 backward compatibility only
function DM.Table:DecompressV2(base,cString,anon)
	while(cString:len()>0) do
		local eType,eData=nil,nil;
		local dType,dData=nil,nil;
		if (anon) then
			dType,dData,cString=self:PullEntry(cString,2);
			dData=self:DecompressV2(base,dData);	-- dData is now without table-tags
		else
			eType,eData,cString=self:PullEntry(cString,2);		-- Entry name
			if (eType==self[2].Number) then eData=tonumber(eData); end
			dType,dData,cString=self:PullEntry(cString,2);		-- Entry data
			if (dType==self[2].Number) then dData=tonumber(dData);
			elseif (dType==self[2].Bool) then if (dData=="true") then dData=true; else dData=false; end
			elseif (dType==self[2].Nil) then dData=nil;
			elseif (dType:sub(1,1)==self[2].sTable) then
				base[eData]={};
				dData=self:DecompressV2(base[eData],dData);	-- dData is now without table-tags
		end end
		if (not base) then return dData; else if (anon) then base=dData; else base[eData]=dData; end end
	end
	return base;
end

duckmoddatabasemetacode={
	__index = function (t,k)
		if (k==nil) then return nil; end
		if (not rawget(t,"__DATA__")) then return nil; end
		if (not rawget(rawget(t,"__DATA__"),k)) then						-- don't have this item key
			if (rawget(rawget(t,"__METADATA__"),"reverse")=="Item") then	-- the question is reversible
				-- traverse for name (case sensitive)
				for _k,_v in rawpairs(t) do									-- traverse without unpacking
					if (tostring(_v):find(_k,1,true)) then					-- textual search
						if (t[_k][rawget(rawget(t,"__METADATA__"),"reverse")]==k) then	-- found item by name: proper read -> recursive metatable
							return t[_k];									-- return unpacked data
			end end end end
			return nil;
		end
		return DM.Table.Read(t,k);
	end,
	__newindex = function (t,k,v)
		if (not rawget(t,"__METADATA__")) then return nil; end
		if (rawget(rawget(t,"__METADATA__"),"nouse")~=nil) then return nil; end		-- section not activated
		if (not rawget(rawget(t,"__DATA__"),k)) then						-- don't have this item key
			if (rawget(rawget(t,"__METADATA__"),"reverse")) then			-- it's a reversible section
				-- traverse for name
				for _k,_v in rawpairs(t) do									-- traverse without unpacking
					if (_v:find(_k,1,true)) then							-- textual search
						if (t[_k][rawget(rawget(t,"__METADATA__"),"reverse")]) then	-- found item by name
							k=_k; break;									-- set true key and proceed as normal
				end end end
			end
		end
		return DM.Table.Write(t,k,v);
	end,
	__tostring = function (t)
		return "DropCount "..t.__METADATA__.sub1.." database";
	end,
-- metamethod "__len" cannot be redefined.
}

duckmodnonfailtable={
	__index = function (t,k)												-- trying to read non-existent entry
		if (not rawget(t,k) and k~="__DATA__" and k~="__METADATA__") then	-- it's not packer-stuff
			rawset(t,k,{}); DM.Table.__ACTIVATE(rawget(t,k));				-- create missing table and activate packer
		end
		return rawget(t,k);													-- return new table
	end,
}

function DM.Table.__ACTIVATE(t)
	t.__DATA__=t.__DATA__ or {};
	t.__METADATA__=t.__METADATA__ or {};
	setmetatable(t,nil);							-- STOP!
	setmetatable(t,duckmoddatabasemetacode);		-- GO!
end

function DM.Table.Enable(root,...)
	DM.Table.MyAddon(root);												-- prep it
	local count=select("#",...);
	if (count>0) then													-- there are specific sub-levels
		local i=1;
		local list,buf,name={},root;
		for i=1,count do
			name=select(i,...);											-- get next parameter
			setmetatable(buf,nil);										-- STOP!
			buf[name]=buf[name] or {};									-- create it if it's missing
			buf=buf[name];												-- grab reference
			setmetatable(buf,duckmodnonfailtable);						-- nested should not fail
			table.insert(list,name);									-- remember name
		end
		DM.Table.__ACTIVATE(buf);											-- activate/create packer
		for k,v in ipairs(list) do										-- walk nested levels
			buf.__METADATA__["sub"..k]=v;								-- add all names
		end
		i=1; while(buf.__METADATA__["sub"..(count+i)]) do buf.__METADATA__["sub"..(count+i)]=nil; i=i+1; end	-- remove others
		return true;
	end
	local function dive(sublevel)
		local ep=true;
		for k,v in pairs(sublevel) do
			if (type(v)=="table" and k~="__DATA__" and k~="__METADATA__") then	-- a regular table
				ep=nil;
				if (v.__DATA__) then									-- it has been used as packer ep before
					DM.Table.__ACTIVATE(v);								-- activate packer
					return;												-- no more nested per definition
				end
				if (not dive(v)) then									-- no pre-registration - so try for more levels
					setmetatable(v,nil);								-- STOP!
					setmetatable(v,duckmodnonfailtable);				-- nested should not fail
		end end end
		if (sublevel==root) then return; end							-- it's the root. we do not pack at this level.
		if (not ep) then return; end									-- there are nested tables, so not end point
		DM.Table.__ACTIVATE(sublevel);										-- activate packer
		return true;
	end
	dive(root);
end

function DM.Table.DecompressV3(cString)
	if (cString=="") then return {}; end
	return DM.Table.v03.DecodeTable(cString);
end

-- 'root' is the root table for the addon in question
function DM.Table.MyAddon(root)
	DM.Table.Addon=DM.Table.Addon or {};
	DM.Table.Addon[root]=DM.Table.Addon[root] or {};
	DM.Table.Addon[root].cache=DM.Table.Addon[root].cache or {};
end

-- Unpack an entry from meta and store it back unpacked.
function DM.Table.Unpack(base,entry)
	if (not base) then return; end
	local raw=base.__DATA__;
	if (not raw or not raw[entry]) then return; end
	if (type(raw[entry])~="string") then return; end
	rawset(raw,entry,base[entry]);				-- store unpacked by meta over packed
	if (type(raw[entry])~="table") then
		chat("Unpacker did not unpack ("..type(raw[entry]).." "..tostring(entry)..")");
	end
end

function DM.Table.GetCache(base)
	if (base.__METADATA__.nocache) then return nil; end					-- cache off
	if (not base.__METADATA__.sub1) then return nil; end				-- metadata not in use
	local buf,i,cache,loc;
	for k,v in pairs(DM.Table.Addon) do									-- check all addons
		buf=k;															-- grab this root for testing
		i=1; loc=base.__METADATA__["sub"..i];
		while(loc) do								-- valid
			if (not buf) then return nil; end		-- don't have this
			buf=buf[loc];							-- dive
			i=i+1;
			loc=base.__METADATA__["sub"..i];
		end
		if (base==buf) then							-- found it
			cache=v;
			for j=1,i-1 do
				loc=base.__METADATA__["sub"..j];
				cache[loc]=cache[loc] or {};		-- verify
				cache=cache[loc];					-- follow
			end
			return cache;
		end
	end
	return nil;																	-- no cache for this entry
end

-- applies correct decompressor and handles cache
-- Table:Read(DM_WHO,k,t,t.__METADATA__.sub1,t.__METADATA__.sub2);
-- Table:Read(_,entry,base)
function DM.Table.Read(base,entry)
	local cache=DM.Table.GetCache(base);
	if (cache and cache[entry]) then return cache[entry]; end			-- got it, so use it
	if (not base.__DATA__[entry]) then return nil; end					-- entry does not exist, so nil
	local raw=base.__DATA__;											-- go to raw level
	if (type(raw[entry])~="string") then								-- not a packed entry
		if (type(raw[entry])=="table") then return raw[entry]; end		-- use unpacked entry directly
		_debug("debug: Got "..type(raw[entry]).." as entry");
		return nil;														-- wrong entry type
	end
	local buf;
	if (base.__METADATA__.comp==1) then buf=DM.Code:Decompress(raw[entry]); else buf=raw[entry]; end
	local ver=buf:sub(2,2) or "0";										-- get compressor version
	local fnDecompress=DM.Table["DecompressV"..ver];					-- find decompressor
	if (not fnDecompress) then return nil; end							-- no decompressor for this data
	if (not cache) then wipe(DM.Table.Scrap); cache=DM.Table.Scrap; end	-- not using cache, so use scrapbook for decompression
	if (ver>="3") then cache[entry]=fnDecompress(buf:sub(3));			-- decompress to cache/scrap
	else cache[entry]=fnDecompress(DM.Table,{},buf:sub(3),true); end	-- decompress to cache/scrap - old schema
	if (base.__METADATA__.datameta) then
		setmetatable(cache[entry],_G[base.__METADATA__.datameta]);		-- got entry-level meta function
	end
	return cache[entry];
end

function DM.Table.PurgeCache(root) wipe(DM.Table.Addon[root] or {}); end		-- in case it's unregistered

DM.Code={};

--	Code with password. It's not complicated, but needs the password
--	to decode, and is totally unreadable after coding.
--		code=mod(byte+seed+pw[next],256);
--		byte=mod(code-pw[next]-seed,256);
function DM.Code:Protect(text,pw)
	local _char,_mod,_byte,_insert=string.char,mod,string.byte,table.insert;
	if (not pw or pw=="") then return text; end
	local seed,pwlen,code,hash,val,pi,textlen,pieces=0,pw:len(),"",0,0,1,text:len(),{}; for i=1,pwlen do seed=seed+_byte(pw,i); end
	for i=1,textlen do
		hash=hash+_byte(text,i);
		_insert(pieces,_char(_mod(_byte(text,i)+(seed+_byte(pw,pi)),256)));
		pi=pi+1; if (pi>pwlen) then pi=1; end
	end
	return string.format("%X:",_mod(hash,0x10000))..table.concat(pieces);
end

function DM.Code:Unprotect(code_,pw)
	local _char,_mod,_byte,_insert=string.char,mod,string.byte,table.insert;
	if (not pw or pw=="") then return code; end
	local hash,code=code_:match("^(%x+):(.+)$");
	if (not hash) then								-- old password protection
		local decoded;
		code,decoded=self:Decode(code_);			-- the old protection applied Base64 at the end
		if (decoded~=true) then return code; end
	else
		hash=tonumber(hash,16);						-- grab hash of unprotected data
	end
	local seed,pwlen,text,texthash,pi,value,pieces,codelen=0,pw:len(),"",0,1,nil,{},code:len(); for i=1,pwlen do seed=seed+_byte(pw,i); end
	for i=1,codelen do
		value=_mod(_byte(code,i)-(seed+_byte(pw,pi)),256);
		while(value<0) do value=value+256; end						-- normalise
		_insert(pieces,_char(value));
		pi=pi+1; if (pi>pwlen) then pi=1; end
		if (hash) then texthash=texthash+value; end					-- we're doing hash check as well
	end
	if (hash and hash~=_mod(texthash,0x10000)) then return ""; end				-- wrong hash
	return table.concat(pieces);
end

DM.Code.Base64scrap={};
DM.Code.Base64={"A","B","C","D","E","F","G","H","I","J","K","L","M","N","O","P","Q","R","S","T","U","V","W","X","Y","Z","a","b","c","d","e","f","g","h","i","j","k","l","m","n","o","p","q","r","s","t","u","v","w","x","y","z","0","1","2","3","4","5","6","7","8","9","+","/"};
DM.Code.Base64rev={};
for k,v in ipairs(DM.Code.Base64) do DM.Code.Base64rev[v]=k-1; end

function DM.Code:Code(text)
	local _insert,_lshift,_rshift,_band,_len=table.insert,bit.lshift,bit.rshift,bit.band,string.len;
	local pad=text:len()%3; if (pad>0) then pad=3-pad; end
	text=text..string.rep(string.char(0),pad);
	local list=self.Base64scrap; wipe(list);
	local p1,p2,p3,p4;
	for i=1,_len(text),3 do
		p1=_lshift(text:byte(i),16)+_lshift(text:byte(i+1),8)+text:byte(i+2);
		p4=_band(p1,63); p1=_rshift(p1,6);
		p3=_band(p1,63); p1=_rshift(p1,6);
		p2=_band(p1,63); p1=_rshift(p1,6);
		_insert(list,self.Base64[p1+1]);	-- lua offsets by 1
		_insert(list,self.Base64[p2+1]);
		_insert(list,self.Base64[p3+1]);
		_insert(list,self.Base64[p4+1]);
	end
	if (pad>=1) then list[(#list)]="="; end
	if (pad>=2) then list[(#list)-1]="="; end
	return table.concat(list);
end

function DM.Code:Decode(code)
	local _lshift,_rshift,_sub,_char,_len,_insert=bit.lshift,bit.rshift,string.sub,string.char,string.len,table.insert;
	if (code:len()%4~=0) then return code,false; end
	local pad=0; if (code:find("==",-2,true)) then pad=2; elseif (code:find("=",-1,true)) then pad=1; end
	local com;
	local text="";
	local p1,p2,p3,p4;
	local pieces={};
	for i=1,_len(code),4 do
		p1=self.Base64rev[_sub(code,i  ,i  )];
		p2=self.Base64rev[_sub(code,i+1,i+1)];
		p3=self.Base64rev[_sub(code,i+2,i+2)] or 0;
		p4=self.Base64rev[_sub(code,i+3,i+3)] or 0;
		com=_lshift(p1,18);
		com=com+_lshift(p2,12);
		com=com+_lshift(p3,6);
		com=com+p4;
		_insert(pieces,_char(_rshift(com,16)%256));
		_insert(pieces,_char(_rshift(com,8)%256));
		_insert(pieces,_char(com%256));
	end
	text=table.concat(pieces);
	if (pad>0) then
		text=text:sub(1,-1-pad);
	end
	return text,true;
end

local bitlookup={ [1]=1 };
for i=2,256 do table.insert(bitlookup,((bitlookup[i-1]+1)*2)-1); end
function DM.Code.Bits(value,bits)
	if (value>bitlookup[bits]) then return bits+1; end
	return bits;
end

-- dictionary: compress
-- key: uncompressed data
-- val: compressed token
local dmcodecompress={};
function DM.Code:Compress(text)
	local _char,_insert,_concat,_floor,_sub,_len,_lshift,_bits=string.char,table.insert,table.concat,floor,string.sub,string.len,bit.lshift,DM.Code.Bits;
	local dictionary={};								-- start new dictionary
	for i=0,255 do dictionary[_char(i)]=i; end	-- init all possible first bytes
	local top=255;										-- upper token
	local bits=_bits(top,8);						-- needed bits for upper token
	local out,unused="",0;								-- output string, unused bits at the end
	local extlen;
	-- look for uncompressed in dictionary
	local find=function(index)
		local thislen,l=_len(text),1;					-- full length, 1 character
		local short;									-- nothing - yet
		local long=_sub(text,index,index);				-- longest test is just one character yet
		while(true) do
			short,long=long,_sub(text,index,index+l);	-- copy longest to shortest
			if (not dictionary[long] or index+l>thislen) then	-- check for the longest and if we're at the end
				return short;							-- longest doesn't exist, so use the last that existed
			end
			l=l+1;										-- advance to next
		end
	end
	-- add compressed token to output
	local add=function(token)
		wipe(dmcodecompress);
		local lshift=unused-bits;						-- 6 unused, 5 to insert, 1 shift left => 1 bit unused
														-- 5 unused, 6 to insert, 1 shift right => 7 unused
		if (lshift<0) then
			while(lshift<0) do lshift=lshift+8; end		-- shuffle - not important how many bytes
		end
		if (lshift>0) then token=_lshift(token,lshift); end	-- position if needed
		extlen=bits+lshift;
		if (extlen>8) then								-- crosses a byte boundary after positioning
			while(extlen>8) do							-- until it fits within a single byte
				_insert(dmcodecompress,1,_char(token%256));	-- add lowest byte to top
				token=_floor(token/256);				-- shave 8 bits off the end
				extlen=extlen-8;
			end
		end
		if (unused>0) then								-- need to blend data
			_insert(dmcodecompress,1,_char(bit.bor(out:byte(-1),token)));	-- merge last byte with new token upper part
			_insert(dmcodecompress,1,_sub(out,1,-2));						-- get all without last byte
		else
			_insert(dmcodecompress,1,_char(token));		-- add left-aligned token
			_insert(dmcodecompress,1,out);				-- get all previous output
		end
		out=_concat(dmcodecompress);					-- spew
		unused=lshift;									-- set new unused space
	end
	-- create token
	local new=function(i,len)
		top=top+1;
		bits=_bits(top,bits);					-- needed bits for upper token
		dictionary[_sub(text,i,i+(len-1))]=top;			-- add new token
	end

	-- run compressor
	local i=1;
	local tmp,len;
	while (i<=_len(text)) do
		tmp=find(i);									-- look for compressed token - it will always exist
		add(dictionary[tmp]);							-- add token to output
		len=_len(tmp)+1;								-- set length for new uncompressed chunk
		if (i+len<=_len(text)) then new(i,len); end		-- add token if there's more to compress
		i=i+(len-1);
	end
	return out;
end

-- dictionary: decompress
-- key: compressed data
-- val: uncompressed token
function DM.Code:Decompress(raw)
	local _ceil,_floor,_len,_byte,_rshift,_sub,_band,_bits=ceil,floor,string.len,string.byte,bit.rshift,string.sub,bit.band,DM.Code.Bits;
	local dictionary={};								-- start new dictionary
	for i=0,255 do dictionary[i]=string.char(i); end	-- init all possible first bytes
	local top=255;										-- upper token
	local bits=_bits(top,8);						-- needed bits for upper token
	local unused=0;										-- unused bits to the left
	local offset=0;										-- no bytes read yet
	local out="";										-- decompressed data
	-- grab next uncompressed data
	local grab=function()
		local bytes=_ceil((bits+unused)/8);				-- number of full bytes to read
		if (offset+bytes>_len(raw)) then offset=offset+bytes;
		return ""; end
		local token=0;									-- init token
		for i=1,bytes do								-- all token bytes
			token=(token*256)+_byte(raw,offset+i);		-- accumulate little-endian
		end
		offset=offset+_floor((bits+unused)/8);			-- number of full bytes completed
		local ending=unused+bits;						-- get size from left
		while(ending>7) do ending=ending-8; end			-- normalise to within byte
		if (ending>0) then								-- if there's right-most unused bits...
			token=_rshift(token,8-ending);			-- shift right to align first bit
		end
		token=_band(token,(2^bits)-1);				-- cut upper used bits
		local val=dictionary[token];					-- grab current uncompressed data
		if (top>255) then									-- something has previously been added
			dictionary[top]=dictionary[top].._sub(val,1,1);	-- add first byte from this data as last byte of previous data
			if (top==token) then val=dictionary[top]; end	-- need last one, so update grabbed value
		end
		-- create new token
		top=top+1; bits=_bits(top,bits);			-- needed bits for upper token
		dictionary[top]=val;							-- add new token without last byte
		unused=ending;									-- used for this token, so remove for next
		return val;
	end

	-- run decompressor
	local len=_len(raw);
	while(offset<len) do out=out..grab(); end			-- concatenate this uncompressed data
	return out;
end

function packtable(t,n)
	if (not t) then t={}; end
	t="03"..DM.Table.CompressV3(t);
	if (n) then t=DM.Code:Protect(t,NATIVE); end
	return t;
end
function unpacktable(s,n)
	if (type(s)=="table") then return DM:CopyTable(s); end
	if (not s or s=="") then return {}; end
	if (n) then s=DM.Code:Unprotect(s,NATIVE); end
	local fnDecompress=DM.Table["DecompressV"..s:sub(2,2)];	-- Look for decompressor
	return fnDecompress(s:sub(3));
end

function DM:Tick(elapsed)
	DM.MT.FastMT=elapsed*0.99;				-- aim at slighly lower fps for fast switching
	if (DM.MT.FastMT>DM.MT.Speed) then DM.MT.FastMT=DM.MT.Speed; end	-- limit to slowest MT
	DM.MT:Next();		-- Run multi-threading
end

function DM:Init()
	if (not self.TheFrame) then
		self.TheFrame=CreateFrame("Frame","DuckMod-Core-Messager"..TV,UIParent);
		if (not self.TheFrame) then DM:Chat("Could not create DuckMod Core UI frame",1); return; end
--		self.TheFrame:SetScript("OnEvent",self.OnEvent);
		self.TheFrame:SetScript("OnUpdate",self.Tick);
	end
	if (not self.TheFrameWorldMap) then
		self.TheFrameWorldMap=CreateFrame("Frame","DuckMod-CoreWM-Messager"..TV,WorldMapDetailFrame);
		if (not self.TheFrameWorldMap) then DM:Chat("Could not create DuckMod Core WM frame",1); return; end
		self.TheFrameWorldMap:SetScript("OnUpdate",self.Tick);
	end

	-- As LUA random can be dodgy at times, it is recommended to make a few calls to it to get it going.
	math.random(); math.random(); math.random();

	if (DM.Com2) then DM.Com2:Init(); end
	if (DM.Database) then DM.Database:Init(); end
	if (DM.WoWnet) then DM.WoWnet:Init(); end
	if (DM.Render) then DM.Render:Init(); end
	if (DM.BrowserServer) then DM.BrowserServer:Init(); end
end
