﻿--[[
	DuckMod - A World of Warcraft add-on library
	Copyright (C) 2009-2016  Dag Bakken

	DuckMod is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	DuckMod is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with DuckMod.  If not, see <http://www.gnu.org/licenses/>.
]]


local TV=3.1;

if (not DuckMod) then DuckMod = {}; end			-- Make it if empty
if (not DuckMod[TV]) then					-- Don't mess with it if it's loaded
	DuckMod[TV]={};
end
local DM=DuckMod[TV];
local _;
DM.RenderWnCounter=1000;

-- Helper colours
DM.Color = {
	White="|cFFFFFFFF",
	Green="|cFF00FF00",
	Red="|cFFFF0000",
	lBlue="|cFF6060FF",				-- Light blue
	hBlue="|cFEA0A0FF",				-- Highlight blue
	Yellow="|cFFFFFF00",
};

DM.RenderQueue={};
DM.Render={
	class={
		-- "=" denotes the named class
		-- "+" denotes included entities
		HaveContainer  = 0x00000001, -- - - - + + - - - - - - - - - - -
		UIObject	   = 0x00000002, -- + + + + + + + + + + + + + + + +
		FontInstance   = 0x00000004, -- = - - + - - + - - + + - - - - +
		Region         = 0x00000008, -- - = + + + + + + + + + + + + + +
		LayeredRegion  = 0x00000010, -- - - = + + - - - - - - - - - - -
		FontString     = 0x00000020, -- - - - = - - - - - - - - - - - -
		Texture        = 0x00000040, -- - - - - = - - - - - - - - - - -
		Frame          = 0x00000080, -- - - - - - = + + + + + + + + + +
		EditBox        = 0x00000100, -- - - - - - - = - - - + - - - - +
		Button         = 0x00000200, -- - - - - - - - = - - - - + + - -
		SimpleHTML     = 0x00000400, -- - - - - - - - - = - - - - - - -
		ScrMsgFrame    = 0x00000800, -- - - - - - - - - - = - - - - - -
		SwapToLink     = 0x00001000, -- - - - - - - - - - - + - - - - -
		DeadEditBox    = 0x00002000, -- - - - - - - - - - - = - - - - -
		NoAutoEvent    = 0x00004000, -- - - - - - - - - - - - - - - - -
		List           = 0x00008000, -- - - - - - - - - - - - = - - - -
		Extended       = 0x00010000, -- - - - - - - - - - - - - + + - -
		ChildText      = 0x00020000, -- - - - - - - - - - - - - + + - -
		CheckBox       = 0x00040000, -- - - - - - - - - - - - - = + - -
		RadioButton    = 0x00080000, -- - - - - - - - - - - - - - = - -
		Bar            = 0x00100000, -- - - - - - - - - - - - - - - = -
		MultilineEdit  = 0x00200000, -- - - - - - - - - - - - - - - - =

		FontString_    = 0x0000003F, -- Full FontString
		Texture_	   = 0x0000005B, -- Full Texture
		iTexture_      = 0x000042DA, -- Full button and texture with ItemButtonTemplate, no auto event
		Frame_         = 0x0000008A, -- Full Frame
		EditBox_       = 0x0000018E, -- Full EditBox
		Button_        = 0x0000028A, -- Full Button
		TextLink_      = 0x0000083F, -- FontString + SwapToLink
		SimpleHTML_    = 0x0000148E, -- Full SimpleHTML + swap-flag
		ScrMsgFrame_   = 0x0000088E, -- Full ScrollingMessageFrame
		DeadEditBox_   = 0x0000318E, -- Full editBox with no interaction
--		List_          = 0x0000808A, -- Full DM_01 list
		CheckBox_      = 0x0007028A,	-- Full native checkbox extended
		RadioButton_   = 0x000F028A,	-- Full native checkbox extended with radiobutton
		Bar_           = 0x0010008A, -- A graphic bar
		MultilineEdit_ = 0x0020018E, -- Full multiline editbox
	},
	frameAPI={},					-- API for a created frame with canvas
};
local ren=DM.Render;

function ren:Init()
	if (not self.TheFrame) then
		self.TheFrame=CreateFrame("Frame","DuckMod-Render-Messager"..TV,UIParent);
		if (not self.TheFrame) then DM:Chat("Could not create DuckMod rendering UI frame",1); return; end
	end
	if (not self.TheFrameWorldMap) then
		self.TheFrameWorldMap=CreateFrame("Frame","DuckMod-RenderWM-Messager"..TV,WorldMapDetailFrame);
		if (not self.TheFrameWorldMap) then DM:Chat("Could not create DuckMod rendering WM frame",1); return; end
	end
	self.TheFrame:SetScript("OnUpdate",self.Tick);
	self.TheFrameWorldMap:SetScript("OnUpdate",self.Tick);
end

function ren:CreateFrame(parent,name,noscroll)
	if (not name) then name=tostring(random()); end
	local sf;
	if (noscroll) then sf=CreateFrame("Frame",name,parent);
	else sf=CreateFrame("ScrollFrame",name,parent,"UIPanelScrollFrameTemplate"); end
	sf.Canvas=CreateFrame("Frame",name.."Canvas",sf);
	if (not noscroll) then sf:SetScrollChild(sf.Canvas); end
	sf.___noscroll=noscroll;
	sf.Canvas:SetSize(64,32);
	sf.Canvas:SetPoint("TOPLEFT");
	sf:SetScript("OnSizeChanged",ren.frameAPI.iFrameSizeChange);
	sf.Contents=ren.frameAPI.SetSectionData;
	ren.SetRenderingEngine(sf);
	return sf;
end

function ren.frameAPI.iFrameSizeChange(frame,w,h)
	if (not frame.Canvas or not frame.Canvas.SetWidth) then return; end
	frame.Canvas:SetWidth(w-2);											-- apply
	if (frame.Canvas:GetHeight()<2) then frame.Canvas:SetHeight(2); end	-- make sure it's bigger than nothing
	ren:Render(frame.Canvas,frame.___LastAppliedData);					-- redraw
end

function ren.frameAPI.SetSectionData(frame,data)
	frame.___LastAppliedData=data;
	ren:Render(frame.Canvas,frame.___LastAppliedData);					-- redraw
end

function ren:Render(frame,sectiondata)
	if (type(sectiondata)~="string") then return; end
	table.insert(DM.RenderQueue,{frame=(frame.Canvas or frame),sectiondata=sectiondata});
end

-- Check the return value for success. "nil" is fail. "true" is victory.
-- If "nil", retry next update-frame.
function ren:Render_S(frame,sectiondata)
	if (not frame or ren.Canvas) then return nil; end	-- Could not render at this time
	if (type(sectiondata)=="table") then			-- SECTIONDATA
	elseif (type(sectiondata)=="string") then		-- SECTIONDATA.Data
		sectiondata={
			Data=sectiondata,
			Keys="",
			Description="Anonymous data",
			Meta={
				Server="",
				Location="",
			},
		};
	else DM.RenderWnCounter=1000; return; end

	ren.Canvas=true;
	frame.SECTIONDATA=sectiondata;
	ren.SetRenderingEngine(frame);
	DM.RenderWnCounter=0;
	DM.RenderWnRemaining=sectiondata.Data;
	if (frame:GetParent().SetVerticalScroll) then frame:GetParent():SetVerticalScroll(0); end
	ren:Render_2(frame,DM.RenderWnRemaining);
end

function ren.SetRenderingEngine(frame)
	if (frame.WnRenderData) then return; end
	frame.WnRenderData={};
	frame.WnRenderData.Widget={};
	frame.WnRenderData.Focus={};
	frame.WnRenderData.WnContext={};
--	frame.GoTo=ren.SetCanvasFocus;
	frame.GoTo=function(frm,...) frm:Queue("SetCanvasFocus",...); end		-- ren.SetCanvasFocus(self,id)
	frame.GetForm=ren.GetContextData;
	frame.GetObject=ren.GetObject;
	frame.GetObjectData=ren.GetObjectData;
	frame.Queue=function(...) ren.QueueFunction(...); end	-- func name [,callback] ,args
end

function ren.QueueFunction(frame,f,cb,...)
	if (type(cb)=="function") then table.insert(DM.RenderQueue,{frame=(frame.Canvas or frame),cb=cb,f=f,arg={...}});	-- callback provided
	else table.insert(DM.RenderQueue,{frame=(frame.Canvas or frame),f=f,arg={cb,...}}); end								-- no callback
end

local CallbackCount=3;		-- do callback at the count of three

function ren:Render_2(canvas,remaining)
--	DM.MT:Run("WoWnet Render",ren.Render_MT_DO,ren,canvas,remaining)
	ren:Render_MT_DO(canvas,remaining);
end

function ren:Render_MT_DO(canvas,remaining)
	if (not canvas) then
		ren.Canvas=nil; DM.RenderWnCounter=1000; return nil;			-- fail
	end
	DM.RenderWnCounter=DM.RenderWnCounter+1;
	canvas.WnRenderData.LockH=0;			-- Far left
	-- Do blocks
	local entryname,param,data;
	entryname,param,remaining=ren:PullSection(remaining);
	if (entryname~="<wn>") then ren.Canvas=nil; DM.RenderWnCounter=1000; return nil; end			-- Wrong data-type waiting
	ren:InitPage(canvas);	-- This will work, as the FontString widget will be set properly before a new redraw
	while (remaining~="") do
		entryname,param,data,remaining=ren:PullSection(remaining);
		if (entryname=="<meta>") then
			local remM=data;						-- Do the meta data
			while (remM~="") do
				entryname,param,data,remM=ren:PullSection(remM);
				if (entryname=="<elements>") then
					local remE=data;				-- Do the elements data
					while (remE~="") do
						entryname,param,data,remE=ren:PullSection(remE);
						-- Overlay parameters
						canvas.WnRenderData.Meta["<elements>"][entryname]=ren:Mix(param,canvas.WnRenderData.Meta["<elements>"][entryname]);
						canvas.WnRenderData.Meta["<elements>"][entryname].class=tonumber(canvas.WnRenderData.Meta["<elements>"][entryname].class,16);
						ren:SetInterface(canvas,entryname,canvas.WnRenderData.Meta["<elements>"][entryname].class)
					end
				elseif (entryname=="<font>") then
					local name="Default";
					if (param.name) then name=param.name; end
					param.name=nil;
					canvas.WnRenderData.Meta.font[name]=ren:Mix(param,canvas.WnRenderData.Meta.font[name]);
				elseif (entryname=="<page>") then
					canvas.WnRenderData.Meta[entryname]=ren:Mix(param,canvas.WnRenderData.Meta[entryname]);
				elseif (entryname=="<package>") then
					if (WoWnet.Scheme[data]) then
						-- load the scheme
					end
				else													-- any non-nested entries
					canvas.WnRenderData.Meta[entryname]=ren:Mix(param,canvas.WnRenderData.Meta[entryname]);
				end
			end
		elseif (entryname=="<body>" or entryname=="<render>") then
			if (not canvas.MeasureFontString) then
				canvas.MeasureFontString=ren:GetWidget(canvas,"<stext>",canvas.WnRenderData.Meta["<elements>"]["<stext>"],nil);
				if (not canvas.MeasureFontString) then
					DM:Chat("Could not create a DuckMod measurement frame",1);
					ren.Canvas=nil; DM.RenderWnCounter=1000; return nil;
				end
				canvas.MeasureFontString.SPECIALNOREUSE=true;
			end
			-- Set the page standards from meta
			if (canvas.WnRenderData.Meta["<page>"]) then
				local root=canvas.WnRenderData.Meta["<page>"];
				if (root.background) then
					local bgwidget=ren:GetWidget(canvas,"<img>",canvas.WnRenderData.Meta["<elements>"]["<img>"],nil);	-- create an <img> element for background
					bgwidget:SetDrawLayer("BACKGROUND");
					bgwidget:ClearAllPoints();
					bgwidget:SetPoint("TOPLEFT",canvas:GetParent(),"TOPLEFT");
					bgwidget:SetPoint("BOTTOMRIGHT",canvas:GetParent(),"BOTTOMRIGHT");
					local r,g,b,a=ren:SplitColor(root.background);	-- Try to make RGBA
					if (r) then bgwidget:SetTexture(r,g,b,a);							-- Single color
					else bgwidget:SetTexture(root.background); end		-- Image-file
					bgwidget:Show();
				end
				if (root.height) then canvas:GetParent():SetHeight(tonumber(root.height)); end
				if (root.width) then canvas:GetParent():SetWidth(tonumber(root.width)); end
			end
			-- Do the contents from data
			local remB=data;						-- Do the body data
			while (remB~="") do
				entryname,param,data,remB=ren:PullSection(remB);
				if (not entryname) then
					DM:Chat("Unknown tag: ?",1);
				elseif (entryname=="<set>") then
					ren:SetContext(canvas,param);	-- Non-visual elements
				elseif (entryname=="<focus>") then
					if (param.id) then canvas.WnRenderData.Focus[param.id]=canvas.WnRenderData.Bottom; end
				elseif (not canvas.WnRenderData.Meta["<elements>"][entryname]) then
					DM:Chat("Unknown tag: "..entryname,1);
				else
					-- handle embedded html
					if (data and data:lower():find("<html>",1,true)) then
						local a,b,c,d=data:lower():find("<html>",1,true),data:lower():find("<body>",1,true),data:lower():find("</body>",1,true),data:lower():find("</html>",1,true);
						if (a and b and c and d and a<b and b<c and c<d) then	-- all present in correct order
							data=strtrim(data:sub(a,(d+7)-a));				-- crop html code
							while(entryname~="<body>") do					-- <body> exists so find it
								entryname,_,data,_=ren:PullSection(data);	-- dig inward until its found
								data=strtrim(data or "");
							end
							--** At this point 'data' contains everything within the body-tags from some html code
							local html={};
							local remH=data;
							local br=1;													-- start with 1 in case the first is a <br> which will need a 2-count immediately
							while(remH~="") do
								if (remH:find("<",1,true)==1) then						-- it's a tag
									entryname,param,data,remH=ren:PullSection(remH);	-- param is always table - can be empty
									local tag=nil;
									if (entryname=="<p>") then entryname="<text>"; end	-- treat as text - as good an estimate as anything
									if (entryname=="<br>") then							-- line-break handling
										br=br+1;
										if (br>1) then tag="<text> </text>"; end		-- design needs empty line
									else
										br=0;											-- not <br>
										if (next(param)) then										-- format with parameters
											tag=entryname:sub(1,-2);								-- "<element"
											for k,v in pairs(param) do
												if (k=="src" and (data or "")=="") then data=v;		-- use parameter "src" as data
												else tag=tag.." "..k.."='"..v.."'"; end				-- " par='par val'"
											end
											tag=tag..">"..(data or "").."</"..entryname:sub(2);		-- ">data</element>"
										else
											tag=entryname..(data or "").."</"..entryname:sub(2);	-- "<element>data</element>"
										end
									end
									table.insert(html,tag);								-- add interpreted html (nil should not lengthen the table)
								else													-- it's not tag, so assume text
									br=0;												-- not <br>
									if (remH:find("<",1,true)) then						-- there is a tag later
										data=remH:find("<",1,true)-1;
										table.insert(html,"<text>"..strtrim(remH:sub(1,data)).."</text>");	-- add text in wn tags
										remH=remH:sub(data);							-- crop at next tag
									end
								end
								remH=strtrim(remH);										-- clean it up
							end
							table.insert(html,remB);									-- append rest of wn-code to the end
							remB=table.concat(html);									-- spew
						end
					else
						local e=canvas.WnRenderData.Meta["<elements>"][entryname];
						local swap=ren:Flag(canvas,e,"SwapToLink");
						-- Visual elements
						if (not e) then DM:Chat("Missing element <"..entryname..">",1); ren.Canvas=nil; DM.RenderWnCounter=1000; return nil; end
						if (not data) then data=e.defaultdata; end;
						-- * Make parameters
						local mix=ren:Mix(param,e);							-- does not alter sources
						-- * Create a widget, or retrieve an old from a stack.
						local widget=ren:GetWidget(canvas,entryname,e,mix);
						-- * Set provided data
						local r,g,b,a=ren:SplitColor(data);	-- Try to make an RGBA
						if (not r or (entryname~="<img>" and entryname~="<hr>")) then
							if (ren:Flag(canvas,e,"FontInstance")) then
								ren:SetFont(canvas,widget,mix);
							elseif (ren:Flag(canvas,e,"ChildText")) then
								ren:SetFont(canvas,_G[widget:GetName().."Text"],mix);
							end
							if (mix.color) then								-- there's a 'color' parameter provided
								r,g,b,a=ren:SplitColor(mix.color);			-- Try to make an RGBA
								if (r) then
									data=string.format("|C%02X%02X%02X%02X",a*255,r*255,g*255,b*255)..data;	-- set new temporary default colour
									-- NOTE 1: The use of capital 'C' will trigger the engine to keep using this colour
									-- if the displayed text uses the '|r' coding after changing colour locally.
									-- NOTE 2: The parameter 'mix.color' is not used directly as the engine also
									-- accepts formats not natively accepted by widgets.
								end
							end
							if (e.IF.SetData) then						-- there's a method
								widget[e.IF.SetData](widget,ren:toASCII(data));	-- Simulate colon
							end
						else
							if (e.IF.SetData) then						-- there's a method
								widget[e.IF.SetData](widget,r,g,b,a);			-- Simulate colon
							end
						end
						-- * Set it for placeholder if needed
						if (swap) then
							ren:InitWidget(canvas.MeasureFontString);
							ren:CopyTextAttributes(widget,canvas.MeasureFontString);
							canvas.MeasureFontString:SetText(ren:toASCII(data));
						end
						-- * Anchor and size
						ren:SetPosition(canvas,widget,mix,swap);
						-- * set data for lists after size has been set
						if (ren:Flag(canvas,e,"List")) then
							widget[e.IF.SetData](widget,ren:toASCII(data));	-- Simulate colon
						end
						-- * Set it for the actual widget if needed
						if (swap) then
							widget[e.IF.SetData](widget,ren:toASCII(data));	-- Simulate colon
						end
	--[[					if (widget.SetBackdrop) then
							widget:SetBackdrop({
								bgFile = "Interface\\DialogFrame\\UI-DialogBox-Gold-Background",
								edgeFile=nil, tile=false, tileSize=1, edgeSize=0,
								insets={ left=-2, right=-2, top=-2, bottom=-2, },
							});
						end
	]]
						-- * Show it
						widget:Show();
						canvas.WnRenderData.LastWidget=widget;
						-- * Callback for widget
						if (mix and mix["callback"] and DM.RenderWnCounter==CallbackCount) then
							DM:StoreByText(mix["callback"],canvas,widget,DM:CopyTable(mix) or {},DM:CopyTable(canvas.WnRenderData.WnContext) or {},data);	-- canvas,widget,param,context,data
						end
					end
				end
--				DM.MT:Yield();
			end				-- end while
			canvas:SetHeight(canvas.WnRenderData.Bottom+1);
		else
			-- Unknown block
		end
	end
	DM.RenderWnCounter=DM.RenderWnCounter+1;

	-- Finalise the page
	if (DM.RenderWnCounter==4) then ren.Canvas=nil; end				-- Done with this canvas
	return true;
end


function ren:Tick(elapsed)
	if (DM.RenderWnCounter==2) then										-- first pass done
		ren:Render_2(DM.RenderQueue[1].frame,DM.RenderWnRemaining);		-- do second pass
		table.remove(DM.RenderQueue,1);									-- remove it
		return;
	elseif (DM.RenderWnCounter<4) then return; end						-- rendering, so go no further
	if (not DM.RenderQueue[1]) then return; end
	local buf=DM.RenderQueue[1];
	if (buf.sectiondata) then ren:Render_S(buf.frame,buf.sectiondata); return;	-- start new render (no entry delete)
	elseif (buf.f) then
		local ret=ren[buf.f](buf.frame,unpack(buf.arg));
		if (buf.cb) then buf.cb(ret,buf.frame,unpack(buf.arg)); end
	end
	table.remove(DM.RenderQueue,1);										-- remove this entry
end
--	table.insert(DM.RenderQueue,{frame=(frame.Canvas or frame),f="GetContextData",arg={key,val}});
--	table.insert(DM.RenderQueue,{frame=(frame.Canvas or frame),cb=cb,f=f,arg={...}});



-- Deviations to the default set-up of parameters are done by table overlay.
ren.DefaultMeta={
	-- Create the building-blocks
	Package="",											-- A package reference
	["<page>"]={
		background="#0000",								-- #AARRGGBB or filename
--		height=60,										-- window height override
--		width=200,										-- window width override
	},
	font={			-- EXCEPTION: Not bracketed, as it is not a tag-to-entry table
		Default={
			file=[[Fonts\FRIZQT__.TTF]],
			size="14",
			color="#ffff",
			flags="",
		},
		Small={
			size="10",
		},
		Smaller={
			size="16",
		},
		Medium={
			size="22",
		},
		Big={
			size="28",
		},
		Edit={
			color="#FAAA",
		},
	},
	-- Create the elements
	["<elements>"]={
		["<text>"]={
			font="Default",
			pad=3,
			widget="FontString",				-- the widget
			class=ren.class.FontString_,
		},
		["<h1>"]={
			widget="FontString",				-- the widget
			font="Big",
			pad=3,
			class=ren.class.FontString_,
		},
		["<h2>"]={
			widget="FontString",				-- the widget
			font="Medium",
			pad=3,
			class=ren.class.FontString_,
		},
		["<h3>"]={
			widget="FontString",				-- the widget
			font="Smaller",
			pad=3,
			class=ren.class.FontString_,
		},
		["<img>"]={
			widget="Texture",					-- the widget
			pad=3,
			width=32,
			height=32,
			defaultdata=[[Interface\Icons\INV_Misc_QuestionMark]],
			class=ren.class.Texture_,
		},
		["<edit>"]={
			widget="EditBox",					-- the widget
			inheritsys="DuckModREN_EditText",	-- Inherit a frame
			width=200,
			height=25,
			font="Edit",
			pad=3,
			class=ren.class.EditBox_,
		},
		["<button>"]={
			widget="Button",					-- the widget
			inheritsys="UIPanelButtonTemplate",
			width=125,
			height=22,
			pad=3,
			class=ren.class.Button_,
		},
		["<hr>"]={
			widget="Texture",					-- the widget
			width="70%",
			height=1,
			align="center",
			defaultdata="#ffff",
			class=ren.class.Texture_,
		},
		["<pad>"]={
			widget="Texture",					-- the widget
			width=10,
			height=1,
			glue="left",
			defaultdata="",
			class=ren.class.Texture_,
		},
		["<frame>"]={
			widget="Frame",						-- the widget
			width=128,
			height=32,
			class=ren.class.Frame_,
		},

		["<none>"]={							-- dummy
			class=0,
		},
		["<stext>"]={
			font="Default",
			pad=3,
			widget="FontString",				-- the widget
			class=ren.class.FontString_,
		},
		["<ltext>"]={
			font="Default",
			pad=3,
			widget="EditBox",					-- the widget
			class=ren.class.DeadEditBox_,
		},
		["<img_i>"]={
			widget="Button",					-- the widget
			inheritsys="ItemButtonTemplate",
			pad=3,
			class=ren.class.iTexture_,
		},
		["<check>"]={
			widget="CheckButton",				-- the widget
			inheritsys="UICheckButtonTemplate",
			width=100,
			height=25,
			pad=3,
			class=ren.class.CheckBox_,
		},
		["<radio>"]={
			widget="CheckButton",				-- the widget
			inheritsys="UICheckButtonTemplate",
			width=100,
			height=25,
			pad=3,
			class=ren.class.RadioButton_,
		},
		["<bar>"]={
			widget="Frame",						-- the widget
			width=100,
			height=14,
			pad=3,
			class=ren.class.Bar_,
		},
		["<mledit>"]={
			widget="EditBox",					-- the widget
--			inheritsys="DuckModREN_EditText",	-- Inherit a frame
			width="100%",
			height=100,
			font="Edit",
			class=ren.class.MultilineEdit_,
		},
	},
};

-- Each canvas (frame) will have the following additional info:
-- canvas
--	WnRenderData
--	   Widget	 - Storage for widget objects
--	   Meta	   - Meta for this page. That is: combined default and provided.
--	   LastWidget - Last rendered widget
--	   Bottom	 - Current end-of-page
--	   WnContext  - Current page context (<set>)
--	   LockH	  - The current LockH while rendering

-- Since each canvas keeps its own registry, multiple rendered pages is supported.
function ren:InitPage(canvas)
	canvas.WnRenderData.Meta=DM:CopyTable(ren.DefaultMeta);	-- Set default meta
	canvas.WnRenderData.LastWidget=nil;				-- nil is the container frame, and true only for the first widget
	canvas.WnRenderData.Bottom=0;					-- Bottom of reendered page
	canvas.WnRenderData.LockH=0;					-- Last locked horizontal
	canvas:SetScript("OnMouseDown",ren.CanvasClicked);
	wipe(canvas.WnRenderData.WnContext);			-- Current run-length context
	wipe(canvas.WnRenderData.Focus);				-- Clear foci
	-- Clear all used elements
	for wEntry,wTable in pairs(canvas.WnRenderData.Widget) do	 -- All widget groups
		for wwEntry,wwTable in pairs(wTable) do		-- All widgets within group
			wwTable.Used=nil;						-- New page, so all unused
			ren:InitWidget(wwTable.widget);
		end
	end
	-- Set interface for all defaults
	for nElement,dElement in pairs(canvas.WnRenderData.Meta["<elements>"]) do
		self:SetInterface(canvas,nElement,dElement.class)
	end
end

function ren.CanvasClicked(frame,button,down)
	for element,eTable in pairs(frame.WnRenderData.Widget) do
		for name,nTable in pairs(eTable) do
			if (nTable.widget and nTable.widget.IsMouseOver) then
				if (nTable.widget:IsMouseOver()) then		-- We're hovering this widget
					if (nTable.widget.WnParam and nTable.widget.WnParam.click) then
						if (nTable.widget.WnParam.click:lower()=="left" and button=="LeftButton") then
							ren.PageButtonClicked(nTable.widget,button);
							return;
						elseif (nTable.widget.WnParam.click:lower()=="right" and button=="RightButton") then
							ren.PageButtonClicked(nTable.widget,button);
							return;
						elseif (nTable.widget.WnParam.click:lower()=="any") then
							ren.PageButtonClicked(nTable.widget,button);
							return;
					end end
					if (nTable.widget.GetText) then
						-- Walk through all links if any
						local text=nTable.widget:GetText();
						local menu=DMMenuCreate();
						for link,atom in text:gmatch("(|c........|H(.-:.-)|h.-|h|r)") do
							menu:Add(link,function() ren.Widget_OnHyperlinkClicked(nTable.widget,atom,link,button); end);
						end
						menu:Show();		-- will not show if empty
						return;
	end end end end end
end

function ren:InitWidget(widget)
	widget:Hide();					-- Hide all
	if (_G[widget:GetName().."NormalTexture"]) then _G[widget:GetName().."NormalTexture"]:SetTexture(""); end
	if (_G[widget:GetName().."PushedTexture"]) then _G[widget:GetName().."PushedTexture"]:SetTexture(""); end
	if (_G[widget:GetName().."HighlightTexture"]) then _G[widget:GetName().."HighlightTexture"]:SetTexture(""); end
	if (widget.SetBackdrop) then widget:SetBackdrop(nil);end
	if (widget.SetScript) then
		widget:SetScript("OnEnter",nil); widget:SetScript("OnLeave",nil);
		if (widget.GetCursorPosition) then
			widget:SetScript("OnCursorChanged",nil);
		end
	end
	if (widget.SetChecked) then widget:SetChecked(nil); end
	if (widget._Extended) then widget._Extended=0; end
	widget:ClearAllPoints();
	widget:SetHeight(0);
	widget:SetWidth(0);
end

ren.ScrMsgFrame={};
function ren.ScrMsgFrame.SetText(self,text,r,g,b)
	text=ren:toASCII(text)
	self:Clear();
	self:AddMessage(text,r,g,b);
end

ren.Bar={};
function ren.Bar._SetBarData(self,data)
	if (not self.WnParam) then return; end
	data=tonumber(data);
	if (not self.WnParam.dir) then self.WnParam.dir="right"; end
	if (not self.WnParam.max) then self.WnParam.max=100; end
	if (not self.WnParam.fill) then self.WnParam.fill="#ffff"; end
	if (not self.WnParam.empty) then self.WnParam.empty="#0000"; end
	self.WnParam.dir=self.WnParam.dir:lower();
	self.WnParam.max=tonumber(self.WnParam.max);
	local A=_G[self:GetName().."A"];
	local B=_G[self:GetName().."B"];
	local TXT=_G[self:GetName().."TXT"];
	local r,g,b,a;
	r,g,b,a=ren:SplitColor(self.WnParam.fill);		-- try to make an RGBA
	if (not r) then A:SetTexture(self.WnParam.fill); else A:SetTexture(r,g,b,a); end
	r,g,b,a=ren:SplitColor(self.WnParam.empty);		-- try to make an RGBA
	if (not r) then B:SetTexture(self.WnParam.empty); else B:SetTexture(r,g,b,a); end

	A:ClearAllPoints();
	B:ClearAllPoints();
	TXT:ClearAllPoints();
	-- => shrink/stretch alignment
	if (self.WnParam.dir=="left") then
		A:SetPoint("TOPRIGHT",self); A:SetPoint("BOTTOMRIGHT",self);
		B:SetPoint("TOPRIGHT",A,"TOPLEFT"); B:SetPoint("BOTTOMLEFT",self);
		A:SetWidth(((data/self.WnParam.max)*(self.WnParam.width-1))+1);
		TXT:SetPoint("TOPLEFT",B); TXT:SetPoint("BOTTOMRIGHT",A);
	elseif (self.WnParam.dir=="up") then
		A:SetPoint("BOTTOMLEFT",self); A:SetPoint("BOTTOMRIGHT",self);
		B:SetPoint("BOTTOMLEFT",A,"TOPLEFT"); B:SetPoint("TOPRIGHT",self);
		A:SetHeight(((data/self.WnParam.max)*(self.WnParam.height-1))+1);
		TXT:SetPoint("TOPLEFT",B); TXT:SetPoint("BOTTOMRIGHT",A);
	elseif (self.WnParam.dir=="down") then
		A:SetPoint("TOPLEFT",self); A:SetPoint("TOPRIGHT",self);
		B:SetPoint("TOPLEFT",A,"BOTTOMLEFT"); B:SetPoint("BOTTOMRIGHT",self);
		A:SetHeight(((data/self.WnParam.max)*(self.WnParam.height-1))+1);
		TXT:SetPoint("TOPLEFT",A); TXT:SetPoint("BOTTOMRIGHT",B);
	else
		A:SetPoint("TOPLEFT",self); A:SetPoint("BOTTOMLEFT",self);
		B:SetPoint("TOPLEFT",A,"TOPRIGHT"); B:SetPoint("BOTTOMRIGHT",self);
		A:SetWidth(((data/self.WnParam.max)*(self.WnParam.width-1))+1);
		TXT:SetPoint("TOPLEFT",A); TXT:SetPoint("BOTTOMRIGHT",B);
	end
end

ren.FontString={};
function ren.FontString._SetText(self,text)
	text=ren:toASCII(text)
	if (text:find("|C",1,true)) then
		local pos=text:find("|C",1,true);
		text=text.."|C";			-- add one to the end for capture
		local newtext="";			-- the new text
		local color,chunk=text:match("|C(........)(.-)|C");		-- capture capital color key
		repeat
			if (chunk:find("|r",1,true)) then
				chunk=chunk:gsub("|r","|r|c"..color);
			end
			newtext=newtext.."|c"..color..chunk;		-- rebuild
			text=text:gsub("|C(........)(.-)|C","|c"..color..chunk.."|C",1);	-- just one
			color,chunk=text:match("|C(........)(.-)|C");		-- capture capital color key
		until (not color or not chunk);
		self:SetText(text:sub(1,pos-1)..newtext);
	else
		self:SetText(text);
	end
end

ren.CheckBox={};
function ren.CheckBox._SetText(self,text)
	text=ren:toASCII(text)
	_G[self:GetName().."Text"]:SetText(text);
end

ren.SimpleHTML={};
function ren.SimpleHTML._SetText(self,text)
	text=ren:toASCII(text)
	self.DMRUTextContents=text;
	self:SetText(text);
end
function ren.SimpleHTML._GetText(self)
	return self.DMRUTextContents;
end
function ren._SetTexture(widget,...)
	if (not _G[widget:GetName().."IconTexture"]) then return; end
	SetItemButtonTexture(widget,...);	-- probably wont support RGBA
end

function ren.SetCanvasFocus(self,id)
	if (not self.WnRenderData) then return; end
	if (not self.WnRenderData.Focus) then return; end
	if (not self.WnRenderData.Focus[id]) then return; end
	if (self:GetParent().SetVerticalScroll) then
--print(self:GetParent():GetVerticalScrollRange(),self.WnRenderData.Focus[id]);
		local pos=self.WnRenderData.Focus[id];
		if (pos>self:GetParent():GetVerticalScrollRange()) then pos=self:GetParent():GetVerticalScrollRange(); end
		self:GetParent():SetVerticalScroll(pos);
	end
end

--	DM:Chat("Clicked: "..linkdata);	-- essential center - "H" part
--	DM:Chat("Clicked: "..link);		-- Complete link text
--	DM:Chat("Clicked: "..button);	-- "LeftButton"
function ren.Widget_OnHyperlinkClicked(widget,linkdata,link,button)
	ren.SendEvent(widget,"DMEVENT_HYPERLINK_CLICKED",linkdata,link,button);
end
function ren.Widget_OnHyperlinkEnter(widget,linkdata,link)
	ren.SendEvent(widget,"DMEVENT_HYPERLINK_ENTER",linkdata,link);
end
function ren.Widget_OnHyperlinkLeave(widget,linkdata,link)
	ren.SendEvent(widget,"DMEVENT_HYPERLINK_LEAVE",linkdata,link);
end

function ren.SendEvent(widget,event,arg1,arg2,arg3,arg4)
	local frame,handler=nil,nil;
	frame=widget; if (frame and frame.GetScript) then handler=frame:GetScript("OnEvent"); end					-- 0
	if (not handler) then frame=widget:GetParent();											-- 1
		if (frame and frame.GetScript) then handler=frame:GetScript("OnEvent");
			if (not handler) then frame=widget:GetParent():GetParent();						-- 2
				if (frame and frame.GetScript) then handler=frame:GetScript("OnEvent");
					if (not handler) then frame=widget:GetParent():GetParent():GetParent();	-- 3
						handler=frame:GetScript("OnEvent");
						if (not handler) then DEFAULT_CHAT_FRAME:AddMessage("No handler found for a canvas"); return; end
	end end end end end
	handler(frame,event,widget:GetParent(),arg1,arg2,arg3,arg4);	-- simulated self
end

function ren.Widget_OnEnter(widget)
	if (widget.WnParam.tip) then
		ren.SendEvent(widget,"DMEVENT_WIDGET_ENTER",widget.WnParam.tip,widget.WnContext,widget.WnParam);
	end
end

function ren.Widget_OnLeave(widget)
	if (widget.WnParam.tip) then
		ren.SendEvent(widget,"DMEVENT_WIDGET_LEAVE");
	end
end

function ren.Widget_OnCursorChanged(widget)
	if (widget.WnParam.onchange) then
		DM:StoreByText(widget.WnParam.onchange,widget,widget:GetText());
	end
end

function ren.GetContextData(canvas,key,val)
	local data={};
	for _,wg in pairs(canvas.WnRenderData.Widget) do	 -- All widget groups
		for _,wData in pairs(wg) do					-- All widgets within group
			if (wData.Used) then
				local element=wData.widget.WnElement;
				local GetMyData=canvas.WnRenderData.Meta["<elements>"][element].IF.GetData;
				if (GetMyData) then
					if (wData.Used and wData.widget.WnContext[key] and wData.widget.WnContext[key]==val) then
						table.insert(data,{
							param=DM:CopyTable(wData.widget.WnParam),
							context=DM:CopyTable(wData.widget.WnContext),
							data=wData.widget[GetMyData](wData.widget),
							element=element,
							widget=wData.widget,
						});
	end end end end end
	return data;
end

function ren.GetObject(canvas,parameter,value)
	for _,wg in pairs(canvas.WnRenderData.Widget) do	 -- All widget groups
		for _,wData in pairs(wg) do					-- All widgets within group
			if (wData.Used and wData.widget.WnParam and wData.widget.WnParam[parameter]==value) then
				return wData.widget;
	end end end
end

function ren.GetObjectData(canvas,parameter,value)
	local object=canvas:GetObject(parameter,value);
	if (object) then
		local GetMyData=canvas.WnRenderData.Meta["<elements>"][object.WnElement].IF.GetData;
		if (GetMyData) then return object[GetMyData](object); end
	end
end

function ren.PageButtonClicked(button,mbutton)
	if (not button) then DM:Chat("No button provided"); return; end
	local canvas=button.WnCanvas;

	-- run base radiobutton functionality
	if (button.WnContext.group) then
		local valid=nil;
		local form=canvas:GetForm("group",button.WnContext.group);
		for _,d in pairs(form) do
			if (d.widget==button and d.element=="<radio>") then valid=true; break; end
		end
		if (valid) then
			for _,d in pairs(form) do
				if (d.element=="<radio>") then d.widget:SetChecked(nil); end
			end
			button:SetChecked(true);
		end
	end

	if (button.WnParam.click) then									-- param "click" supplied
		if (button.WnParam.click:lower()=="left" and mbutton~="LeftButton") then return; end
		if (button.WnParam.click:lower()=="right" and mbutton~="RightButton") then return; end
	end
	ren.SendEvent(button,"DMEVENT_BUTTON_CLICKED",button.WnContext,button.WnParam,mbutton);
	if (not button.WnParam) then									-- No params in XML
		DM:Chat("No action associated with this button.");
	elseif (button.WnParam.call) then								-- param "call" supplied
		DM:ReadByText(button.WnParam.call,mbutton);					-- make the call
	elseif (button.WnParam.send) then								-- param "send" supplied
		if (button.WnContext[button.WnParam.send]) then				-- It's send-data exists in the context
			local val=button.WnContext[button.WnParam.send];
			local transData={};
			local stored=nil;
			for wEntry,wTable in pairs(canvas.WnRenderData.Widget) do	 -- All widget groups
				local GetMyData=canvas.WnRenderData.Meta["<elements>"][button.WnElement].IF.GetData;
				for wwEntry,wwTable in pairs(wTable) do		-- All widgets within group
					if (canvas.WnRenderData.Meta["<elements>"][wwTable.widget.WnElement].IF.GetData) then
						if (wwTable.Used and wwTable.widget.WnParam.id and wwTable.widget.WnContext[button.WnParam.send]==val) then
							transData[wwTable.widget.WnParam.id]=wwTable.widget[GetMyData](wwTable.widget);
							stored=true;
			end end end end
			-- We've got data, so send it
			if (stored) then
				local canvas=button:GetParent();
				if (not canvas.SECTIONDATA) then return; end			-- No meta-data
				if (not canvas.SECTIONDATA.Meta) then return; end			-- No meta-data
				if (not canvas.SECTIONDATA.Meta.Server) then return; end		-- No server to send to
				-- Make a queue for polling and sending depending on result
				DM.WoWnet.Session:SendPublic("ping:"..canvas.SECTIONDATA.Meta.Server);	-- Do ping
				if (not DM.WoWnet.Session.ResolveWait) then				-- Nothing else waiting
					DM.WoWnet.Session.ResolveWait={
						Table=transData,					-- Actual data to send
						Server=canvas.SECTIONDATA.Meta.Server,	-- Server to send to
						PipeToon=nil;						-- Who's the toon
					};
		end end end
	else
--		DM:Chat("No known action associated with this button.");
	end
end


-- The following code enables context-code, like so:
-- <set form="guild-application"/>  -- Starting form for guild application (multiple forms allowed)
-- <edit req="1"/>				  -- Assigned to form "guild-application" (requires at least one character data)
-- <button send="form"/>			-- Assigned to form "guild-application"
-- <set form="guild-log-in"/>	   -- Starting form for log-in, guild application form is done
-- <edit ID="LIname"/>
-- <edit ID="LIpass"/>
-- <set form=""/>					-- No form active
-- Following this, pressing the button will collect data from all active elements
-- associated with the same context. In this case "form".
function ren:SetContext(canvas,param)
	if (not param) then param={}; end
	canvas.WnRenderData.WnContext=ren:Mix(param,canvas.WnRenderData.WnContext);
	for entry,eVal in pairs(canvas.WnRenderData.WnContext) do
		if (eVal=="") then canvas.WnRenderData.WnContext[entry]=nil; end		-- Remove emptied parameters
	end
end

function ren:Mix(paramA,paramB)
	if (not paramA) then paramA={}; end
	local param=DM:CopyTable(paramB);		-- Make copy
	return DM:CopyTable(paramA,param);		-- Merge and return
end

function ren:CopyTextAttributes(wF,wT)
	wT:SetFont(wF:GetFont());
	wT:SetSpacing(wF:GetSpacing());
end

function ren:toHTML(s)
	s=s or "";
	s=s:gsub("\"","&quot;");
	s=s:gsub("<","&lt;");
	s=s:gsub(">","&gt;");
	s=s:gsub("  ","&nbsp; ");
	s=s:gsub("&","&amp;");
	return s;
end

function ren:toASCII(s)
	s=s or "";
	s=s:gsub("&quot;","\"");
	s=s:gsub("&lt;","<");
	s=s:gsub("&gt;",">");
	s=s:gsub("&nbsp;"," ");
	s=s:gsub("&amp;","&");
	return s;
end

-- Handle "glue" and "flow"
function ren:SetPosition(canvas,widget,param,swap,onlysize)
	if (not param) then param={}; end
	local glue,flow,align,pad=param.glue,param.flow,param.align,param.pad;		-- Make copies
	local bottomtype="stack";
	if (swap) then
		swap=widget;
		widget=canvas.MeasureFontString;
	end
	widget:ClearAllPoints();
	local cWidth=canvas:GetWidth();
	local hSet,wSet;
	if (param.width) then
		local value=tostring(param.width);
		if (value:sub(-1)=="%") then
			value=tonumber(value:sub(1,-2));			-- Get the percentage
			value=(cWidth/100)*value;	-- Convert to points
		end
		if (widget._Extended) then
			widget._Extended=value-param.height;
			value=param.height;
		end
		widget:SetWidth(tonumber(value));				-- Set provided value directly
		wSet=true;
	end
	if (param.height) then
		widget:SetHeight(tonumber(param.height));		-- Set provided value directly
		hSet=true;
	end
	if (onlysize) then return; end

	-- Check provided information
	if (not glue and not flow and not align) then align="LEFT"; end
	if (not align) then align=""; end
	if (not glue) then glue=""; end
	if (not flow) then flow=""; end
	if (not pad) then pad=0; end
	align=align:upper();
	glue=glue:upper();
	flow=flow:upper();

	-- Set text adjustment. "align" has presedence as it does that for positioning as well.
	local adjust;
	adjust="LEFT";		-- Since we are left-to-right readers
	if (flow=="LEFT") then adjust="RIGHT"; end
	if (flow=="DOWN") then adjust="CENTER"; end
	if (align=="LEFT" or align=="RIGHT" or align=="CENTER") then adjust=align; end

	local lastE=0; if (canvas.WnRenderData.LastWidget) then lastE=(canvas.WnRenderData.LastWidget._Extended or 0); end
	local lockside,bottomadd="LEFT",pad;
	-- Align with canvas in stead of widget
	if (align=="LEFT" or (not canvas.WnRenderData.LastWidget and glue=="LEFT")) then
		widget:SetPoint("TOPLEFT",canvas,"TOPLEFT",0,0-(canvas.WnRenderData.Bottom+bottomadd));
		canvas.WnRenderData.LockH=0;
		adjust="LEFT";
	elseif (align=="RIGHT" or (not canvas.WnRenderData.LastWidget and glue=="RIGHT")) then
		widget:SetPoint("TOPRIGHT",canvas,"TOPRIGHT",0,0-(canvas.WnRenderData.Bottom+bottomadd));
		lockside="RIGHT";
		canvas.WnRenderData.LockH=cWidth-0;
		adjust="RIGHT";
	elseif (align=="CENTER" or (not canvas.WnRenderData.LastWidget and glue=="TOP")) then
		widget:SetPoint("TOP",canvas,"TOP",0,0-(canvas.WnRenderData.Bottom+bottomadd));
		lockside="CENTER";
		canvas.WnRenderData.LockH=cWidth/2;
		adjust="CENTER";
	elseif (glue=="LEFT") then
		if (flow=="DOWN") then widget:SetPoint("TOPLEFT",canvas.WnRenderData.LastWidget,"TOPRIGHT",pad+lastE,0); bottomtype="TOPDOWN";
		elseif (flow=="UP") then widget:SetPoint("BOTTOMLEFT",canvas.WnRenderData.LastWidget,"BOTTOMRIGHT",pad+lastE,0); bottomtype="BOTTOMUP";
		else widget:SetPoint("LEFT",canvas.WnRenderData.LastWidget,"RIGHT",pad+lastE,0); bottomtype="CENTERSIDE"; end
		adjust="LEFT";
		bottomadd=0;
	elseif (glue=="RIGHT") then
		if (flow=="DOWN") then widget:SetPoint("TOPRIGHT",canvas.WnRenderData.LastWidget,"TOPLEFT",-(pad+lastE),0); bottomtype="TOPDOWN";
		elseif (flow=="UP") then widget:SetPoint("BOTTOMRIGHT",canvas.WnRenderData.LastWidget,"BOTTOMLEFT",-(pad+lastE),0); bottomtype="BOTTOMUP";
		else widget:SetPoint("RIGHT",canvas.WnRenderData.LastWidget,"LEFT",-(pad+lastE),0); bottomtype="CENTERSIDE"; end
		lockside="RIGHT";
		adjust="RIGHT";
		bottomadd=0;
	elseif (glue=="TOP") then
		if (flow=="DOWN") then widget:SetPoint("TOP",canvas.WnRenderData.LastWidget,"BOTTOM",0,0-(pad+lastE)); lockside="CENTER"; adjust="CENTER";
		elseif (flow=="LEFT") then widget:SetPoint("TOPRIGHT",canvas.WnRenderData.LastWidget,"BOTTOMRIGHT",0,0-(pad+lastE)); lockside="RIGHT"; adjust="RIGHT";
		else widget:SetPoint("TOPLEFT",canvas.WnRenderData.LastWidget,"BOTTOMLEFT",0,0-(pad+lastE)); lockside="LEFT"; adjust="LEFT"; end
	end
	if (widget.SetJustifyV) then widget:SetJustifyV("TOP"); end
	if (widget.SetJustifyH) then widget:SetJustifyH(adjust); end

	-- Expand widget if text is bigger than said area
	if (not wSet and widget.GetStringWidth) then
		local sWidth=widget:GetStringWidth();
		if (lockside=="LEFT") then
			canvas.WnRenderData.LockH=canvas.WnRenderData.LockH+pad;
			if (canvas.WnRenderData.LockH+sWidth>cWidth) then
				widget:SetWidth(cWidth-canvas.WnRenderData.LockH);
			end
		elseif (lockside=="RIGHT") then
			canvas.WnRenderData.LockH=canvas.WnRenderData.LockH-pad;
			if (canvas.WnRenderData.LockH-(sWidth)<0) then
				widget:SetWidth(canvas.WnRenderData.LockH);
			end
		elseif (lockside=="CENTER") then
			if (canvas.WnRenderData.LockH-(sWidth/2)<0) then
				widget:SetWidth(cWidth);
			end
		end
	end
	if (not hSet and widget.GetStringHeight) then
		widget:SetHeight(widget:GetStringHeight());
	end

	-- Set new lock-side
	if (lockside=="LEFT") then
		canvas.WnRenderData.LockH=canvas.WnRenderData.LockH+widget:GetWidth();
	elseif (lockside=="RIGHT") then
		canvas.WnRenderData.LockH=canvas.WnRenderData.LockH-widget:GetWidth();
	elseif (lockside=="CENTER") then
		-- Assume continuous position, even though that may be wrong when the center is center of an element
	end

	-- Set the visually lowest point of the rendering
	canvas.WnRenderData.Bottom=canvas.WnRenderData.Bottom+bottomadd;		-- Add any vertical padding
	if (bottomtype=="stack") then
		canvas.WnRenderData.Bottom=canvas.WnRenderData.Bottom+widget:GetHeight();
	elseif (bottomtype=="TOPDOWN") then
		if (widget:GetHeight()>canvas.WnRenderData.LastWidget:GetHeight()) then
			canvas.WnRenderData.Bottom=canvas.WnRenderData.Bottom+(widget:GetHeight()-canvas.WnRenderData.LastWidget:GetHeight());
		end
	elseif (bottomtype=="CENTERSIDE") then
		if (widget:GetHeight()>canvas.WnRenderData.LastWidget:GetHeight()) then
			canvas.WnRenderData.Bottom=canvas.WnRenderData.Bottom+(widget:GetHeight()-canvas.WnRenderData.LastWidget:GetHeight())/2;
		end
	end

	if (swap) then
		swap:ClearAllPoints();
		swap:SetParent(widget:GetParent());
		swap:SetWidth(widget:GetWidth()*1.0065+0.5);
		swap:SetHeight(widget:GetHeight()+0.5);
		local index=1;
		while(index<=widget:GetNumPoints()) do swap:SetPoint(widget:GetPoint(index)); index=index+1; end
		widget=swap;
	end
end

-- Find an unused widget or create a new
function ren:GetWidget(canvas,element,widget,param)
	if (not param) then param={}; end

	if (element=="<img>" and param.tip) then
		element="<img_i>";
		widget=canvas.WnRenderData.Meta["<elements>"][element];
	end

	local typename=widget.widget;
	local inheritsys=widget.inheritsys;
	if (param.inheritsys) then inheritsys=param.inheritsys; end		-- Note the lower case
	if (not canvas.WnRenderData.Widget[element]) then canvas.WnRenderData.Widget[element]={}; end
	local TheName=nil;
	for wEntry,wTable in pairs(canvas.WnRenderData.Widget[element]) do
		if (not wTable.Used and not canvas.WnRenderData.Widget[element][wEntry].widget.SPECIALNOREUSE) then
			if (canvas.WnRenderData.Widget[element][wEntry].inheritsys==inheritsys) then
				TheName=wEntry;
				break;
	end end end
	-- Create new if needed
	local e=nil;
	if (not TheName) then
		TheName="WoWnet-widget-"..DM:Random(0,1000000);		-- Fractional million
		canvas.WnRenderData.Widget[element][TheName]={};
		e=canvas.WnRenderData.Widget[element][TheName];
		if (self:Flag(canvas,widget,"HaveContainer")) then
			local creator="Create"..typename;
			e.widget=canvas[creator](canvas,TheName);	-- Simulate colon
		else
			e.widget=CreateFrame(typename,TheName,canvas,inheritsys);
			e.inheritsys=inheritsys;
			if (self:Flag(canvas,widget,"Extended")) then
				e.widget._Extended=0;
			end
		end
		if (self:Flag(canvas,widget,"List")) then
--		["<mledit>"]={
--			widget="EditBox",					-- the widget
--			inheritsys="DuckModREN_EditText",	-- Inherit a frame
--			width=200,
--			height=100,
--			font="Edit",
--			class=ren.class.MultilineEdit_,
--		},
		elseif (self:Flag(canvas,widget,"MultilineEdit")) then
			e.widget:SetHyperlinksEnabled(true);
			e.widget:SetTextInsets(-1,-2,-1,0);	-- lrtb
			e.widget:SetMultiLine(true);
			e.widget:EnableKeyboard(true);
			e.widget:SetAutoFocus(nil);
			e.widget._SetText=ren.FontString._SetText;
		elseif (self:Flag(canvas,widget,"SimpleHTML")) then
			e.widget._SetText=ren.SimpleHTML._SetText;
			e.widget._GetText=ren.SimpleHTML._GetText;
		elseif (self:Flag(canvas,widget,"DeadEditBox")) then
			e.widget:SetHyperlinksEnabled(true);
			e.widget:SetTextInsets(-1,-2,-1,0);	-- lrtb
			e.widget:SetMultiLine(true);
			e.widget:EnableKeyboard(nil);
			e.widget:SetAutoFocus(nil);
			e.widget:SetScript("OnHyperlinkClick",ren.Widget_OnHyperlinkClicked);
			e.widget:SetScript("OnHyperlinkEnter",ren.Widget_OnHyperlinkEnter);
			e.widget:SetScript("OnHyperlinkLeave",ren.Widget_OnHyperlinkLeave);
		elseif (self:Flag(canvas,widget,"ScrMsgFrame")) then
			e.widget.SetText=ren.ScrMsgFrame.SetText;
			e.widget:SetHyperlinksEnabled(true);
			e.widget:SetFading(nil);
			e.widget:SetIndentedWordWrap(nil);
			e.widget:SetMaxLines(10000);
			e.widget:SetResizable(nil);	-- user resizable off
		elseif (self:Flag(canvas,widget,"CheckBox")) then
			e.widget._SetText=ren.CheckBox._SetText;
		elseif (self:Flag(canvas,widget,"Bar")) then
			e.widget._SetBarData=ren.Bar._SetBarData;
			e.widget:CreateTexture(TheName.."A");
			e.widget:CreateTexture(TheName.."B");
			e.widget:CreateFontString(TheName.."TXT");
		elseif (self:Flag(canvas,widget,"EditBox")) then
			e.widget:EnableKeyboard(true);
			e.widget:SetPropagateKeyboardInput(nil);
			e.widget:SetScript("OnEscapePressed",function(edit) edit:ClearFocus(); end);
			e.widget:SetScript("OnTabPressed",function(edit) ren.SendEvent(edit,"DMEVENT_EDIT_TABPRESSED",edit:GetText(),edit.WnContext,edit.WnParam); end);
			e.widget:SetScript("OnEnterPressed",function(edit) ren.SendEvent(edit,"DMEVENT_EDIT_ENTERPRESSED",edit:GetText(),edit.WnContext,edit.WnParam); end);
			e.widget:SetScript("OnChar",function(edit,key) ren.SendEvent(edit,"DMEVENT_EDIT_KEYPRESSED",key,edit:GetText(),edit.WnContext,edit.WnParam); end);
			e.widget:SetScript("OnMouseWheel",function(edit,delta) ren.SendEvent(edit,"DMEVENT_EDIT_MOUSEWHEEL",delta,edit.WnContext,edit.WnParam); end);
			e.widget:SetAutoFocus(nil);
			e.widget._SetText=ren.FontString._SetText;
		elseif (self:Flag(canvas,widget,"FontInstance")) then
			e.widget._SetText=ren.FontString._SetText;
		end
	else
		e=canvas.WnRenderData.Widget[element][TheName];
		e.widget:SetParent(canvas);			-- In case it was moved earlier
		wipe(e.widget.WnContext);				-- Clear the context for the element
		e.widget.WnContext=nil;				-- And throw it away
		if (e.widget.WnParam) then
			wipe(e.widget.WnParam);				-- Clear the parameters for the element
			e.widget.WnParam=nil;					-- And throw it away
		end
	end
	-- Set up the widget a bit
	e.widget.WnElement=element;
	if (e.widget.SetDrawLayer) then
		e.widget:SetDrawLayer("ARTWORK");
	end
	e.widget.WnContext=DM:CopyTable(canvas.WnRenderData.WnContext);	-- Insert current context in the element
	e.Used=true;
	e.widget.WnParam=DM:CopyTable(param);
	e.widget.WnCanvas=canvas;
	-- Make fake texture handler for those who can't
	if (not e.widget.SetTexture) then e.widget.SetTexture=ren._SetTexture; end
	if (e.widget.SetNonSpaceWrap) then e.widget:SetNonSpaceWrap(true); end
	-- Set some font for widget with fontstrings
	if (self:Flag(canvas,widget,"FontString")) then
		e.widget:SetFontObject("GameFontNormal");
	end
	ren:InitWidget(e.widget);
	-- Set mouse-over handler for those who can
	if (e.widget.SetScript) then
		e.widget:SetScript("OnEnter",ren.Widget_OnEnter);
		e.widget:SetScript("OnLeave",ren.Widget_OnLeave);
		if (e.widget.GetCursorPosition) then
			e.widget:SetScript("OnCursorChanged",ren.Widget_OnCursorChanged);
		end
	end
	-- Set handlers
	if (self:Flag(canvas,widget,"Button")) then
		-- check for button that must not have a click handler
		if (self:Flag(canvas,widget,"NoAutoEvent") and (param and not param.click)) then
			e.widget:SetScript("OnClick",nil);
		else
			e.widget:SetScript("OnClick",ren.PageButtonClicked);
			if (e.widget.RegisterForClicks) then e.widget:RegisterForClicks("AnyUp"); end
		end
	end

	return e.widget;
end

-- This function takes a parameter table, and assumes that it inherits
-- all the way back to default.
function ren:SetFont(canvas,widget,param)
	local file,size,flags,r,g,b,a;
	if (not param) then param={}; end
	if (not widget.SetFont) then
	end

	-- Start with default settings
	if (canvas.WnRenderData.Meta.font.Default.system) then
		widget:SetFontObject(canvas.WnRenderData.Meta.font.Default.system);
		file,size,flags=widget:GetFont();
	end
	file=canvas.WnRenderData.Meta.font.Default.file or file;
	size=canvas.WnRenderData.Meta.font.Default.size or size;
	flags=canvas.WnRenderData.Meta.font.Default.flags or flags;
	r,g,b,a=self:SplitColor(canvas.WnRenderData.Meta.font.Default.color);
	if (param.font and canvas.WnRenderData.Meta.font[param.font]) then			-- Font supplied and it exists
		file,size,flags,r,g,b,a=self:MakeFont(widget,file,size,flags,canvas.WnRenderData.Meta.font[param.font],canvas.WnRenderData.Meta,r,g,b,a);
	end
	widget:SetFont(file,tonumber(size),flags);
	widget:SetTextColor(r,g,b,a);					-- set it
	widget:SetShadowOffset(0,0);
end

function ren:MakeFont(widget,file,size,flags,font,meta,r,g,b,a)
	if (not font) then return file,size,flags; end
	if (font.system) then												-- system font supplied
		-- Start with system settings
		widget:SetFontObject(font.system);
		file,size,flags=widget:GetFont();
	end
	if (font.inherits and meta.font[font.inherits]) then	-- Inherits from some other font
		file,size,flags,r,g,b,a=ren:MakeFont(widget,file,size,flags,meta.font[font.inherits],meta,r,g,b,a);
	end
	if (font.file) then file=font.file; end
	if (font.size) then size=font.size; end
	if (font.flags) then flags=font.flags; end
	if (font.color) then r,g,b,a=self:SplitColor(font.color); end
	return file,size,flags,r,g,b,a;
end

function ren:SplitColor(color)
	local r,g,b,a;
	if (not color) then return; end
	if (color:find("#",1,true)==1) then
		color=color:sub(2);
		if (color:len()==8) then		-- Standard full-resolution mode 0-255
			a=tonumber("0x"..color:sub(1,2)); if (a>0) then a=a/255; end
			r=tonumber("0x"..color:sub(3,4)); if (r>0) then r=r/255; end
			g=tonumber("0x"..color:sub(5,6)); if (g>0) then g=g/255; end
			b=tonumber("0x"..color:sub(7,8)); if (b>0) then b=b/255; end
		elseif (color:len()==4) then	-- Simple html-mode 0-15
			a=tonumber("0x"..color:sub(1,1)); if (a>0) then a=a/15; end
			r=tonumber("0x"..color:sub(2,2)); if (r>0) then r=r/15; end
			g=tonumber("0x"..color:sub(3,3)); if (g>0) then g=g/15; end
			b=tonumber("0x"..color:sub(4,4)); if (b>0) then b=b/15; end
		end
	end
	return r,g,b,a;
end

function ren:SetInterface(canvas,element,class)
	if (not canvas.WnRenderData.Meta["<elements>"][element]) then return; end
	canvas.WnRenderData.Meta["<elements>"][element].class=class;
	if (not canvas.WnRenderData.Meta["<elements>"][element].IF) then canvas.WnRenderData.Meta["<elements>"][element].IF={}; end
	-- Check for implied interfaces
	if (not bit.band(class,ren.class.UIObject)) then DM:Chat("<"..element.."> does not have UIObject",1); return; end
	if (not bit.band(class,ren.class.Region)) then DM:Chat("<"..element.."> does not have Region",1); return; end
	-- Clear the common programming interface
	canvas.WnRenderData.Meta["<elements>"][element].IF.SetData=nil;		-- No method
	canvas.WnRenderData.Meta["<elements>"][element].IF.GetData=nil;		-- No method

	-- Set the common programming interface
	if (bit.band(class,ren.class.Bar)==ren.class.Bar) then
		canvas.WnRenderData.Meta["<elements>"][element].IF.SetData="_SetBarData";
	elseif (bit.band(class,ren.class.CheckBox)==ren.class.CheckBox) then
		canvas.WnRenderData.Meta["<elements>"][element].IF.SetData="_SetText";
		canvas.WnRenderData.Meta["<elements>"][element].IF.GetData="GetChecked";
	elseif (bit.band(class,ren.class.SimpleHTML)==ren.class.SimpleHTML) then
		canvas.WnRenderData.Meta["<elements>"][element].IF.SetData="_SetText";	-- These are hooked
		canvas.WnRenderData.Meta["<elements>"][element].IF.GetData="_GetText";
	elseif (bit.band(class,ren.class.Texture)==ren.class.Texture) then
		canvas.WnRenderData.Meta["<elements>"][element].IF.SetData="SetTexture";
	elseif (bit.band(class,ren.class.Button)==ren.class.Button) then
		canvas.WnRenderData.Meta["<elements>"][element].IF.SetData="SetText";
		canvas.WnRenderData.Meta["<elements>"][element].IF.GetData="GetText";
	elseif (bit.band(class,ren.class.ScrMsgFrame)==ren.class.ScrMsgFrame) then
		canvas.WnRenderData.Meta["<elements>"][element].IF.SetData="SetText";
	elseif (bit.band(class,ren.class.FontInstance)==ren.class.FontInstance) then
		canvas.WnRenderData.Meta["<elements>"][element].IF.SetData="_SetText";
		canvas.WnRenderData.Meta["<elements>"][element].IF.GetData="GetText";
	elseif (bit.band(class,ren.class.List)==ren.class.List) then
		canvas.WnRenderData.Meta["<elements>"][element].IF.SetData="_SetText";
	elseif (bit.band(class,ren.class.Button)==ren.class.EditBox) then
		canvas.WnRenderData.Meta["<elements>"][element].IF.SetData="SetText";
		canvas.WnRenderData.Meta["<elements>"][element].IF.GetData="GetText";
	else
		canvas.WnRenderData.Meta["<elements>"][element].IF.SetData="SetData";
	end
end

function ren:Flag(canvas,widget,flag)
	if (bit.band(widget.class,ren.class[flag])==ren.class[flag]) then
		return true;
	end
	return nil;
end

function ren:PullParameters(text)
	local param={};
	local start=0;
	text=strtrim(text);
	local pE,dS,dE;		-- parameter end, data start, data end
	local tmp,enclose;
	repeat
		start=start+1;
		pE=text:find("=",start,true)-1;				-- grab the end of this parameter
		enclose=nil;
		tmp=text:find("=\"",start,true);			-- look for double quoted data
		if (tmp==pE+1) then enclose="\"";			-- it's double quoted
		else
			tmp=text:find("='",start,true);			-- look for single quoted data
			if (tmp==pE+1) then enclose="'"; end	-- it's single quoted
		end
		if (enclose) then
			dS=pE+3;								-- point at data start character
			dE=text:find(enclose,dS,true)-1;		-- find next enclosure character
		else
			dS=pE+2;								-- point at data start character
			dE=text:find(" ",dS,true);				-- find next space separator
			if (dE) then dE=dE-1; end				-- point at last character
			dE=dE or text:len();					-- use everything if no such separator exists
		end
		param[text:sub(start,pE):lower()]=text:sub(dS,dE);
		start=dE+((enclose and 2) or 1);			-- Position at space between parameters
	until (text:sub(start,start)~=" ");
	return param;
end


-- * The section must start at the first character
-- * Does NOT support nested same-type top-tags
-- Returns:
--   Section name: String. Contents of first tag without parameters
--	 Parameters: Table. All parameters where entry is param-name and value is string value
--		   Data: The entire contents between starting and ending tags
--	  Remaining: The remainder of the data-block
function ren:PullSection(data)
	local lowData=data:lower();
	local locS=data:find(" ",1,true);
	local locE=data:find(">",1,true);
	-- locTagStart is by definition always 1
	-- locTagNameStart is by definition always 2
	local locTagNameEnd,locTagEnd
	local params=nil;

	if (locS and locS<locE) then	-- With parameters
		locTagNameEnd=locS-1;
		locTagEnd=locE;
		if (data:find("/>",locTagEnd-1,true)==locTagEnd-1) then		-- No data included
			params=self:PullParameters(data:sub(locS+1,locTagEnd-2));
			-- Return the distilled data (tag,params,data,remaining)
			return lowData:sub(1,locTagNameEnd)..">",params,nil,strtrim(data:sub(locTagEnd+1));
		end
		params=self:PullParameters(data:sub(locS+1,locTagEnd-1));
	else 							-- Without parameters
		locTagNameEnd=locE-1;
		locTagEnd=locE;
		if (data:find("/>",locTagEnd-1,true)==locTagEnd-1) then		-- No data or parameters included ( <tag/> )
			locTagNameEnd=locE-2;
			return lowData:sub(1,locTagNameEnd)..">",nil,nil,strtrim(data:sub(locTagEnd+1));
		end
		-- got data, no params
	end
	local locEndTagStart,locEndTagEnd=data:find("</"..data:sub(2,locTagNameEnd)..">",locE,true);
	return lowData:sub(1,locTagNameEnd)..">",params,strtrim(data:sub(locTagEnd+1,locEndTagStart-1)),strtrim(data:sub(locEndTagEnd+1));
end
