﻿--[[
	MessageBoard - A World of Warcraft In-game Forum
	Copyright (C) 2009-2016  Devilhoy

	MessageBoard is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	MessageBoard is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with MessageBoard.  If not, see <http://www.gnu.org/licenses/>.
]]

local _, MessageBoard = ...
local L = MessageBoard.L;
local addonName, PRIVATE = ... -- Blizzard passes us the addon name and a table to all addon files in the ellipsis variable
PRIVATE.Functions = {} -- create a subtable to store our utility functions so we can call them from anywhere in our addon without making them global
local MBFunc = PRIVATE.Functions -- get a reference to our functions table with a friendly/short name
DuckLibMB=DuckMod[3.1];
local DuckLib=DuckLibMB;
local Net=DuckLib.Net;
local DB=DuckLib.Database;
local ren=DuckLibMB.Render;
local mt=DuckLib.MT;
local pairs=d_pairs;
local ipairs=d_ipairs;
local pairs_up=d_pairs_up;
local pairs_dn=d_pairs_dn;
local type=d_type;
local CategoryLevel;
BINDING_HEADER_MESSAGEBOARD = "MessageBoard bindings";
BINDING_NAME_TOGGLEMESSAGEBOARD = "Toggle MessageBoard";
SLASH_MESSAGEBOARD1 = "/mboard";
SLASH_MESSAGEBOARD2 = "/mb";
local MB_PREFIX = "MB3004";
local MB_HEARTBEAT = 1/3;
local L = LibStub("AceLocale-3.0"):GetLocale("MessageBoard")
local version = "3.04"
local iVersion = 3004
local MBDialog = LibStub("LibDialog-1.0")
local Ambiguate = Ambiguate
local _MBLDB
local _minimapIcon
local vface
local MBverRX
local NATIVE="/mboard";
local MBWhere=nil;

MessageBoard_LS = {
	IconPosition=180,
	Debug=false,
	Quoted=nil,
	NewPostNotify=true,
	NewPostNotifyPopup=true,
	Background="Interface\\Addons\\MessageBoard\\Images\\MessageBoardBG-LEGION",
	OptBackground="Interface\\Addons\\MessageBoard\\Images\\MessageBoard_Options-LEGION",
	buttonbg="Interface\\Addons\\MessageBoard\\Images\\buttonbg-LEGION",
	BackgroundName=L["BACKGROUND_LEGION"],
	Draft=nil,
	DraftTitle=nil,
	FontSize=nil,
};

MessageBoard_DB = {
}

MessageBoardLDBIconDB = {
}

MessageBoardXML = {
	Editor={
		ID_MANAGEMENTNOTE = 1, -- management note
		ID_NEWTOPIC       = 2, -- new topic
		ID_NEWMESSAGE     = 3, -- reply
		ID_EDITMESSAGE    = 4, -- edit
		ID_NEWPOLL		  = 5, -- new poll
		ID_NEWPM		  = 6, -- new private message
	},
	Admin={},
	MyRank=nil,
	Administrator=nil;
	Avatar,
};

local MessageBoard = {
	HaveRoster=nil,
	GuildName=nil,
	Ready=nil,
	Selected = {
		Topic=nil,
	},
	Timers = {
		Last=0,
	},
	Administrator=nil,
	selected=nil,
	disabledButton=nil,
	FirstRun=true,
	LoggedIn=false,
};

local Icon={
	ANNOUNCE = "Interface\\Buttons\\UI-GuildButton-MOTD-Disabled",
	ANNOUNCE_new = "Interface\\Buttons\\UI-GuildButton-MOTD-Up",
	STICKY = "Interface\\Buttons\\UI-GuildButton-OfficerNote-Disabled",
	STICKY_new = "Interface\\Buttons\\UI-GuildButton-OfficerNote-Up",
	NORMAL = "Interface\\Buttons\\UI-GuildButton-PublicNote-Disabled",
	NORMAL_new = "Interface\\Buttons\\UI-GuildButton-PublicNote-Up",
	LOCKED = "Interface\\AddOns\\MessageBoard\\Images\\topic-icon-lockedread",
	LOCKED_new = "Interface\\AddOns\\MessageBoard\\Images\\topic-icon-lockedunread",
	STICKYLOCKED = "Interface\\AddOns\\MessageBoard\\Images\\topic-icon-stickylockread",
	STICKYLOCKED_new = "Interface\\AddOns\\MessageBoard\\Images\\topic-icon-stickylockunread",
	POLL = "Interface\\AddOns\\MessageBoard\\Images\\topic-icon-pollread",
	POLL_new = "Interface\\AddOns\\MessageBoard\\Images\\topic-icon-pollunread",
};

function _debug(text,r,g,b)
	if (not MessageBoard_LS.Debug) then return; end
	DuckLib:Chat("MBDEBUG: "..tostring(text),r,g,b);
end

function MessageBoard_OnLoad(self)
	self:RegisterEvent("ADDON_LOADED");
end

function MessageBoard_OnEvent(self,event,...)
	if (MessageBoard[event]) then MessageBoard[event](self,...); end
end

function MessageBoard.ADDON_LOADED(self,addon)
	if (addon~="MessageBoard") then return; end
	self:UnregisterEvent("ADDON_LOADED");
	DuckLib:Init();
	mt:Run("ADDON_LOADED_MT",MessageBoard.ADDON_LOADED_MT,self);
end

function MessageBoard.ADDON_LOADED_MT(self)
	while(not MessageBoard:ValidateBoard()) do mt:Yield(true); end		-- wait here until database is ready
	vface=DuckLib.Com2:CreateInterface("MBMBV",MBverRX);   -- create communication interface
	SlashCmdList["MESSAGEBOARD"]=function(msg) MessageBoard.Slasher(msg) end;
	if MessageBoard_LS.Quoted==nil then MessageBoard_LS.Quoted=false; end -- setup default quoted text
	if MessageBoardLDBIconDB.hide==nil then MessageBoardLDBIconDB.hide=true; end
	if MessageBoard_LS.NewPostNotify==nil then MessageBoard_LS.NewPostNotify=true; end -- setup default notification text
	if MessageBoard_LS.FontSize==nil then MessageBoard_LS.FontSize=14; end -- setup default font size
	if MessageBoard_LS.NewPostNotifyPopup==nil then MessageBoard_LS.NewPostNotifyPopup=true; end -- setup default notification text popup
	if MessageBoard_LS.Background==nil then MessageBoard_LS.Background="Interface\\Addons\\MessageBoard\\Images\\MessageBoardBG-LEGION"; end
	if MessageBoard_LS.OptBackground==nil then MessageBoard_LS.OptBackground="Interface\\Addons\\MessageBoard\\Images\\MessageBoard_Options-LEGION"; end
	if MessageBoard_LS.buttonbg==nil then MessageBoard_LS.buttonbg="Interface\\Addons\\MessageBoard\\Images\\buttonbg-LEGION"; end
	if MessageBoard_LS.BackgroundName==nil then MessageBoard_LS.BackgroundName="Legion"; end
	tinsert(UISpecialFrames,"MessageBoard_Main");
	tinsert(UISpecialFrames,"MessageBoard_OFrame");
	MessageBoard_MainLowerFrameBG:SetTexture(MessageBoard_LS.Background); -- set bg
	for i=1,9 do 
		_G["MessageBoard_MainNavBarButton"..i.."BG"]:SetTexture(MessageBoard_LS.buttonbg)
	end
	MessageBoard_MainNewTopic:SetText(L["MAIN_NEW_TOPIC"]);
	MessageBoard_MainMsgReply:SetText(L["MAIN_NEW_REPLY"]);
	MessageBoard_MainNewPoll:SetText(L["MAIN_NEW_POLL"]);
	MessageBoard_MainUsers:SetText(L["MAIN_USERS"]);
	MessageBoard_MainPMS:SetText("PMs");
	self:RegisterEvent("GUILD_ROSTER_UPDATE");
	self:RegisterEvent("CHAT_MSG_ADDON");
	MessageBoard.LoggedIn=true;
	-- MBFunc.InitializeMinimap()
	-- MBFunc.GetAdmin()
	-- MessageBoard:SetPrunes()
end

function MBFunc.GetAdmin()
	_debug("Get Admin()")
	local info = GetGuildInfoText()
	if not info then
		_debug("no guild info")
		return
	end
	local lines = {string.split("\n", info)}
	--local mbsettings = false
	for _,line in pairs(lines) do
		--if line == "-MB-" then
	        --mbsettings = not mbsettings
		--elseif mbsettings then
			if (line:find("#MB:A:",1,true)) then
				local admins={strsplit(",",line:sub(7))} --sub 11
				for _,toon in ipairs(admins) do
					local name=UnitName("player");
					name = Ambiguate(name,"none")
					if (name==toon) then
						MessageBoard.Administrator=true; MessageBoardXML.Administrator=true;
						break;
					else
						MessageBoard.Administrator=nil; MessageBoardXML.Administrator=nil;
					end
				end
			end
		--end
	end
end

-- function MessageBoard.GUILD_ROSTER_UPDATE(self)
-- _debug("event - GUILD_ROSTER_UPDATE")
	-- MBFunc.GetAdmin()
-- end

function MessageBoard.DMEVENT_HYPERLINK_ENTER(self,canvas,linkdata,link,button)
	if (linkdata:find("http:",1,true)==1) then
		GameTooltip:SetOwner(UIParent,"ANCHOR_CURSOR");
		GameTooltip:AddLine("Click to Copy");
		GameTooltip:Show();
	else
		local ttlink=linkdata..link;
		GameTooltip:SetOwner(UIParent,"ANCHOR_CURSOR");
		GameTooltip:SetHyperlink(ttlink);
		GameTooltip:Show();
	end
end

function MessageBoard.DMEVENT_HYPERLINK_LEAVE(self)
	GameTooltip:Hide();
end

function MessageBoard.DMEVENT_HYPERLINK_CLICKED(self,canvas,linkdata,link,button)
	MBDialog:Register("MessageBoardUrlCopyDialog", {
	  text = L["POPUP_URL_COPY"],
	  buttons = {
		{
		  text = "Close",
		},
	  },
	  editboxes = {
		{
		auto_focus = true,
		width=200,
		},
	  },
	  on_show = function(self, data)
		self.text:SetFormattedText("\n"..L["POPUP_URL_COPY"].."\n")
		self.editboxes[1]:SetText(linkdata)
		self.editboxes[1]:HighlightText()
	  end,
	  hide_on_escape = true,
	  show_while_dead = true,
	})
	if (linkdata:find("http:",1,true)==1) then
		linkdata=string.sub(linkdata,6); -- removes the temporary http: in front
		MBDialog:Spawn("MessageBoardUrlCopyDialog")
	else MessageBoardXML:OnHyperlinkShow(self,linkdata,link,button); end
end

function MessageBoard.DMEVENT_WIDGET_LEAVE(self)
	GameTooltip:Hide()
end

MessageBoard.DMEVENT_WIDGET_ENTER_id={}

function MessageBoard.DMEVENT_WIDGET_ENTER(self,canvas,tip,context,params)
	if (not tip) then return; end										-- just a fail-safe
	if (MessageBoard.DMEVENT_WIDGET_ENTER_id[tip]) then					-- if we have this type of tip, show it
		MessageBoard.DMEVENT_WIDGET_ENTER_id[tip](self,canvas,tip,context,params);
		return;
	end
end

-- show toon information
-- To enable this tooltip for a widget, add these parameters to it:
--    tip=author
--    author='author name'
function MessageBoard.DMEVENT_WIDGET_ENTER_id.author(self,canvas,tip,context,params)
	local stats=dbPath["//MBPR/Users/"..params.author]; if (not stats) then return; end
	GameTooltip:SetOwner(UIParent,"ANCHOR_CURSOR") -- Set the header to the entry-text. This also clears all previous text in it (I think)
	GameTooltip:SetText(stats.nameclasscolor)
	GameTooltip:AddLine(L["TOOLTIP_LEVEL"].." "..tostring(stats.level).." "..stats.class)
	GameTooltip:AddLine(L["TOOLTIP_RANK"].." "..tostring(stats.guildrank))
	GameTooltip:AddLine(L["TOOLTIP_POSTS"].." "..tostring(MessageBoardXML:UserMessageCount(params.author)))
	GameTooltip:SetWidth(425)
	GameTooltip:Show()
end

-- show a text excerpt
-- To enable this tooltip for a widget, add these parameters to it:
--    tip=excerpt
--    topic='topic name'
--    author='author name'
--    term='search-term'
--    text='text with term inside'
function MessageBoard.DMEVENT_WIDGET_ENTER_id.excerpt(self,canvas,tip,context,params)
	if (not params or not params.category or not params.topic or not params.key) then _debug("Malformed XML"); return; end
	GameTooltip:SetOwner(UIParent,"ANCHOR_CURSOR") -- Set the header to the entry-text. This also clears all previous text in it (I think)
	GameTooltip:SetText(params.topic)
	GameTooltip:AddLine(params.author)  -- Set some text
	local term=params.term
	local text=dbPath["//MBPR/Base/"..params.category.."/"..params.topic.."/"..params.key].Text;
	local txts,txte=text:lower():find(term:lower(),1,true)          -- Find the searched term in the text
	if (not txts) then txts=0; txte=2; end
	local showoffstart,showoffend="(...) "," (...)"                 -- Just some show-off in case we need to trunc
	txts=math.floor((txte+txts)/2)-50                               -- Go 50 characters to the left from term center
	if (txts<1) then txts=1; showoffstart=""; end                   -- Clip to beginning
	txte=txts+100; if (text:len()<=txte) then showoffend=""; end    -- Clip to end
	text=text:sub(txts,txte)                                        -- Copy 100 characters
	txts,txte=text:lower():find(term:lower(),1,true)                -- Re-find the term in the new string
	if (txts) then
		text=text:sub(1,txts-1).."|cffffff00"..text:sub(txts,txte).."|r"..text:sub(txte+1)    -- Add the yellow color
	end
	GameTooltip:AddLine(showoffstart..text..showoffend,1,1,1,1)     -- Add text to tooltip
	GameTooltip:SetWidth(425)
	GameTooltip:Show()
end

MessageBoard.DMEVENT_BUTTON_CLICKED_id={};
function MessageBoard.DMEVENT_BUTTON_CLICKED(self,canvas,context,params)
	if (not context or not params) then return; end    -- Data not provided, so something is wrong somewhere
	local base=dbPath["//MBPR/Base"];
	-- poll
	if (context.pollid and params.action=="submit") then
		local form=canvas:GetForm("pollid",context.pollid);
		for _,t in pairs(form) do
			if (t.param.id and t.data) then
				-- you've just found a checked box, so save it
				local buf=base[tonumber(t.context.pollc)][t.context.pollt][t.context.pollm];    -- get the message
				local name=UnitName("player");
				name = Ambiguate(name,"none")
				if (type(buf.Voted)~="table") then buf.Voted={}; end    -- create voting table
				buf.Voted[name]=tonumber(t.param.id);	-- add myself as voter with the vote
				base[tonumber(t.context.pollc)][t.context.pollt][t.context.pollm]=buf;    -- put message back and sync db
				MessageBoard:ViewThread(tonumber(t.context.pollc),t.context.pollt);
				break;    -- only one option allowed selected by this method
			end
		end
		return
	end
	if (context.category) then context.category=tonumber(context.category); end
	if (params.id and MessageBoard.DMEVENT_BUTTON_CLICKED_id[params.id]) then
		MessageBoard.DMEVENT_BUTTON_CLICKED_id[params.id](self,base,canvas,context,params);
		return;
	end
end

function MessageBoard.DMEVENT_BUTTON_CLICKED_id.topic(self,base,canvas,context,params)
	--MessageBoardXML:SetCategoryLevel(params.category)
	MessageBoard:ViewThread(tonumber(params.category),params.topic);
end

function MessageBoard.DMEVENT_BUTTON_CLICKED_id.user(self,base,canvas,context,params)
	MessageBoard_ViewUserStats(params.user)
end

function MessageBoard.DMEVENT_BUTTON_CLICKED_id.pms(self,base,canvas,context,params)
	MessageBoard:ViewPM(params.category,params.topic);
end

function MessageBoard.DMEVENT_BUTTON_CLICKED_id.pm(self,base,canvas,context,params)
	MessageBoardXML.Editor:Init(MessageBoardXML.Editor.ID_NEWPM,params.user,"",{});
	MessageBoardXML.Editor:ResetPoll();
	MessageBoard_Editor_Quote:Hide();MessageBoard_Editor_Announce:Hide(); MessageBoard_Editor_Sticky:Hide();
	MessageBoard_Editor_StickyLock:Hide(); MessageBoard_Editor_Locked:Hide(); MessageBoard_EditorTopicName:Hide();MessageBoard_EditorTopicNameNonEdit:Hide();
end

function MessageBoard.DMEVENT_BUTTON_CLICKED_id.mod(self,base,canvas,context,params)
	local user=dbPath["//MBPR/Users/"..params.user] or {};	-- grab specified user or create if absent
	if user["mod"]==true then
		user["mod"]=false
		DuckLibMB:Chat("MessageBoard:|cFFFFFFFF "..tostring(params.user).." is no longer a moderator.",0,0.9,0.9)
	else
		user["mod"]=true
		DuckLibMB:Chat("MessageBoard:|cFFFFFFFF "..tostring(params.user).." is now a moderator.",0,0.9,0.9)
	end
	dbPath["//MBPR/Users/"..params.user]=user;				-- save it and sync
	MessageBoard_ViewUserStats(params.user)
end

function MessageBoard.DMEVENT_BUTTON_CLICKED_id.ban(self,base,canvas,context,params)
	local user=dbPath["//MBPR/Users/"..params.user] or {};	-- grab specified user or create if absent
	local toon=params.user
	if MBFunc.GetBan(toon) then
		MBFunc.SetBan(toon,nil)
	else
		MBDialog:Register("MESSAGEBOARD_SET_BAN", {
		  text = L["POPUP_ENTER_BANREASON"],
		  buttons = {
			{
				text = _G.ACCEPT,
				on_click = function(self,data)
				    MBFunc.SetBan(data.toon,self.editboxes[1]:GetText())
					
				end,
			},
			{
			  text = _G.CANCEL,
			},
		  },
			editboxes = {
				{
				auto_focus = true,
				},
			},
		  on_show = function(self,data)
			self.text:SetFormattedText(L["POPUP_ENTER_BANREASON"])
			self.editboxes[1]:SetText("")
		  end,
		  hide_on_escape = true,
		  show_while_dead = true,
		})	
		MBDialog:Spawn("MESSAGEBOARD_SET_BAN", {
					toon = params.user,
				})		
	end
end

function MessageBoard.DMEVENT_BUTTON_CLICKED_id.deletetopic(self,base,canvas,context,params)
	base[tonumber(params.category)][params.topic]=nil;    -- make the topic nil
	base[tonumber(params.category)]:DM_SetRedundancy(params.topic,true);    -- it's now redundant
	MessageBoardXML:Update();
	MessageBoardXML:ViewTopics();        -- view remaining topics
end

function MessageBoard.DMEVENT_BUTTON_CLICKED_id.deleteconversation(self,base,canvas,context,params)
	if (params.sender and MessageBoard_LS.PM[params.sender]) then
		MessageBoard_LS.PM[params.sender]=nil;
	end
	MessageBoardXML:Update();
	MessageBoardXML:ViewPMs();								-- redraw
end

function MessageBoard.DMEVENT_BUTTON_CLICKED_id.movetopic(self,base,canvas,context,params)
	local menu=DMMenuCreate()
	menu.Timer=0.2    -- value in seconds
	menu:Texture("Highlight","Interface\\FriendsFrame\\UI-FriendsFrame-HighlightBar-Blue")
	menu:Header("Select A Category")
	menu:Separator()
	for i=1,9 do
		local temp=MessageBoardXML:Setting("Category"..i.."Value")
		menu:Button(temp,function () MessageBoard:MoveTopic(i,temp,params.topic,params.category) end,nil,nil,nil,nil)
	end			
	menu:Show()
end

function MessageBoard.DMEVENT_BUTTON_CLICKED_id.edit(self,base,canvas,context,params)
	MessageBoardXML.Editor:Init(	MessageBoardXML.Editor.ID_EDITMESSAGE,
									context.category,
									context.topic,
									base[context.category][context.topic][context.msg],
									dbPath["//MBPR/Base/"..context.category]:DM_Property(context.topic,1),
									context.r,context.g,context.b);
	MessageBoard_EditorTopicName:Hide();MessageBoard_Editor_Quote:Hide();MessageBoard_EditorTopicNameNonEdit:Show();MessageBoard_EditorTopicNameNonEdit:SetText(ren:toASCII(context.topic));
	MessageBoard_PollEditor:Hide();
end

function MessageBoard.DMEVENT_BUTTON_CLICKED_id.reply(self,base,canvas,context,params)
	if (context.pm=="yes") then									-- it's from a PM conversation view
		if (not context.sender) then return; end				-- no sender provided
		--local buf=unpacktable(MessageBoard_LS.PM[context.sender][tonumber(context.msg)]);	-- grab message
		--buf.Text=DuckLib.Code:Protect(buf.Text,MessageBoard_LS.PMman[category] or NATIVE);	-- make
		MessageBoardXML.Editor:Init(MessageBoardXML.Editor.ID_NEWPM,context.sender,"",{});	-- run editor
		MessageBoard_Editor_Quote:Hide();MessageBoard_Editor_Announce:Hide(); MessageBoard_Editor_Sticky:Hide(); MessageBoard_Editor_Locked:Hide(); MessageBoard_Editor_StickyLock:Hide();
		MessageBoard_EditorTopicName:Hide();MessageBoard_EditorTopicNameNonEdit:Hide();
		return;
	end
	local buf=base[context.category][context.topic][context.msg]
	local writepermission=MessageBoard:GetCategoryWritePermissions(CategoryLevel);
	if writepermission then
		buf.Text="\n\n\n|Cff888888=== "..L["EDITOR_MESSAGE_QUOTE"].." "..buf.Author.." ===\n\n"..buf.Text.."|r";
		MessageBoardXML.Editor:Init(MessageBoardXML.Editor.ID_NEWMESSAGE,CategoryLevel,context.topic,buf); --color was |cFFB8860B
		MessageBoard_Editor_Quote:Show();MessageBoard_Editor_Announce:Hide(); MessageBoard_Editor_Sticky:Hide(); MessageBoard_Editor_Locked:Hide(); MessageBoard_Editor_StickyLock:Hide();
		MessageBoard_EditorTopicName:Hide();MessageBoard_EditorTopicNameNonEdit:Show(); MessageBoard_EditorTopicNameNonEdit:SetText(ren:toASCII(context.topic));
	else DuckLibMB:Chat("MessageBoard: |cFFFFFFFF"..L["NO_REPLY_ALLOWED"],0,0.9,0.9);
	end
end

function MessageBoard.DMEVENT_BUTTON_CLICKED_id.delete(self,base,canvas,context,params)
	if (context.pm=="yes") then									-- it's from a PM conversation view
		if (not context.sender) then return; end				-- no sender provided
		local pm=MessageBoard_LS.PM[context.sender]; if (not pm) then return; end	-- no such conversation
		table.remove(pm,tonumber(context.msg));					-- delete and close gap
		if (next(pm)) then										-- not empty
			MessageBoard_Main_MessageScroller.SkipScroller=MessageBoard_Main_MessageScrollerScrollBar:GetValue();
			MessageBoard:ViewPM(context.sender);				-- redraw conversation
		else													-- no more messages
			MessageBoard_LS.PM[context.sender]=nil;				-- remove sender completely
			MessageBoardXML:ViewPMs();							-- refresh sender list
		end
		return;
	end
	base[CategoryLevel][context.topic][context.msg]=nil;
	local dbcount=0;
	listType("table");			-- we want only these
	for msg,mTable in pairs(base[CategoryLevel][context.topic]) do
		dbcount=dbcount+1;
		if (dbcount>1) then break; end
	end
	if dbcount < 1 then
		base[CategoryLevel][context.topic]=nil;     -- make the topic nil
		base[CategoryLevel]:DM_SetRedundancy(context.topic,true);    -- it's now redundant
		MessageBoardXML:Update();
		MessageBoardXML:ViewTopics();        -- view remaining topics
	else
		base[CategoryLevel][context.topic]:DM_SetRedundancy(context.msg,true);    -- it's now redundant
		MessageBoard_Main_MessageScroller.SkipScroller=MessageBoard_Main_MessageScrollerScrollBar:GetValue();
		MessageBoard:ViewThread(CategoryLevel,context.topic); 
	end
end

function MessageBoard.DMEVENT_BUTTON_CLICKED_id.copy(self,base,canvas,context,params)
	local copytext
	local t = { }
	listType("table");			-- we want only these
	for msg,v in pairs(base[context.category][context.topic]) do
		t[#t+1] = tostring(v.Author..":\n"..v.Text)
	end
	copytext=table.concat(t,"\n\n")
	MBDialog:Register("MessageBoardTextEditorCopyDialog", {
	  text = L["POPUP_TEXT_COPY"],
	  buttons = { { text = "Close", }, },
	  editboxes = { { auto_focus = true, width=200, }, },
	  on_show = function(self, data)
		self.text:SetFormattedText("\n"..L["POPUP_TEXT_COPY"].."\n")
		self.editboxes[1]:SetText(copytext)
		self.editboxes[1]:HighlightText()
	  end,
	  hide_on_escape = true,
	  show_while_dead = true,
	})
	MBDialog:Spawn("MessageBoardTextEditorCopyDialog")
end

MessageBoard.DMEVENT_DATABASE_SYNC_type={}
function MessageBoard.DMEVENT_DATABASE_SYNC(self,handle,key,newdata)
	local base=dbPath["//MBPR"];
	local unitname=UnitName("player");
	 _debug("key: "..tostring(key)) -- auk_xxxxxx_auk or State
	 _debug("handle[key]: '"..tostring(type(handle[key])).."' "..tostring(handle[key])) -- text, time, and author / or state number
	 _debug("path: "..handle.Path)
	 _debug("newdata: "..tostring(newdata)) -- newdata seems to be true or nil
	if (handle.Path:match("^//MBPR%d+/") and newdata) then				-- category sync
		MessageBoard.DMEVENT_DATABASE_SYNC_type.Category(base.Base,handle,key,newdata);
	elseif (handle.Path=="//MBPR" and key=="Settings") then				-- settings
		MessageBoard.DMEVENT_DATABASE_SYNC_type.Settings(base,handle,key,newdata);
	elseif (handle.Path=="//MBPR/Avatar") then							-- avatars
		MessageBoard.DMEVENT_DATABASE_SYNC_type.Avatar(base,handle,key,newdata);
	elseif (handle.Path=="//MBPR/Users") then							-- users
		MessageBoard.DMEVENT_DATABASE_SYNC_type.Users(base,handle,key,newdata);
	elseif (handle.Path:match("^//MBPR/PMtransfer/"..unitname.."/")) then		-- it's a PM for me
		_debug("Sync: Received a PM")
		MessageBoard.DMEVENT_DATABASE_SYNC_type.PMtransfer(base,handle,key,newdata);
	elseif type(handle[key])=="table" then 
		for k,v in pairs(handle[key]) do
			if (k=="Voted" and newdata==nil) then
				MessageBoard.DMEVENT_DATABASE_SYNC_type.Poll(base,handle,key,newdata)
			end
		end
	end
	MessageBoardXML:Update();											-- redraw
end

function MessageBoard.DMEVENT_DATABASE_SYNC_type.Poll(base,handle,key,newdata)
	local tree=DuckLib.Database.Util:Name(handle);
	_debug("---Poll Vote Received---");
	local cat = string.sub (tree[1],5,5)
	if (MessageBoard:GetCategoryReadPermissions(cat) and tree[2] and type(handle[key])=="table" and key ~= "State" and not(MBFunc.GetBan(name))) then
		local newcat = MessageBoardXML:Setting("Category"..cat.."Value")
			if newcat then 
			if MessageBoard_LS.NewPostNotify==true then -- if new post notify option is on
				DuckLibMB:Chat("MessageBoard: |cFFFFFFFF"..L["NEW_POLL_IN"]..ren:toASCII(newcat),0,0.9,0.9);   -- print notification to chat
			end
			if MessageBoard_LS.NewPostNotifyPopup==true then -- if new post notify option is on
				MBFunc.SpawnToast("MessageBoard","|cFFFFFFFF"..L["NEW_POLL_IN"]..ren:toASCII(newcat)..".|r", 1,1,1);  -- print notification to popup
			end
		end
		--MessageBoard:SetReadFlag(cat,tree[2],key,not handle[key]);
		if MBWhere=="ViewTopics" then MessageBoardXML:ViewTopics(); end
		MessageBoardXML:Update();
	end
end

function MessageBoard.DMEVENT_DATABASE_SYNC_type.Category(base,handle,key,newdata)
	local tree=DuckLib.Database.Util:Name(handle);
	_debug("---Post Received---");
	local cat = string.sub (tree[1],5,5)
	if (MessageBoard:GetCategoryReadPermissions(cat) and tree[2] and type(handle[key])=="table" and key ~= "State" and not(MBFunc.GetBan(name))) then
		local newcat = MessageBoardXML:Setting("Category"..cat.."Value")
			if newcat then 
			if MessageBoard_LS.NewPostNotify==true then -- if new post notify option is on
				DuckLibMB:Chat("MessageBoard: |cFFFFFFFF"..L["NEW_POST_IN"]..ren:toASCII(newcat),0,0.9,0.9);   -- print notification to chat
			end
			if MessageBoard_LS.NewPostNotifyPopup==true then -- if new post notify option is on
				MBFunc.SpawnToast("MessageBoard","|cFFFFFFFF"..L["NEW_POST_IN"]..ren:toASCII(newcat)..".|r", 1,1,1);  -- print notification to popup
			end
		end
		MessageBoard:SetReadFlag(cat,tree[2],key,not handle[key]);
		if MBWhere=="ViewTopics" then MessageBoardXML:ViewTopics(); end
		MessageBoardXML:Update();
	end
end

function MessageBoard.DMEVENT_DATABASE_SYNC_type.Settings(base,handle,key,newdata)
	_debug("---Settings Received---");	
	MessageBoard:UpdatePermissions()
	MessageBoardXML:UpdateCategories();
	--if (key=="MessageLimit") then
	--	base:DM_Set("Limit",handle[key]);
	--end
end

function MessageBoard.DMEVENT_DATABASE_SYNC_type.Avatar(base,handle,key,newdata)
	_debug("---Avatar List Received---");
end

function MessageBoard.DMEVENT_DATABASE_SYNC_type.Users(base,handle,key,newdata)
	_debug("---User Info Received---");
	if MBWhere=="ViewUsers" then MessageBoard_ViewUsers() end
end

function MessageBoard.DMEVENT_DATABASE_SYNC_type.PMtransfer(base,handle,key,newdata)
	_debug("---PM Received---");
	if (handle[key]) then
		local buf=handle[key]
		MessageBoard_LS.PM[buf.Author]=MessageBoard_LS.PM[buf.Author] or {};	-- create sender if first msg
		buf.how="receive";										-- we received this one
		table.insert(MessageBoard_LS.PM[buf.Author],packtable(buf,true));	-- add it scrambled: LS.PM[sender][index]=msg
		local name=UnitName("player");
		name = Ambiguate(name,"none")
		if not(MBFunc.GetBan(name)) then
			if MessageBoard_LS.NewPostNotify==true then -- if new post notify option is on
				DuckLibMB:Chat("MessageBoard: |cFFFFFFFF"..L["NEW_PM"].." "..buf.Author..".|r",0,0.9,0.9);   -- print notification to chat
			end
			if MessageBoard_LS.NewPostNotifyPopup==true then -- if new post notify option is on
				MBFunc.SpawnToast("MessageBoard","|cFFFFFFFF"..L["NEW_PM"].." "..buf.Author..".|r",1,1,1);  -- print notification to popup
			end
		end
		handle[key]=nil;										-- remove from public (trigger sync)
		MessageBoard:SetReadFlag("fakePMcategory",buf.Author,#(MessageBoard_LS.PM[buf.Author]),false);		-- set read flag to false for a message you just received
	end
	if MBWhere=="viewpms" then MessageBoardXML:ViewPMs(); end
end

function MessageBoard:UpdateRoster() -- Update guild roster
	_debug("Updating Guild Roster");
	if (not IsInGuild()) then return; end
	GuildRoster();
end

function MessageBoard.Slasher(msg) -- There's slashing to be done
	if (not msg) then msg=""; end   -- Validate input
	local omsg=msg;
	if (strlen(msg)>0) then msg=strlower(msg); end
	if (msg=="ver" or msg=="version") then
		MessageBoardXML:CheckV()
		return;
	end
	if (msg=="debug") then
		if (MessageBoard_LS.Debug==true) then MessageBoard_LS.Debug=false; DuckLib:Chat("Debug OFF");
		else MessageBoard_LS.Debug=true; DuckLib:Chat("Debug ON"); end
		return;
	end
	if (msg=="?" or msg=="help") then
		DuckLibMB:Chat("MessageBoard: |cFFFFFFFF"..L["SLASH_HELP_1"],0,0.9,0.9)
		DuckLibMB:Chat("MessageBoard: |cFFFFFFFF"..L["SLASH_HELP_2"],0,0.9,0.9)
		--DuckLibMB:Chat("MessageBoard: |cFFFFFFFF"..L["SLASH_HELP_3"],0,0.9,0.9)
		DuckLibMB:Chat("MessageBoard: |cFFFFFFFF"..L["SLASH_HELP_4"],0,0.9,0.9)
		DuckLibMB:Chat("MessageBoard: |cFFFFFFFF"..L["SLASH_HELP_5"],0,0.9,0.9)
		return;
	end
	if msg=="time" then
		print("time()",time()," offset:",DB.bgc.offset)
	end
	MessageBoard:ToggleMainFrame();
end

-- Generic mouse-wheel handler for scrolling-panes.
function MessageBoardXML:MouseWheel(slider,delta)
 if (not slider) then slider=self; end
 local min,max=slider:GetMinMaxValues(); if (not min or not max) then return; end
 local step=slider:GetValueStep(); if (not step) then return; end
 local value=slider:GetValue(); if (not value) then return; end
 if (delta==1 and (value-step)>=min) then slider:SetValue(value-step);
 elseif (delta==-1 and (value+step)<=max) then slider:SetValue(value+step); end
end

function MessageBoard_Main_OnShow() -- When the main window is shown
	if (not IsInGuild()) then MessageBoard_WaitingText:SetText(L["JOIN_A_GUILD"]); MessageBoard:DisableBoard(); return; end
	--if (not MessageBoard.HaveRoster) then MessageBoard:UpdateRoster(); end
	--MessageBoard:UpdatePermissions();
	MessageBoardXML:UpdateCategories();
	MessageBoardXML:Update();
end

function MessageBoard_Main_OnHide() -- When the main window is hidden
	MessageBoard_OFrame:Hide();
end

-- MessageBoard heartbeat
-- o The naming-convention is because the use it not as an OnUpdate
--   handler. It's a slow beat going on MB_HEARTBEAT speed, and written
--   so that it will get slowed down by hiccups like changing zone and
--   heavy loading from disk when entering a highly populated zone.
function MessageBoard_HeartBeat(self,elapsed)
	MessageBoard.Timers.Last=MessageBoard.Timers.Last+elapsed;      -- Update this
	if (MessageBoard.Timers.Last<MB_HEARTBEAT) then return; end;    -- Heartbeat is 1/3 sec
	MessageBoard.Timers.Last=0;      								-- Restart
	if (MessageBoard.Ready~=true and MessageBoard.LoggedIn) then -- and MessageBoard.HaveRoster) then    -- "LoggedIn" must guarrantee data has been loaded
		--MessageBoard:UpdatePermissions();
		MessageBoardXML:Update();
		MessageBoard.Ready=true;
		DuckLibMB:Chat("MessageBoard: |cFFFFFFFF"..L["CONNECTED_TO_THE"]..MessageBoard:GetGuildName()..L["GUILD_FORUM"],0,0.9,0.9);
		MBFunc.SpawnToast("MessageBoard","|cFFFFFFFF"..L["CONNECTED_TO_THE"]..MessageBoard:GetGuildName()..L["GUILD_FORUM"], 1,1,1);
		MessageBoard_MainTitle:SetText(MessageBoard:GetGuildName()..L["GUILD_FORUM"]);
		self:RegisterEvent("PLAYER_ALIVE", MessageBoard:ScanStats())
		self:RegisterEvent("UNIT_INVENTORY_CHANGED", MessageBoard:ScanStats())
	end
end

-- Retrieves the total number of messages by user. Don't count deleted messages
-- Author provided returns that user's message count.
function MessageBoardXML:UserMessageCount(author)
	local count=0;
	for category,cTable in pairs(dbPath["//MBPR/Base"]) do
		listType("database");			-- we want only these
		for topic,tTable in pairs(cTable) do
			listType("table");			-- we want only these
			for msg,mTable in pairs(tTable) do
				if (not author or (author and mTable.Author==author)) then
					count=count+1;
				end
			end
		end
	end
	return count;
end

function MessageBoard:CleanUp()
	_debug("Cleaning Up...",1);
	-- Kill empty topics
	for category,cTable in pairs(dbPath["//MBPR/Base"]) do
		listType("database");			-- we want only these
		for topic,tTable in pairs(cTable) do
			local found=nil;
			listType("table");			-- we want only these
			for _ in pairs(tTable) do
				found=true;
				break;
			end
			if (not found) then
				_debug("Killing: "..tostring(cTable[topic]))
				cTable[topic]=nil;
			end
		end
	end
end

-- Only runs when the main window is visible
function MessageBoard_Main_OnUpdate()
	if (DuckLib.RenderWnCounter and DuckLib.RenderWnCounter>=4) then
		if (MessageBoard_Main_MessageScroller.SkipScroller) then
            MessageBoard_Main_MessageScrollerScrollBar:SetValue(MessageBoard_Main_MessageScroller.SkipScroller);
            MessageBoard_Main_MessageScroller.SkipScroller=nil;
		end
    end
	if MessageBoard_Main_MessageScroller:IsShown() then
		if MessageBoard_Main_MessageScroller:GetVerticalScrollRange()-1 > MessageBoard_Main_MessageScroller:GetVerticalScroll() then
			MBMorePostsBelow:Show(); UIFrameFlash(MBMorePostsBelow, 1, 1, 9999, nil);
		elseif MBMorePostsBelow:GetAlpha()<0.2 then
			UIFrameFlashStop(MBMorePostsBelow); MBMorePostsBelow:Hide();
		elseif MessageBoard_Main_MessageScroller:GetVerticalScrollRange()==0 then
			MBMorePostsBelow:Hide();
		end
		--flasher:Play();flasher:SetScript("OnFinished", function() MBMorePostsBelow:SetAlpha(1) end) 
		--flasher:Stop(); --MBMorePostsBelow:Hide();
	end
end

-- Item-link support
function MessageBoardXML:OnHyperlinkShow(frame,link,text,button)
	ChatFrame_OnHyperlinkShow(frame,link,text,button);
end

-- Return the state of the provided topic
-- o Returns State,Icon,Unread
function MessageBoard:TopicState(category,topic)
	local top=dbPath["//MBPR/Base/"..category.."/"..topic];
	if (not top) then return 0,"",nil; end
	local state=dbPath["//MBPR/Base/"..category]:DM_Property(topic,1)
	--local state=top.State or 0; if (state < 0) then return 0,"",nil; end
	local newmsg=false;
	listType("table");			-- we want only these
	for msg,mTable in pairs(top) do
		if (not self:GetReadFlag(category,topic,msg)) then newmsg=true; break; end
	end
	local nm=""; if (newmsg) then nm="_new"; end
	if state== 0 then return state,Icon["NORMAL"..nm],newmsg; end
	if state==10 then return state,Icon["STICKY"..nm],newmsg; end
	if state==15 then return state,Icon["STICKYLOCKED"..nm],newmsg; end
	if state==20 then return state,Icon["ANNOUNCE"..nm],newmsg; end
	if state==30 then return state,Icon["LOCKED"..nm],newmsg; end
	if state==40 then return state,Icon["POLL"..nm],newmsg; end
	return 0,"",nil;
end

-- Some helpers to get past later updates to the WoW code
function MessageBoardXML.SetButtonText(button,text)
	_G[button:GetName().."Text"]:SetText(text);
end
function MessageBoardXML.GetButtonText(button,text)
	return _G[button:GetName().."Text"]:GetText(text);
end
function MessageBoardXML.SetButtonTextColor(button,r,g,b,a)
	_G[button:GetName().."Text"]:SetTextColor(r,g,b,a);
end

-- Update the main window
-- Use when anything visual has changed
function MessageBoardXML:Update()
	MBFunc.MBUpdateMinimap()
	MBFunc.MBGuildFrame_UpdateTabard();
	if (not dbPath["//MBPR"]) then return; end
	MessageBoard:UpdatePermissions()
	MessageBoardXML:SetCategories()
	MBFunc.GetAdmin()
	_debug("Checking for Ban")
	if MessageBoardXML.MyRank=="Banned" then MessageBoard:DisableBoard(); end
    if (not(MessageBoardXML.MyRank=="Banned") and MessageBoard_Banned:IsShown()) then MessageBoard.FirstRun=true; MessageBoard:EnableBoard(); end
end

function MessageBoardXML:GetNewMsgCount()
	local unread=nil;
	local unreadcount=0;
	listType("database");			-- we want only these
	for category,cTable in pairs(dbPath["//MBPR/Base"]) do
		for topic,tTable in pairs(cTable) do
			local readpermission=MessageBoard:GetCategoryReadPermissions(category);
			local _,_,unread=MessageBoard:TopicState(category,topic);
			if unread and readpermission then
				unreadcount=unreadcount+1
			end
		end
	end
	return unreadcount
end

function MessageBoardXML:MarkAllRead()
	local unread=nil;
	local unreadcount=0;
	listType("database");			-- we want only these
	for category,cTable in pairs(dbPath["//MBPR/Base"]) do
		for topic,tTable in pairs(cTable) do
			local readpermission=MessageBoard:GetCategoryReadPermissions(category);
			local _,_,unread=MessageBoard:TopicState(category,topic);
			if unread and readpermission then
				listType("table");			-- we want only these
				for msg,d in pairs(tTable) do
					MessageBoard:SetReadFlag(category,topic,msg,true);
				end
			end
		end
	end
	MessageBoardXML:Update()
	if MBWhere=="ViewTopics" then MessageBoardXML:ViewTopics(); end
end

-- Format TOD
function MessageBoard:FormatTime(native)
	if (not native) then return "<unknown time>"; end
	return date(MessageBoardLANG.Widget.NearDate,native);
end

function MessageBoard:GetCategoryReadPermissions(category)
	if (not category) then return false; end
	if MessageBoardXML.MyRank=="Administrator" then	return true;
	elseif MessageBoardXML.MyRank=="Moderator" then	return true;
	elseif MessageBoardXML.MyRank=="Officer" then return MessageBoardXML:Setting("Category"..category.."OfficerRead");
	elseif MessageBoardXML.MyRank=="Normal User" then return MessageBoardXML:Setting("Category"..category.."UserRead");
	elseif MessageBoardXML.MyRank=="Alt" then return MessageBoardXML:Setting("Category"..category.."UserRead");
	elseif MessageBoardXML.MyRank=="No Access" then	return false;
	end
end

function MessageBoard:GetCategoryWritePermissions(category)
	if (not category) then return false; end
	if MessageBoardXML.MyRank=="Administrator" then	return true;
	elseif MessageBoardXML.MyRank=="Moderator" then	return true;
	elseif MessageBoardXML.MyRank=="Officer" then return MessageBoardXML:Setting("Category"..category.."OfficerWrite");
	elseif MessageBoardXML.MyRank=="Normal User" then return MessageBoardXML:Setting("Category"..category.."UserWrite");
	elseif MessageBoardXML.MyRank=="Alt" then return false;
	elseif MessageBoardXML.MyRank=="No Access" then	return false;
	end
end

function MessageBoardXML:MakeNewTopic()
	if (CategoryLevel==nil) then return; end
	if (MessageBoard:GetCategoryWritePermissions(CategoryLevel)) then -- check ranks
		MessageBoardXML.Editor:Init(MessageBoardXML.Editor.ID_NEWTOPIC,CategoryLevel,"",{});
		MessageBoard_Editor_Quote:Hide();MessageBoard_EditorTopicNameNonEdit:Hide();
		MessageBoardXML.Editor:ResetPoll();
	else DuckLibMB:Chat("MessageBoard: |cFFFFFFFF"..L["NOT_ALLOWED_POST"],0,0.9,0.9);
	end
end

function MessageBoardXML:MakeNewArt()
	if (CategoryLevel==nil) then return; end
	if (MessageBoard:GetCategoryWritePermissions(CategoryLevel)) then -- check ranks
		MessageBoardXML.Editor:Init(MessageBoardXML.Editor.ID_NEWART,CategoryLevel,"",{});
		MessageBoard_Editor_Quote:Hide();MessageBoard_EditorTopicNameNonEdit:Hide();
		MessageBoardXML.Editor:ResetPoll();
	else DuckLibMB:Chat("MessageBoard: |cFFFFFFFF"..L["NOT_ALLOWED_POST"],0,0.9,0.9);
	end
end

function MessageBoardXML:MakeNewPoll()
	if (CategoryLevel==nil) then return; end
	if (MessageBoard:GetCategoryWritePermissions(CategoryLevel)) then -- check ranks
		MessageBoardXML.Editor:Init(MessageBoardXML.Editor.ID_NEWPOLL,CategoryLevel,"",{});
		MessageBoard_Editor_Quote:Hide();MessageBoard_EditorTopicNameNonEdit:Hide();MessageBoard_Editor_Announce:Hide();MessageBoard_Editor_Sticky:Hide();MessageBoard_Editor_Locked:Hide();MessageBoard_Editor_StickyLock:Hide();
		MessageBoardXML.Editor:ResetPoll();
	else DuckLibMB:Chat("MessageBoard: |cFFFFFFFF"..L["NOT_ALLOWED_POST"],0,0.9,0.9);
	end
end

function MessageBoardXML:Reply() -- new reply with no quotes
	if (not MessageBoard.Selected.Topic) then return; end
	local state=dbPath["//MBPR/Base/"..CategoryLevel]:DM_Property(MessageBoard.Selected.Topic,1)
	if (state==30 or state==15) then -- if locked
		DuckLibMB:Chat("MessageBoard: |cFFFFFFFF"..L["TOPIC_IS_LOCKED"],0,0.9,0.9);
		return;
	end
	if (MessageBoard:GetCategoryWritePermissions(CategoryLevel)) then -- check ranks
		MessageBoardXML.Editor:Init(MessageBoardXML.Editor.ID_NEWMESSAGE,CategoryLevel,MessageBoard.Selected.Topic,{});
		MessageBoard_Editor_Quote:Hide();MessageBoard_Editor_Announce:Hide();MessageBoard_Editor_Sticky:Hide();
		MessageBoard_Editor_Locked:Hide();MessageBoard_EditorTopicName:Hide();MessageBoard_EditorTopicNameNonEdit:Show();
		MessageBoard_Editor_StickyLock:Hide();
		MessageBoard_EditorTopicNameNonEdit:SetText(ren:toASCII(MessageBoard.Selected.Topic));
	else DuckLibMB:Chat("MessageBoard: |cFFFFFFFF"..L["NO_REPLY_ALLOWED"],0,0.9,0.9);
	end
end

-- Entry-function when editor is done (no validateboard)
-- o The text will be saved accordingly
function MessageBoardXML:EditedText(id,category,topic,data,state,colorr,colorg,colorb)
	local now=time();
	local cat=dbPath["//MBPR/Base/"..category];
	local top;
	if (not cat) then
		if (id~=MessageBoardXML.Editor.ID_NEWPM) then _debug("Category \""..category.."\" was not found",1); return; end
	else
		top=cat[topic];
	end
	if (id==MessageBoardXML.Editor.ID_NEWTOPIC or id==MessageBoardXML.Editor.ID_NEWPOLL) then
		if (top) then
			DuckLibMB:Chat("MessageBoard: |cFFFFFFFF"..L["TOPIC_IS_EXIST"],0,0.9,0.9);
			return;   -- can't allow same-named topics in one category
		end
		cat[topic]={ }; -- old state cat[topic]={ State=state };
		dbPath["//MBPR/Base/"..category]:DM_Property(topic,1,state) -- new state
		top=cat[topic];
		top:DM_EndPoint(true);
	end
	if (id==MessageBoardXML.Editor.ID_NEWTOPIC or id==MessageBoardXML.Editor.ID_NEWMESSAGE or id==MessageBoardXML.Editor.ID_NEWPOLL) then
		_debug("Called EditedText NEWTOPIC NEWMESSAGE OR NEWPOLL")
		if (not top) then _debug("Topic \""..topic.."\" was not found",1); return; end
		local msg=top:DM_UniqueKey();
		local name=UnitName("player");
		name = Ambiguate(name,"none")
		data.Time=now;
		data.Author=name;
		if MessageBoardXML:Setting("AllowColorPicker") then
			data.colorR=colorr;
			data.colorG=colorg;
			data.colorB=colorb;
		else
			data.colorR=1;
			data.colorG=1;
			data.colorB=1;
		end
		data.Edited=nil;
		top[msg] = data;
		MessageBoard:SetReadFlag(category,topic,msg,true); -- set read flag to true for a message you just wrote
	elseif (id==MessageBoardXML.Editor.ID_EDITMESSAGE) then
		_debug("Called EditedText EDITMESSAGE")
		local name=UnitName("player");
		name = Ambiguate(name,"none")
		 if (not top) then _debug("Topic \""..topic.."\" was not found",1); return; end
		-- top.State=state; -- old state
		dbPath["//MBPR/Base/"..category]:DM_Property(topic,1,state) -- new state
		local tempmsg,msg;
		listType("table");			-- we want only these
		for msg,unpmsg in pairs_dn(top) do
			if (unpmsg.Text~=data.Text) or (data.colorR~=colorr or data.colorG~=colorg or data.colorB~=colorb) then
				data.Edited="\n\n|Cff888888"..L["TOPIC_SCROLLER_EDITED"].." "..name.." on "..MessageBoard:FormatTime(now);
				data.colorR=colorr;
				data.colorG=colorg;
				data.colorB=colorb;
			end
            if (unpmsg.Time==data.Time) then tempmsg=msg; break; end
		end
		if (not tempmsg) then _debug("Original message not found",1); return; end
		top[tempmsg]=data;
		MessageBoard:SetReadFlag(category,topic,tempmsg,true); -- set read flag to true for a message you just edited
	elseif (id==MessageBoardXML.Editor.ID_NEWPM) then
		_debug("Called EditedText NEWPM")
		local pm=dbPath["//MBPR/PMtransfer"];
		local name=UnitName("player");
		if (not pm[category]) then pm[category]={ }; end					-- "category" is receiver - create
		if (not pm[category][name]) then pm[category][name]={}; end			-- create sender
		pm[category][name]:DM_EndPoint(true);								-- The line must be draw here! No further!
		local msg=pm[category][name]:DM_UniqueKey();
		data.Time=now; data.Author=name;									-- set standards
		if MessageBoardXML:Setting("AllowColorPicker") then
			data.colorR=colorr; data.colorG=colorg; data.colorB=colorb;		-- set standards
		end
		pm[category][name][msg]=data;										-- sync
		MessageBoard_LS.PM[category]=MessageBoard_LS.PM[category] or {};	-- create recipient if first msg
		table.insert(MessageBoard_LS.PM[category],packtable(data,true));	-- shove packed coded
		MessageBoard:SetReadFlag("fakePMcategory",category,#(MessageBoard_LS.PM[category]),true);	-- set read flag to true for a pm you just wrote
		DuckLibMB:Chat("MessageBoard: |cFFFFFFFF"..L["PM_SENT"],0,0.9,0.9);
	end
	MessageBoardXML:Update();
	if (MBWhere=="viewpms") then MessageBoard:ViewPMs() end
	if (MBWhere=="viewpm") then MessageBoard:ViewPM(category); end -- MessageBoard_Main_MessageScroller.SkipScroller=MessageBoard_Main_MessageScrollerScrollBar:GetValue(); end
	if (MBWhere=="ViewTopics" and CategoryLevel==category) then MessageBoardXML:ViewTopics(); end
	if (MBWhere=="ViewThread" and MessageBoard.Selected.Topic==topic) then MessageBoardXML:ViewTopics(); end --MessageBoard:ViewThread(CategoryLevel,MessageBoard.Selected.Topic); end
end

function MessageBoardXML:Button_OnEnter()
	if (not self.Help) then return; end
	GameTooltip:SetOwner(self,"ANCHOR_RIGHT");
	GameTooltip:SetText(MessageBoardLANG.Help.Usage);
	GameTooltip:AddLine(self.Help,1,1,1,true);
	GameTooltip:Show();
end

function MessageBoard:GetGuildName()
	MessageBoard.GuildName=GetGuildInfo("player");
	if (not MessageBoard.GuildName) then MessageBoard.GuildName="<UnGuilded>"; end
	return MessageBoard.GuildName;
end

-- Check and create board-base only
function MessageBoard:ValidateBoard(namesonly)
	if (not MessageBoard_DB) then return nil; end						-- can't work without that
	local gn=MessageBoard:GetGuildName(); if (not gn or gn=="<UnGuilded>") then return nil; end	-- no guild, no share
	local dbn=gn;														-- NOTE: swapped realm and guild to not kill old stuff
	if (namesonly) then return dbn; end									-- don't remember what this was for, but I kept it in there
	if (MessageBoard_DB[dbn] and MessageBoard_DB[dbn].DM_Registered) then return MessageBoard_DB[dbn]; end
	if (not next(MessageBoard_DB[dbn] or {})) then						-- database does not exists or is empty
		MessageBoard_DB[dbn]=todatabase({ Avatar={}, Base={}, });		-- so create it
		for i=1,9 do													-- do 9
			MessageBoard_DB[dbn..i]={};									-- create this category
			DB:Register("MBPR"..i,MessageBoard_DB[dbn..i],MessageBoard_Main);	-- register it as a database
		end
		DB:Register("MBPR",MessageBoard_DB[dbn],MessageBoard_Main);		-- register root database
		for i=1,9 do MessageBoard_DB[dbn].Base:DM_SetLink(i,MessageBoard_DB[dbn..i]); end	-- link all 9 categories
	else
		for i=1,9 do DB:Register("MBPR"..i,MessageBoard_DB[dbn..i],MessageBoard_Main); end	-- register categories' databases
		DB:Register("MBPR",MessageBoard_DB[dbn],MessageBoard_Main);		-- register root database
	end
	for i=1,9 do MessageBoard_DB[dbn..i]:DM_Set("Native",true); end		-- native categories
	MessageBoard_DB[dbn]:DM_Set("Native",false);						-- native main
	if (not MessageBoard_DB[dbn].Users) then MessageBoard_DB[dbn].Users={}; end
	MessageBoard_DB[dbn].Users:DM_EndPoint(true);
	if (not MessageBoard_DB[dbn].PMtransfer) then MessageBoard_DB[dbn].PMtransfer={}; end	-- create PM db
	MessageBoard_LS.PM=MessageBoard_LS.PM or {};						-- create PM container
	MessageBoard_LS.PMman=MessageBoard_LS.PMman or {};					-- create PM man-pages
	-- backward compatibility adaption
	for _,c in pairs(MessageBoard_DB[dbn].Base) do			-- walk through all categories
		for _,t in pairs(c) do								-- walk through all topics
			t:DM_EndPoint(true);							-- set endpoint
		end
	end
	if (type(MessageBoard_DB[dbn].Settings)=="table") then
        MessageBoard_DB[dbn].Settings=packtable(MessageBoard_DB[dbn].Settings);
    end

	-- Init
	MBFunc.InitializeMinimap()
	MBFunc.GetAdmin()
	MessageBoard:SetPrunes()
	
	return MessageBoard_DB[dbn];							-- upsies
end

function MessageBoard:ToggleMainFrame()
	if (MessageBoard_Main:IsVisible()) then MessageBoard_Main:Hide();
	else MessageBoard_Main:Show(); end
	DB.EventFrame=MessageBoard_Main;
end

-- MessageBoard Debug
function MessageBoardXML:Debug(...)
	_debug(...);
end

local LibStub = _G.LibStub;
local Toast = LibStub("LibToast-1.0");

-- Update User Permissions
function MessageBoard:UpdatePermissions()
	_debug("Updating Permissions");
	local name=UnitName("player");
	name = Ambiguate(name,"none")
	local guildName, guildRankName, guildRankIndex = GetGuildInfo("player");
	local myrank = guildRankIndex + 1;
	if MessageBoard.Administrator then MessageBoardXML.MyRank="Administrator";
	elseif MBFunc.GetBan(name) then MessageBoardXML.MyRank="Banned";
	elseif MessageBoardXML:GetMod() then MessageBoardXML.MyRank="Moderator";
	elseif (MessageBoardXML:Setting("OfficerRank"..myrank) == true or IsGuildLeader()) then	MessageBoardXML.MyRank="Officer";
	elseif MessageBoardXML:Setting("AltRank"..myrank) == true then MessageBoardXML.MyRank="Alt";
	elseif MessageBoardXML:Setting("NoPostRank"..myrank) == true then MessageBoardXML.MyRank="No Access";
	elseif MessageBoardXML:Setting("NormalRank"..myrank) == true then MessageBoardXML.MyRank="Normal User";
	end
end

function MessageBoardXML:UpdateCategories()
	_debug("Updating Categories");
	local i
	local count=0
	for i=1,9 do
		if (MessageBoardXML:Setting("Category"..i.."Value")==nil or MBFunc.isBlank(MessageBoardXML:Setting("Category"..i.."Value"))) then
			_G["MessageBoard_MainNavBarButton"..i]:Hide();
			count=count+1
		else
			_G["MessageBoard_MainNavBarButton"..i]:Show();
		end
	end
	if count==9 then MessageBoard:DisableBoard(); return end
	if MessageBoard.Administrator then
		MessageBoardXML:SetCategories();
		--if MessageBoard_Waiting:IsShown() then
			MessageBoard:EnableBoard();
		--end
	else
		MessageBoardXML:SetCategories();
		--if MessageBoard_Waiting:IsShown() then
			MessageBoard:EnableBoard();
		--end
	end
	--if MessageBoardXML.MyRank=="Banned" then MessageBoard:DisableBoard(); end
end

function MessageBoard:DisableBoard() -- run when settings are not yet received/setup
	_debug("No settings, not Administrator, or not in a guild, disabling messageboard");
	MessageBoard_Banned:Hide();
	MessageBoard_MainNavBar:Hide();
	MessageBoard_CurrentTopic:Hide();
	MessageBoard_Main_MessageScroller:Hide();
	MessageBoard_Main_MessageScroller_Browser_Frame:Hide();
	MBMorePostsBelow:Hide();
	MessageBoard_Main_LeftPanel:Hide();
	MessageBoard_Main_RightPanel:Hide();
	MessageBoard_Main_LeftDisplay:Hide();
	MessageBoard_Main_RightDisplay:Hide();
	if MessageBoard.Administrator then
		MessageBoard_Waiting:Show();
		MessageBoard_OpenSettings:Show();
	else
		MessageBoard_OpenSettings:Hide();
		MessageBoard_Waiting:Show();
	end
	if MessageBoardXML.MyRank=="Banned" then
		MessageBoard_Waiting:Hide();
		MessageBoard_Banned:Show();
		local name=UnitName("player");
		name = Ambiguate(name,"none")
		MessageBoard_BannedText:SetText(L["SORRY_BANNED"].."\n\n"..L["BANNED_REASON"]..": "..tostring(MBFunc.GetBan(name)));		
	end	
end

function MessageBoard:EnableBoard()
	_debug("EnableBoard")
	if MessageBoard.FirstRun then
		_debug("-- First Run --")
		MessageBoard_OpenSettings:Show();
		MessageBoard_Waiting:Hide();
		MessageBoard_Banned:Hide();
		MessageBoard_MainNavBar:Show();
		MessageBoard_Main_MessageScroller:Hide();
		MessageBoard_Main_MessageScroller_Browser_Frame:Hide();
		MessageBoard_CurrentTopic:Hide();
		MessageBoard_Main_LeftPanel:Show();
		MessageBoard_Main_RightPanel:Show();
		MessageBoard_Main_LeftDisplay:Show();
		MessageBoard_Main_RightDisplay:Show();
		-- set initial category based on first accessible
		local i=1;
		while i<10 do
			local readpermission=MessageBoard:GetCategoryReadPermissions(i);
			if readpermission then
				MessageBoardXML:SetCategoryLevel(i);
				_G["MessageBoard_MainNavBarButton"..i]:Click(); break;
			end
			i=i+1;
		end
		MessageBoard_Back:SetText(MessageBoardXML:Setting("Category"..i.."Value") or "");
		MessageBoard.FirstRun=false;
	end
	if MessageBoard.Selected.Topic then MessageBoard:ViewThread(CategoryLevel,MessageBoard.Selected.Topic); end
	if MessageBoard_Waiting:IsShown() then MessageBoard_Waiting:Hide() end
end

function MessageBoard_Search()
	MBWhere="Search"
	for i = 1,9 do 	_G["MessageBoard_MainNavBarButton"..i].SelectedBG:Hide() end
	MessageBoard_MainNewTopic:Disable()
	MessageBoard_MainNewPoll:Disable()
	MessageBoard_MainMsgReply:Disable()
	CategoryLevel=nil
	MessageBoard_Back:SetScript('OnClick', function() MessageBoard_Search(); end)
	MessageBoard_Back:SetText("Search");
	MessageBoard_Back:SetSize(80, 30);
	MessageBoard_Main_MessageScroller_Browser_Frame:Hide();MessageBoard_CurrentTopic:Hide();
	local text=MessageBoard_MainNavBarSearchBox:GetText() or "";
	text=text:lower();
	local count=0;
	local page=[[<wn><meta>
		<page background="#00000000"/>
		<font color="#FFFFFFFF" size="14"/>
		<elements>
			<button inheritSys="MB_Topic_Buttons2"/>
		</elements>
	</meta><body><img height=5></img>]];
	local avatar=dbPath["//MBPR/Avatar"];
	for category,cTable in pairs(dbPath["//MBPR/Base"]) do -- cycle through the categories
		listType("database");			-- we want only these
		for topic,tTable in pairs(cTable) do -- cycle through the topics
		local state=dbPath["//MBPR/Base/"..category]:DM_Property(topic,1)
			listType("table");			-- we want only these
			for k,buf in pairs(tTable) do
				local readpermission=MessageBoard:GetCategoryReadPermissions(category);
				if (type(buf.Text)=="string" and type(buf.Author)=="string" and buf.Text and readpermission and state) then -- dont search deleted messages --
					local addit=nil;
					if (buf.Text:lower():find(text,1,true)) then     -- Check article text
						_debug("Found article in: "..topic,0,1,1); -- name of the topic with the text debug
						addit=true;
					elseif (buf.Author:lower():find(text,1,true)) then      -- Check article author
						_debug("Found author: "..buf.Author,0,1,1); -- name of the topic with the text debug
						addit=true;
					end
					if (addit) then
						local pic=nil;
						pic=avatar[buf.Author] or "Interface\\Icons\\INV_Misc_QuestionMark";
						page=page..[[<img align="left" pad="0" height="20" width="20">]]..pic..[[</img><button height="30" key="]]..k..[[" term="]]..text..[[" tip=excerpt author="]].. buf.Author..[[" category="]]..category..[[" topic="]]..topic..[[" id="topic" pad="0" width="95%" glue="left">]]..topic..[[</button>]];
						count=count+1;
					end
				end
			end
		end
	end
	page=page..[[</body></wn>]];
	MessageBoard_Main_MessageScroller:Show();MessageBoard_Main_MessageScroller_Browser_Frame:Show();
    DuckLib.Render:Render(MessageBoard_Main_MessageScroller_Browser_Frame,page);
end

function MessageBoardXML:GetMod()
	local name=UnitName("player"); name = Ambiguate(name,"none")
	local stats=dbPath["//MBPR/Users/"..name]; if (not stats) then return; end
	return stats.mod;
end

	
function MessageBoardXML:ViewTopics()
	MessageBoardXML:Update()
	mt:Run("ViewTopics",self.ViewTopicsMT,self);
end

local pageT={};
function MessageBoardXML:ViewTopicsMT()
	--local debugstarted=debugprofilestop();
	_debug("ViewTopics()")
	wipe(pageT);
	MessageBoard_Back:SetScript('OnClick', function() MessageBoardXML:ViewTopics(); end)
	MessageBoard_MainMsgReply:Disable()
	if (not MessageBoard:GetCategoryReadPermissions(CategoryLevel)) then return; end
    CategoryLevel=CategoryLevel or 1;    -- set to 1 if it is nil
    MessageBoard_Main_MessageScroller:Hide();MessageBoard_Main_MessageScroller_Browser_Frame:Hide();MessageBoard_CurrentTopic:Hide();
	MBWhere="ViewTopics"
	table.insert(pageT,[[<wn><meta><page background=#0000/><font color=#FFFF size=14/><elements><button inheritSys=MB_Topic_Buttons2/></elements></meta><body><img height=5></img>]]);
	local _state,_unread=20,true;										-- first selection in the generated list
	local tlist={};        -- NEWNOW: list for sorting
	local cacheGetMod=MessageBoardXML:GetMod()
	local topicStates={};
	for topic,tTable in pairs(dbPath["//MBPR/Base/"..CategoryLevel]) do
		topicStates[topic]={};
		topicStates[topic].state,topicStates[topic].icon,topicStates[topic].unread=MessageBoard:TopicState(CategoryLevel,topic);
		topicStates[topic].tim=MessageBoard:GetTopicTime(tTable);
	end
	repeat
	    wipe(tlist);        -- NEWNOW: clear!
		listType("database");			-- we want only these
		for topic in pairs(dbPath["//MBPR/Base/"..CategoryLevel]) do -- was pairs_up
			if topicStates[topic].state==_state and topicStates[topic].unread==_unread then
				table.insert(tlist,{icon=topicStates[topic].icon,topic=topic,tim=topicStates[topic].tim});
			end
		end
		table.sort(tlist,function(a,b) return (a.tim or 0)>(b.tim or 0); end); -- NEWNOW: sort with biggest on top
		-- spew xml data
		for k,v in pairs(tlist) do
			table.insert(pageT,[[<img align="left" pad="0" height="20" width="20">]]);
			table.insert(pageT,v.icon);
			table.insert(pageT,[[</img><button height="30" category="]]);
			table.insert(pageT,tostring(CategoryLevel));
			table.insert(pageT,[[" topic="]]);
			table.insert(pageT,v.topic);
			table.insert(pageT,[[" id="topic" pad="0" width="80%" glue="left">]]);
			table.insert(pageT,v.topic);
			table.insert(pageT,[[</button>]]);
			if MessageBoard.Administrator or cacheGetMod then
				table.insert(pageT,[[<button inheritSys="MessageBoard_PostButtonsTemplate" category="]]);
				table.insert(pageT,tostring(CategoryLevel));
				table.insert(pageT,[[" topic="]]);
				table.insert(pageT,v.topic);
				table.insert(pageT,[[" id="movetopic" pad="10" width="50" glue="left">]]);
				table.insert(pageT,L["TOPIC_SCROLLER_MOVE"]);
				table.insert(pageT,[[</button>]]);
				table.insert(pageT,[[<button inheritSys="MessageBoard_PostButtonsTemplate" category="]]);
				table.insert(pageT,tostring(CategoryLevel));
				table.insert(pageT,[[" topic="]]);
				table.insert(pageT,v.topic);
				table.insert(pageT,[[" id="deletetopic" pad="10" width="50" glue="left">]]);
				table.insert(pageT,L["TOPIC_SCROLLER_DELETE"]);
				table.insert(pageT,[[</button>]]);
			end
		end
		if (not _unread) then 
			if _state==20 then _state=15;					-- third selection in list
			elseif _state==15 then _state=10;				-- fifth selection in list
			elseif _state==10 then _state=40;				-- seventh selection in list
			elseif _state==40 then _state=0;				-- ninth selection in list
			elseif _state==0 then _state=30;				-- eleventh selection in list
			elseif _state==30 then _state=nil;              -- thirteeth selection in list
			end
		end
		_unread=not _unread									-- up odd number selections to even
		mt:Yield();
	until(not _state);
    table.insert(pageT,[[</body></wn>]]);
    MessageBoard_Main_MessageScroller:Show(); MessageBoard_Main_MessageScroller_Browser_Frame:Show();
    DuckLib.Render:Render(MessageBoard_Main_MessageScroller_Browser_Frame,table.concat(pageT));
	--print(string.format("%.1f",debugprofilestop()-debugstarted).." ms");
end

function MessageBoardXML:ViewTopicsMT_old()
	--local debugstarted=debugprofilestop();
	_debug("ViewTopics()")
	MessageBoard_Back:SetScript('OnClick', function() MessageBoardXML:ViewTopics(); end)
	MessageBoard_MainMsgReply:Disable()
	if (not MessageBoard:GetCategoryReadPermissions(CategoryLevel)) then return; end
    CategoryLevel=CategoryLevel or 1;    -- set to 1 if it is nil
    MessageBoard_Main_MessageScroller:Hide();MessageBoard_Main_MessageScroller_Browser_Frame:Hide();MessageBoard_CurrentTopic:Hide();
	MBWhere="ViewTopics"
	local page=[[<wn><meta><page background=#0000/><font color=#FFFF size=14/><elements><button inheritSys=MB_Topic_Buttons2/></elements></meta><body><img height=5></img>]];
	local _state,_unread=20,true;										-- first selection in the generated list
	local tlist={};        -- NEWNOW: list for sorting
	local cacheGetMod=MessageBoardXML:GetMod()
	local topicStates={};
	for topic,tTable in pairs(dbPath["//MBPR/Base/"..CategoryLevel]) do
		topicStates[topic]={};
		topicStates[topic].state,topicStates[topic].icon,topicStates[topic].unread=MessageBoard:TopicState(CategoryLevel,topic);
		topicStates[topic].tim=MessageBoard:GetTopicTime(tTable);
	end
	repeat
	    wipe(tlist);        -- NEWNOW: clear!
		listType("database");			-- we want only these
		for topic in pairs_up(dbPath["//MBPR/Base/"..CategoryLevel]) do
			if topicStates[topic].state==_state and topicStates[topic].unread==_unread then
				table.insert(tlist,{icon=topicStates[topic].icon,topic=topic,tim=topicStates[topic].tim});
			end
		end
		table.sort(tlist,function(a,b) return (a.tim or 0)>(b.tim or 0); end); -- NEWNOW: sort with biggest on top
		-- spew xml data
		for k,v in pairs(tlist) do
			page=page..[[<img align="left" pad="0" height="20" width="20">]]..v.icon..[[</img><button height="30" category="]]..CategoryLevel..[[" topic="]]..v.topic..[[" id="topic" pad="0" width="80%" glue="left">]]..v.topic..[[</button>]];
			if MessageBoard.Administrator or cacheGetMod then
				page=page..[[<button inheritSys="MessageBoard_PostButtonsTemplate" category="]]..CategoryLevel..[[" topic="]]..v.topic..[[" id="movetopic" pad="10" width="50" glue="left">]]..L["TOPIC_SCROLLER_MOVE"]..[[</button>]];
				page=page..[[<button inheritSys="MessageBoard_PostButtonsTemplate" category="]]..CategoryLevel..[[" topic="]]..v.topic..[[" id="deletetopic" pad="10" width="50" glue="left">]]..L["TOPIC_SCROLLER_DELETE"]..[[</button>]];
			end
		end
		if (not _unread) then 
			if _state==20 then _state=15;					-- third selection in list
			elseif _state==15 then _state=10;				-- fifth selection in list
			elseif _state==10 then _state=40;				-- seventh selection in list
			elseif _state==40 then _state=0;				-- ninth selection in list
			elseif _state==0 then _state=30;				-- eleventh selection in list
			elseif _state==30 then _state=nil;              -- thirteeth selection in list
			end
		end
		_unread=not _unread									-- up odd number selections to even
		mt:Yield();
	until(not _state);
    page=page..[[</body></wn>]];
    MessageBoard_Main_MessageScroller:Show(); MessageBoard_Main_MessageScroller_Browser_Frame:Show();
    DuckLib.Render:Render(MessageBoard_Main_MessageScroller_Browser_Frame,page);
	--print(string.format("%.1f",debugprofilestop()-debugstarted).." ms");
end

function MessageBoard:ViewThread(category,topic)
	mt:Run("ViewThread",self.ViewThreadMT,self,category,topic);
end

function MessageBoard:ViewThreadMT(category,topic)
	CategoryLevel=category
	_debug("ViewThread()")
	MessageBoard_Back:SetScript('OnClick', function() MessageBoardXML:ViewTopics();	end)
	MessageBoard_Back:SetText(MessageBoardXML:Setting("Category"..category.."Value") or "");
	local length = strlen(MessageBoard_Back:GetText())
	if length > 12 then MessageBoard_Back:SetSize(130, 30) end
	--MessageBoardXML:SetCategoryLevel(category)
	MessageBoard_MainMsgReply:Enable()
	MessageBoard_MainNewTopic:Enable()
	MessageBoard_MainNewPoll:Enable()
	for i = 1,9 do
		_G["MessageBoard_MainNavBarButton"..i].SelectedBG:Hide()
	end
	_G["MessageBoard_MainNavBarButton"..category].SelectedBG:Show()
	MessageBoard.Selected.Topic = topic
	MBMorePostsBelow:Hide();
	MBWhere="ViewThread"
	local top=dbPath["//MBPR/Base/"..category.."/"..topic];
	local otopic=topic;
	topic=topic:gsub("<"," "); topic=topic:gsub(">"," "); topic=topic:gsub('"'," ");
	if (otopic~=topic) then
		top=dbPath["//MBPR/Base/"..category];		-- grab category
		top[topic]=top[otopic];						-- copy to new
		top[otopic]=nil;							-- delete old
		top=top[topic];								-- change variable
	end
	local page=[[<wn><meta><page background=#0000/><font size=]]..MessageBoard_LS.FontSize..[[/><elements><button inheritSys=MessageBoard_PostButtonsTemplate/></elements></meta><body>]];
	page=page..[[<h2 glue=left pad=10>|cFFC0C0C0]]..topic..[[|r</h2>]];
	local icon=Icon.NORMAL;
	if (not top) then _debug("Topic has been deleted."); MessageBoardXML:ViewTopics(); return; end -- topic has been deleted
	local state=dbPath["//MBPR/Base/"..category]:DM_Property(topic,1)
	if state==10 then icon=Icon.STICKY;
	elseif state==15 then icon=Icon.STICKYLOCKED;
	elseif state==20 then icon=Icon.ANNOUNCE;
	elseif state==30 then icon=Icon.LOCKED;
	elseif state==40 then icon=Icon.POLL; end	
	page=page..[[<img glue=left pad=10 width=25 height=25>]]..icon..[[</img>]];
	page=page..[[<img align=center pad=10 height=12 width=100%>Interface\FriendsFrame\UI-FriendsFrame-OnlineDivider</img>]];
	local msglist={};			-- local unpacked storage
	listType("table");			-- we want only these
	for msg,d in pairs(top) do			-- get all message keys and data
		table.insert(msglist,{msg=msg,Time=d.Time});		-- insert key and time
	end
	table.sort(msglist,function(A,B)			-- let lua sort it
		if (type(A)~="table" or type(B)~="table") then return nil; end	-- only sort tables (messages)
		return (A.Time<B.Time);				-- the sorting parameters
	end );
	local avatar=dbPath["//MBPR/Avatar"];
	local msgcount={};
	local writepermission=MessageBoard:GetCategoryWritePermissions(CategoryLevel); -- check permission --
	local cacheGetMod=MessageBoardXML:GetMod();
	for _,mit in ipairs(msglist) do			-- read all sorted message indexes
		local msg,buf=mit.msg,top[mit.msg];
		local msgTxt=buf.Text
		if (msgTxt:match("(%a+://%S+)")) then -- grab the http links
			msgTxt=msgTxt:gsub("(%a+://%S+)","|cFF4682B4|Hhttp:%1|h[%1]|h|r"); end
		if (msgTxt:match("%f[%S](www%.[-%w_%%]+%.%S+)")) then -- grab the http links
			msgTxt=msgTxt:gsub("%f[%S](www%.[-%w_%%]+%.%S+)","|cFF4682B4|Hhttp:%1|h[%1]|h|r"); end
		page=page..[[<focus id="]]..topic..msg..[["/>]];
		if buf.colorR then
			page=page..[[<set id=topic r=]]..buf.colorR..[[ g=]]..buf.colorG..[[ b=]]..buf.colorB..[[ author="]]..buf.Author..[[" category="]]..category..[[" topic="]]..topic..[[" msg="]]..msg..[["/>]];
		else
			page=page..[[<set author="]]..buf.Author..[[" category="]]..category..[[" topic="]]..topic..[[" msg="]]..msg..[["/>]];
		end
		page=page.."<img click=left width=60 height=60 pad=10 id=user tip=author user='"..buf.Author.."' author='"..buf.Author.."'>";
		page=page..(avatar[buf.Author] or [[Interface\Icons\INV_Misc_QuestionMark]]);
		page=page..[[</img><text size=12 pad=10 glue=left>|cFFC0C0C0by: ]];
		page=page..buf.Author.." - "..MessageBoard:FormatTime(buf.Time).."|r</text>";
		if (not msgcount[buf.Author]) then msgcount[buf.Author]=MessageBoardXML:UserMessageCount(buf.Author); end
		page=page..[[<text size=8>|cFFC0C0C0Posts: ]]..msgcount[buf.Author];
		page=page..[[|r</text>]];
		-- check for polls and display --
		if buf.Poll then
			-- check if previously answered or have permission to write
			local name=UnitName("player");
			name = Ambiguate(name,"none")
			if (not buf.Voted or not buf.Voted[name] and writepermission) then
				-- set context variables that indicates widgets that are members of the poll
				page=page..[[<set pollid="]]..tostring(category)..topic..msg..[[" pollc="]]..category..[[" pollt="]]..topic..[[" pollm="]]..msg..[["/>]];
				page=page..[[<h2 align=center pad=10>|cFFC0C0C0]]..buf.Poll.Question..[[|r</h2>]];
				for polln,polla in ipairs(buf.Poll) do	-- do numeric indices (skipping the question) -- was ipairs
					if ((polla or "")~="") then
						page=page..[[<set group="]]..buf.Poll.Question..[["/><radio align=center id="]]..polln..[[" pad=5 width=150 height=50>|cFFC0C0C0]]..polla..[[|r</radio>]];
					end
				end
				page=page..[[<img height=10></img>]];
				page=page..[[<button align=center action=submit>]]..L["TOPIC_SCROLLER_SUBMIT"]..[[</button>]];	-- the "Submit" button
				page=page..[[<set pollid="" pollc="" pollt="" pollm=""/>]];	-- clear poll context
			else -- if answered then show results
				local votecount=0;
				page=page..[[<h2 align="center" pad="10">|cFFC0C0C0]]..buf.Poll.Question..[[|r</h2>]];
				local bar={};
				for i,o in ipairs(buf.Poll) do -- find index of last option -- was ipairs
					if (o~="") then bar[i]=0; else break; end	-- init bar sizes
				end
				local high=1;                                 	-- minimum max value
				for _,i in pairs(buf.Voted) do	            -- cycle all votes
					votecount=votecount+1;
					bar[i]=bar[i]+1;                     	-- add vote to correct bar
					if (bar[i]>high) then high=bar[i] end	-- update highest vote-count
				end
				-- draw bars on screen
				for loopcount,i in ipairs(bar) do
					page=page..[[<bar width=150 pad=20 align=center empty=#F222 inheritsys=Votingbarborder fill=Interface\AddOns\MessageBoard\Images\Smoothv2 max=]]..high..[[>]]..i..[[</bar><text font=GameFontLarge size=16 pad=20 glue=right>|cFFC0C0C0]]..buf.Poll[loopcount]..[[  (|r|cFF00FF96]]..i..[[|r|cFFC0C0C0)|r</text>]]; --#FF69CCF0
				end
				page=page..[[<text align=center pad=20 font=GameFontHighlightSmall size=8>|cFFC0C0C0 Votes: ]]..votecount..[[|r</text>]];
			end
		end
		local mcolor
		if buf.colorR then
			mcolor = string.format("FF%02X%02X%02X",buf.colorR*255,buf.colorG*255,buf.colorB*255)
		else
			mcolor="FFF5F5F5"
		end
		page=page..[[<text pad=25>|C]]..mcolor..msgTxt..[[|r</text>]];
		if buf.Edited then page=page..[[<text pad=25>]]..buf.Edited..[[</text>]]; end
		-- Buttons Start --
		local buttoncount=0;
		local left;
		local name=UnitName("player"); name = Ambiguate(name,"none")
		local reply,edit,delete;
		if not(state==30 or state==15) then -- if not locked state then
			reply=true;
			if (buf.Author==name) then edit=true; delete=true; end	-- non-locked own post
		end
		if (MessageBoard.Administrator or cacheGetMod) then -- if locked and admin then add edit and delete
			edit=true; delete=true;
		end
		buttoncount=buttoncount+((reply and 1) or 0);
		buttoncount=buttoncount+((edit and 1) or 0);
		buttoncount=buttoncount+((delete and 1) or 0);
		if (buttoncount==1) then left=255; elseif (buttoncount==2) then left=200; else left=113; end -- show buttons
		page=page..[[<img color=0000 pad=40 height=1 width=]]..left..[[></img>]];
		if (reply) then page=page..[[<button id=reply pad=40 glue=left>]]..L["TOPIC_SCROLLER_REPLY"]..[[</button>]]; end
		if (edit) then page=page..[[<button id=edit pad=17 glue=left>]]..L["TOPIC_SCROLLER_EDIT"]..[[</button>]]; end
		if (delete) then page=page..[[<button id=delete pad=17 glue=left>]]..L["TOPIC_SCROLLER_DELETE"]..[[</button>]]; end
		-- Buttons End --
		page=page..[[<img align=center pad=40 height=12 width=90%>Interface\FriendsFrame\UI-FriendsFrame-OnlineDivider</img>]];
		MessageBoard:SetReadFlag(category,topic,msg,true);
		mt:Yield();
	end
	page=page..[[<button id=copy pad=17 align=center>Copy</button>]];
	page=page..[[</body></wn>]];
	MessageBoard_Main_MessageScroller:Show(); MessageBoard_Main_MessageScroller_Browser_Frame:Show(); MessageBoard_CurrentTopic:Show();
	MessageBoard_CurrentTopicText:SetText(ren:toASCII(topic));
	MessageBoard_CurrentTopic:SetSize(MessageBoard_CurrentTopic:GetTextWidth()+30,30)
	MessageBoard_CurrentTopic:Disable();
	MessageBoardXML:Update()
	DuckLib.Render:Render(MessageBoard_Main_MessageScroller_Browser_Frame,page);
	wipe(msglist)
end

function MessageBoard_ViewUsers()
	MBWhere="ViewUsers"
	for i = 1,9 do _G["MessageBoard_MainNavBarButton"..i].SelectedBG:Hide(); end
	MessageBoard_MainNewTopic:Disable()
	MessageBoard_MainNewPoll:Disable()
	MessageBoard_MainMsgReply:Disable()
	MessageBoard_Back:SetScript('OnClick', function() MessageBoard_ViewUsers();	end)
	MessageBoard_Back:SetText("Users");
	MessageBoard_Back:SetSize(80, 30);
	MessageBoard_Main_MessageScroller:Hide();MessageBoard_Main_MessageScroller_Browser_Frame:Hide();MessageBoard_CurrentTopic:Hide();
	local page=[[<wn><meta><page background=#0000/><font color=#FFFF size=14/><elements><button inheritSys=MB_Topic_Buttons2/></elements></meta><body><img height=5></img>]];
	local avatar=dbPath["//MBPR/Avatar"];
	for user in pairs(dbPath["//MBPR/Users"]) do -- cycle through the users
		page=page..[[<img align=left pad=0 height=20 width=20>]]..(avatar[user] or "Interface\\Icons\\INV_Misc_QuestionMark")..[[</img><button height=30 user="]]..user..[[" id=user pad=0 width=85% glue=left>]]..user..[[</button>]];
	end
	page=page..[[</body></wn>]];
	MessageBoard_Main_MessageScroller:Show(); MessageBoard_Main_MessageScroller_Browser_Frame:Show();
    DuckLib.Render:Render(MessageBoard_Main_MessageScroller_Browser_Frame,page);
end

function MessageBoard_ViewUserStats(user)
	--MessageBoard_Back:SetScript('OnClick', function() MessageBoard_ViewUsers() end)
	local stats=dbPath["//MBPR/Users/"..user]; if not(stats) then DuckLibMB:Chat("MessageBoard: |cFFFFFFFF"..L["NO_STATS"],0,0.9,0.9); return; end
	local avatar=dbPath["//MBPR/Avatar/"..user] or [[Interface\Icons\INV_Misc_QuestionMark]];
	MessageBoard_Main_MessageScroller:Hide();MessageBoard_Main_MessageScroller_Browser_Frame:Hide();MessageBoard_CurrentTopic:Hide();
	MBWhere="ViewUserStats"
	-- display it
	local page=[[<wn><meta><page background=#0000/><font name=Large size=18/><font name=Huge size=30/><font name=small size=12/></meta><body><img height=10></img>]];
	page=page..[[<img width=60 height=60 pad=0>]]..avatar..[[</img><text pad=15 glue=left font=Huge>]]..stats.nameclasscolor..[[ - ]]..L["TOOLTIP_LEVEL"]..[[ ]]..stats.level.." "..stats.class..[[</text>]];
	page=page..[[<text font=Large>]]..L["TOOLTIP_RANK"]..[[: ]]..stats.guildrank..[[</text>]]
	page=page..[[<text font=Large>]]..L["TOOLTIP_POSTS"]..[[: ]]..MessageBoardXML:UserMessageCount(user)..[[</text>]]
	page=page..[[<button inheritSys=MessageBoard_UserButtonsTemplate user=]]..user..[[ id=pm pad=10>]]..L["SEND_MESSAGE"]..[[</button>]]
	if not(user==UnitName("player")) then
		if MessageBoard.Administrator then
			page=page..[[<button inheritSys=MessageBoard_UserButtonsTemplate glue=left user=]]..user..[[ id=mod pad=10>]]..((stats["mod"] and L["REMOVE_MODERATOR"]) or L["ADD_MODERATOR"])..[[</button>]];
			if (MessageBoardXML:GetMod() or MessageBoard.Administrator) then
				page=page..[[<button inheritSys=MessageBoard_UserButtonsTemplate glue=left user=]]..user..[[ id=ban pad=10>]]..((MBFunc.GetBan(user) and L["UNBAN_USER"]) or L["BAN_USER"])..[[</button>]];
			end
		end
	end
	page=page..[[<img height=15></img>]]
	page=page..[[<text font=Large align=center>- ]]..L["STATS"]..[[ -</text>]]
	page=page..[[<img height=15></img>]]
	page=page..[[<text width=25%>]]..L["ITEM_LEVEL"]..[[: ]]..stats.itemlevel..[[</text>]]
	page=page..[[<img height=10></img>]]
	page=page..[[<text width=25%>|cffffd100]]..L["ATTRIBUTES"]..[[</text><text width=25% glue=left flow=up>|cffffd100]]..L["DEFENSE"]..[[</text><text width=25% glue=left flow=up>|cffffd100]]..L["ATTACK"]..[[</text><text width=25% glue=left flow=up>|cffffd100]]..L["SPELL"]..[[</text>]]
	page=page..[[<img height=5></img>]]
	page=page..[[<text width=25%>]]..L["STRENGTH"]..[[: ]]..stats.strength..[[</text><text width=25% glue=left flow=down>]]..L["ARMOR"]..[[: ]]..stats.armor..[[</text><text width=25% glue=left flow=down>]]..L["DAMAGE"]..[[: ]]..stats.damage..[[</text><text width=25% glue=left flow=down>]]..L["SPELLPOWER"]..[[: ]]..stats.spellpower..[[</text>]]
	page=page..[[<text width=25%>]]..L["AGILITY"]..[[: ]]..stats.agility..[[</text><text width=25% glue=left flow=down>]]..L["DODGE"]..[[: ]]..stats.dodge..[[</text><text width=25% glue=left flow=down>]]..L["ATTACK_POWER"]..[[: ]]..stats.attackpower..[[</text><text width=25% glue=left flow=down>]]..L["MANA_REGEN"]..[[: ]]..stats.manaregen..[[</text>]]
	page=page..[[<text width=25%>]]..L["INTELLECT"]..[[: ]]..stats.intellect..[[</text><text width=25% glue=left flow=down>]]..L["PARRY"]..[[: ]]..stats.block..[[</text><text width=25% glue=left flow=down>]]..L["SPEED"]..[[: ]]..stats.speed..[[</text><text width=25% glue=left flow=down></text>]]
	page=page..[[<text width=25%>]]..L["STAMINA"]..[[: ]]..stats.stamina..[[</text><text width=25% glue=left flow=down>]]..L["BLOCK"]..[[: ]]..stats.parry..[[</text><text width=25% glue=left flow=down>  </text><text width=25% glue=left flow=down></text>]]
	page=page..[[<img height=10></img>]]
	page=page..[[<text width=25%>|cffffd100]]..L["ENHANCEMENTS"]..[[</text>]]
	page=page..[[<img height=5></img>]]
	page=page..[[<text width=25%>]]..L["CRIT"]..[[: ]]..stats.crit..[[</text><text width=25% glue=left flow=down>]]..L["HASTE"]..[[: ]]..stats.haste..[[</text><text width=25% glue=left flow=down>]]..L["MASTERY"]..[[: ]]..stats.mastery..[[</text><text width=25% glue=left flow=down>]]..L["VERSATILITY"]..[[: ]]..stats.versatility..[[</text>]]
	
    -- get users posts
	page=page..[[<text pad=50 font=Large align=center>- ]]..user..[['s ]]..L["POSTS"]..[[ -</text>]];
	page=page..[[<img height=15></img>]];
	for category,cTable in pairs(dbPath["//MBPR/Base"]) do -- cycle through the categories
		local readpermission=MessageBoard:GetCategoryReadPermissions(category);
		listType("database");			-- we want only these
		for topic,tTable in pairs(cTable) do -- cycle through the topics
		local state=dbPath["//MBPR/Base/"..category]:DM_Property(topic,1)
			listType("table");			-- we want only these
			 for k,buf in pairs(tTable) do
				if (type(buf.Text)=="string" and type(buf.Author)=="string" and readpermission and state) then -- dont search deleted messages --
					if (user==buf.Author) then -- Check article author
						--page=page..[[<img align=left pad=0 height=20 width=20>]]..avatar..[[</img><button inheritSys=MB_Topic_Buttons3 height=30 author="]].. user..[[" category="]]..category..[[" topic="]]..topic..[[" id=topic width=85% glue=left>]]..topic..[[</button>]];
						page=page..[[<img align="left" pad="0" height="20" width="20">]]..avatar..[[</img><button inheritSys=MB_Topic_Buttons3 height="30" key="]]..k..[[" term="]]..buf.Text..[[" tip=excerpt author="]].. buf.Author..[[" category="]]..category..[[" topic="]]..topic..[[" id="topic" pad="0" width="95%" glue="left">]]..topic..[[</button>]];
					end
				end
			end
		end
	end	
	page=page..[[</body></wn>]];
	MessageBoard_Main_MessageScroller:Show(); MessageBoard_Main_MessageScroller_Browser_Frame:Show();
	MessageBoard_CurrentTopic:Show();
	MessageBoard_CurrentTopicText:SetText(ren:toASCII(user));
	MessageBoard_CurrentTopic:SetSize(MessageBoard_CurrentTopic:GetTextWidth()+30,30)
	MessageBoard_CurrentTopic:Disable()
    DuckLib.Render:Render(MessageBoard_Main_MessageScroller_Browser_Frame,page);
end

function MessageBoardXML:SetCategories()
	_debug("Setting Categories")
	local unreadcount={0,0,0,0,0,0,0,0,0};
	-- Set Category Buttons Names
	for index=1,9 do
		_G["MessageBoard_MainNavBarButton"..index.."Text"]:SetText(MessageBoardXML:Setting("Category"..index.."Value") or " ");
		if MessageBoardXML:Setting("Category"..index.."Value") then
			if strlen(MessageBoardXML:Setting("Category"..index.."Value")) > 0 then
				local length = strlen(_G["MessageBoard_MainNavBarButton"..index.."Text"]:GetText())
				if length > 12 then _G["MessageBoard_MainNavBarButton"..index.."Text"]:SetSize(130, 30) end
			else
				_G["MessageBoard_MainNavBarButton"..index]:Hide()
			end
		end
	end

	local o=1;
	while o<10 do -- Set Category Buttons Icons and Lock Icon if needed
		if MessageBoardXML:Setting("Category"..o.."Icon") then
			_G["MessageBoard_MainNavBarButton"..o.."Icon"]:SetTexture(MessageBoardXML:Setting("Category"..o.."Icon"))
		end
		if MessageBoard:GetCategoryReadPermissions(o) then
			_G["MessageBoard_MainNavBarButton"..o.."LockIcon"]:Hide()
		else
			_G["MessageBoard_MainNavBarButton"..o.."LockIcon"]:Show()
		end
		o=o+1;
	end

	MessageBoard_MainPMSUnreadBox:Hide();
	local i=1; -- Set Category Button Unread Post Count
	while i<10 do
		_G["MessageBoard_MainNavBarButton"..i.."UnreadBox"]:Hide();
		_G["MessageBoard_MainNavBarButton"..i.."Unread"]:Hide();
		listType("database");			-- we want only these
		for topic,tTable in pairs(dbPath["//MBPR/Base/"..i]) do
			local _,_,unread=MessageBoard:TopicState(i,topic);          -- get unread
			if (unread and MessageBoard:GetCategoryReadPermissions(i)) then unreadcount[i]=unreadcount[i]+1;
				_G["MessageBoard_MainNavBarButton"..i.."Unread"]:Show();
				_G["MessageBoard_MainNavBarButton"..i.."Unread"]:SetText(unreadcount[i]);
				_G["MessageBoard_MainNavBarButton"..i.."UnreadBox"]:Show();
			end
		end
		i=i+1;
	end
	local want=true;					                            	-- want unread messages
	for seq=1,2 do
		for k,v in pairs(MessageBoard_LS.PM) do							-- walk all senders
			local unread=nil;											-- no unread yet
			for i in ipairs(v) do										-- walk all conversation with this sender
				if (not MessageBoard:GetReadFlag("fakePMcategory",k,i)) then _debug("Got an unread PM"); unread=true; break; end	-- look for unread messages
			end
			if (unread==want and unread==true) then						-- got what we want
				_debug("Show the PM notification")
				MessageBoard_MainPMSUnreadBox:Show();
			elseif (unread==want and unread==false) then
				_debug("Hide the PM notification")
				MessageBoard_MainPMSUnreadBox:Hide();
			end
		end
		want=nil;						                            	-- look for read messages
	end	
end

function MessageBoard:SetReadFlag(cat,top,msg,flag)
	local gn=self:GetGuildName(); if (not gn or gn=="<UnGuilded>") then return nil; end
	if (not MessageBoard_DB) then return nil; end
	local bn=gn;
	-- Verify local storage
	if (not MessageBoard_LS[bn]) then MessageBoard_LS[bn]={}; end
	if (not MessageBoard_LS[bn].Base) then MessageBoard_LS[bn].Base={}; end
	if (not MessageBoard_LS[bn].Base[cat]) then MessageBoard_LS[bn].Base[cat]={}; end
	if (not MessageBoard_LS[bn].Base[cat][top]) then MessageBoard_LS[bn].Base[cat][top]={}; end
	MessageBoard_LS[bn].Base[cat][top][msg]=flag;
end

function MessageBoard:GetReadFlag(cat,top,msg)
	local gn=self:GetGuildName(); if (not gn or gn=="<UnGuilded>") then return nil; end
	if (not MessageBoard_DB) then return nil; end
	local bn=gn;
	-- Verify local storage
	if (not MessageBoard_LS[bn]) then return nil; end
	if (not MessageBoard_LS[bn].Base) then return nil; end
	if (not MessageBoard_LS[bn].Base[cat]) then return nil; end
	if (not MessageBoard_LS[bn].Base[cat][top]) then return nil; end
	return MessageBoard_LS[bn].Base[cat][top][msg];
end

function MessageBoard:SetPrunes()
	_debug("Init Prune Settings");
	--local limit=MessageBoardXML:Setting("MessageLimit") or 10;
	local limit=10000
	dbPath["//MBPR"]:DM_Set("Limit",limit);
	for category,cTable in pairs(dbPath["//MBPR/Base"]) do
		cTable:DM_Set("Limit",limit);
		listType("database");			-- we want only these
		for topic,tTable in pairs(cTable) do
			if tTable.State then cTable:DM_Property(topic,1,tTable.State); tTable.State=nil; tTable:DM_SetRedundancy("State"); end
			local state=cTable:DM_Property(topic,1)
			--if state==0 then tTable:DM_PruneMode(DM_PRUNE_FIRST+10);		-- normal
			if state==0 then tTable:DM_PruneMode(DM_PRUNE_NEVER);		-- normal
			--elseif state==40 then tTable:DM_PruneMode(DM_PRUNE_FIRST+11);	-- poll
			elseif state==40 then tTable:DM_PruneMode(DM_PRUNE_NEVER);	-- poll
			elseif state==10 then tTable:DM_PruneMode(DM_PRUNE_NEVER);		-- sticky
			elseif state==15 then tTable:DM_PruneMode(DM_PRUNE_NEVER);		-- sticky locked
			elseif state==20 then tTable:DM_PruneMode(DM_PRUNE_NEVER);		-- announcement
			elseif state==30 then tTable:DM_PruneMode(DM_PRUNE_NEVER);		-- locked
			end
		end
	end	
	dbPath["//MBPR"]:DM_PruneMode(DM_PRUNE_NEVER);
	dbPath["//MBPR/Avatar"]:DM_PruneMode(DM_PRUNE_NEVER);
	dbPath["//MBPR/Base"]:DM_PruneMode(DM_PRUNE_NEVER);
	dbPath["//MBPR/Users"]:DM_PruneMode(DM_PRUNE_NEVER);
end

function MBFunc.MBGuildFrame_UpdateTabard()
	_debug("Updating Guild Emblem");
	MBFunc.MBSetLargeGuildTabardTextures("player", MessageBoard_MainTabardEmblem, MessageBoard_MainTabardBackground, MessageBoard_MainTabardBorder);
	MBFunc.MBSetLargeGuildTabardTextures("player", MessageBoard_EditorTabardEmblem, MessageBoard_EditorTabardBackground, MessageBoard_EditorTabardBorder);
	--MBFunc.MBSetLargeGuildTabardTextures("player", MBSideTabardEmblem, MBSideTabardBackground, MBSideTabardBorder);
	MessageBoard_EditorTabardEmblem:SetAlpha(.10);
	--MBSideTabardEmblem:SetAlpha(1.0);
end

function MBFunc.MBSetLargeGuildTabardTextures(unit, emblemTexture, backgroundTexture, borderTexture, tabardData)
	-- texure dimensions are 1024x1024, icon dimensions are 64x64
	local emblemSize, columns, offset;
	if ( emblemTexture ) then
		emblemSize = 64 / 1024;
		columns = 16
		offset = 0;
		emblemTexture:SetTexture("Interface\\GuildFrame\\GuildEmblemsLG_01");
	end
	MBFunc.MBSetGuildTabardTextures(emblemSize, columns, offset, unit, emblemTexture, backgroundTexture, borderTexture, tabardData);
end

function MBFunc.MBSetSmallGuildTabardTextures(unit, emblemTexture, backgroundTexture, borderTexture, tabardData)
	-- texure dimensions are 256x256, icon dimensions are 16x16, centered in 18x18 cells
	local emblemSize, columns, offset;
	if ( emblemTexture ) then
		emblemSize = 18 / 256;
		columns = 14;
		offset = 1 / 256;
		emblemTexture:SetTexture("Interface\\GuildFrame\\GuildEmblems_01");
	end
	MBFunc.MBSetGuildTabardTextures(emblemSize, columns, offset, unit, emblemTexture, backgroundTexture, borderTexture, tabardData);
end

function MBFunc.MBSetDoubleGuildTabardTextures(unit, leftEmblemTexture, rightEmblemTexture, backgroundTexture, borderTexture, tabardData)
	if ( leftEmblemTexture and rightEmblemTexture ) then
		SetGuildTabardTextures(nil, nil, nil, unit, leftEmblemTexture, backgroundTexture, borderTexture, tabardData);
		rightEmblemTexture:SetTexture(leftEmblemTexture:GetTexture());
		rightEmblemTexture:SetVertexColor(leftEmblemTexture:GetVertexColor());
	end
end

function MBFunc.MBSetGuildTabardTextures(emblemSize, columns, offset, unit, emblemTexture, backgroundTexture, borderTexture, tabardData)
	local bkgR, bkgG, bkgB, borderR, borderG, borderB, emblemR, emblemG, emblemB, emblemFilename;
	if ( tabardData ) then
		bkgR = tabardData[1];
		bkgG = tabardData[2];
		bkgB = tabardData[3];
		borderR = tabardData[4];
		borderG = tabardData[5];
		borderB = tabardData[6];
		emblemR = tabardData[7];
		emblemG = tabardData[8];
		emblemB = tabardData[9];
		emblemFilename = tabardData[10];
	else
		bkgR, bkgG, bkgB, borderR, borderG, borderB, emblemR, emblemG, emblemB, emblemFilename = GetGuildLogoInfo(unit);
	end
	if ( emblemFilename ) then
		if ( backgroundTexture ) then
			backgroundTexture:SetVertexColor(bkgR / 255, bkgG / 255, bkgB / 255);
		end
		if ( borderTexture ) then
			borderTexture:SetVertexColor(borderR / 255, borderG / 255, borderB / 255);
		end
		if ( emblemSize ) then
			local index = emblemFilename:match("([%d]+)");
			if ( index) then
				index = tonumber(index);
				local xCoord = mod(index, columns) * emblemSize;
				local yCoord = floor(index / columns) * emblemSize;
				emblemTexture:SetTexCoord(xCoord + offset, xCoord + emblemSize - offset, yCoord + offset, yCoord + emblemSize - offset);
			end
			emblemTexture:SetVertexColor(emblemR / 255, emblemG / 255, emblemB / 255);
		elseif ( emblemTexture ) then
			emblemTexture:SetTexture(emblemFilename);
			emblemTexture:SetVertexColor(emblemR / 255, emblemG / 255, emblemB / 255);
		end
	else
		-- tabard lacks design
		if ( backgroundTexture ) then
			backgroundTexture:SetVertexColor(0.2245, 0.2088, 0.1794);
		end
		if ( borderTexture ) then
			borderTexture:SetVertexColor(0.2, 0.2, 0.2);
		end
		if ( emblemTexture ) then
			if ( emblemSize ) then
				if ( emblemSize == 18 / 256 ) then
					emblemTexture:SetTexture("Interface\\GuildFrame\\GuildLogo-NoLogoSm");
				else
					emblemTexture:SetTexture("Interface\\GuildFrame\\GuildLogo-NoLogo");
				end
				emblemTexture:SetTexCoord(0, 1, 0, 1);
				emblemTexture:SetVertexColor(1, 1, 1, 1);
			else
				emblemTexture:SetTexture("");
			end
		end
	end
end

Toast:Register("MessageBoard_Toast", function(toast, ...)
	toast:SetTitle("MessageBoard");
	toast:SetText(...);
	toast:SetIconTexture([[Interface\Icons\Inv_holiday_brewfestbuff_01]]);
end)

function MBFunc.SpawnToast(source, text, r, g, b, ...)
	Toast:Spawn("MessageBoard_Toast", text);
end

function MBFunc.isBlank(x)
  return not not tostring(x):find("^%s*$")
end

local settingspack,settingscache;
function MessageBoardXML:Setting(key)
    if (not settingspack or settingspack~=dbPath["//MBPR/Settings"]) then settingspack=dbPath["//MBPR/Settings"]; settingscache=unpacktable(settingspack); end
    return settingscache[key];
end

function MessageBoardCategory_OnClick(self,button,down)
	_debug("Click Category"..tostring(self:GetID()))
	local newId = self:GetID()
	local oldId = CategoryLevel
	local readpermission=MessageBoard:GetCategoryReadPermissions(newId)
	if readpermission then
		if ( oldId ~= newId ) then
			selectedEntryID = nil
		end
		selectedPageNum = 1
		for i = 1,9 do
			_G["MessageBoard_MainNavBarButton"..i].SelectedBG:Hide()
		end
		self.SelectedBG:Show()
		MessageBoardXML:SetCategoryLevel(newId)
		PlaySound("UI_igStore_PageNav_Button")
	end
end


-- Set Category Level
function MessageBoardXML:SetCategoryLevel(category)
	_debug("Set Category Level")
	if MessageBoard:GetCategoryReadPermissions(category) then
		CategoryLevel=category;	MessageBoard.Selected.Topic=nil; MessageBoardXML:ViewTopics();
		MessageBoard_Back:SetText(MessageBoardXML:Setting("Category"..category.."Value") or "");
		MessageBoard_Back:SetSize(MessageBoard_Back:GetTextWidth()+40,30)
		MessageBoard_MainNewTopic:Enable()
		MessageBoard_MainNewPoll:Enable()
		MessageBoard_MainMsgReply:Disable()
	else DuckLibMB:Chat("MessageBoard: |cFFFFFFFF"..L["NOT_ALLOWED_ACCESS"],0,0.9,0.9); end
end

function MessageBoardXML:MakeBool(value)
	if (type(value)=="boolean") then return value; end
	if (not value) then return nil; end
	if (type(value)=="string") then value=tonumber(value); end
	if (type(value)=="number") then
		if (value==0) then return nil; end
		return true;
	end
	return nil;
end

local verTable={}
local verTimer=0

function MessageBoardXML:CheckV()
	wipe(verTable);
    verTimer=debugprofilestop();
	vface:Send(nil,"SendVersion");
	local me=tostring(UnitName("player"))
	me = Ambiguate(me,"none")
	DuckLibMB:Chat("MessageBoard: |cFFFFFFFF"..L["VER_CHECK_TITLE"],0,0.9,0.9)
	DuckLibMB:Chat("MessageBoard: |cFFFFFFFF ---------------------------------------",0,0.9,0.9)
	DuckLibMB:Chat("MessageBoard: |cFFFFFFFF"..me..L["VER_CHECK_USER"].."|cFFFFFFFF"..iVersion,0,0.9,0.9)
end

function MBverRX(iface,data,name,_,toon)    -- handler for incoming communication
    if (name=="SendVersion") then           -- the entry used for version request
        iface:Send(iVersion,"Version");    -- assuming variable "iVersion" contains your own version
	end
	if (name=="Version" and debugprofilestop()-verTimer<2000) then               -- The name of the data in this example is "Version", so the data should be version information for "toon".
		verTable[toon]=data
		verTimer=debugprofilestop();
		DuckLibMB:Chat("MessageBoard: |cFFFFFFFF"..toon..L["VER_CHECK_USER"].."|cFFFFFFFF"..data,0,0.9,0.9);
	end
end

-- Messageboard Minimap Button
function MBFunc.InitializeMinimap()
    _MBLDB = LibStub("LibDataBroker-1.1"):NewDataObject("MessageBoard", {
        type = "launcher",
        text = "MessageBoard",
        icon = "Interface\\Buttons\\UI-GuildButton-PublicNote-Disabled",
        OnClick = function()            
            MessageBoard:ToggleMainFrame()
        end,
        OnTooltipShow = function(tt)
            tt:AddLine("MessageBoard", 1, 1, 1);
        end	
    });
   -- MBFunc.MBUpdateMinimap()
end

function MBFunc.MBUpdateMinimap()
	_debug("Updating Minimap")
	if not _minimapIcon then 
		_minimapIcon = LibStub("LibDBIcon-1.0")
        _minimapIcon:Register("MessageBoard", _MBLDB, MessageBoardLDBIconDB)
	end     
	local count = MessageBoardXML:GetNewMsgCount() or 0
	if count > 0 then
		_MBLDB.icon = 'Interface\\Buttons\\UI-GuildButton-PublicNote-Up'
		_MBLDB.OnTooltipShow = function(tt)
            tt:AddLine("MessageBoard", 1, 1, 1);
            tt:AddLine(" ");
            tt:AddLine(count.." "..L["NEW_MESSAGES"])
        end	
	else
		_MBLDB.icon = 'Interface\\Buttons\\UI-GuildButton-PublicNote-Disabled'
		_MBLDB.OnTooltipShow = function(tt)
            tt:AddLine("MessageBoard", 1, 1, 1);
            tt:AddLine(" ");
            tt:AddLine(L["NO_NEW_MESSAGES"])
        end	
	end	
	local icon = LibStub("LibDBIcon-1.0", true)
	if MessageBoardLDBIconDB.hide then
		icon:Show("MessageBoard")
	else
		icon:Hide("MessageBoard")
	end
end

function MessageBoard:GetTopicTime(topic)
	local tim=nil;
	listType("table");			-- we want only these
	for k,v in pairs(topic) do
		if (not tim or tim<v.Time) then tim=v.Time; end        -- grab newest post in this topic
	end
    return tim
end

local function GetAppropriateDamage()
	if IsRangedWeapon() then
		local attackTime, minDamage, maxDamage, bonusPos, bonusNeg, percent = UnitRangedDamage("player");
		return minDamage, maxDamage, nil, nil, 0, 0, percent;
	else
		return UnitDamage("player");
	end
end

function MessageBoard:ScanStats()
	_debug("Scanning Stats")
	local colorPos = "|cff20ff20";
	local colorNeg = "|cffff2020";
	local name=UnitName("Player");
	name = Ambiguate(name,"none")
	local user=dbPath["//MBPR/Users/"..name];
	local buf={}
	
	local avgItemLevel, avgItemLevelEquipped = GetAverageItemLevel();
	avgItemLevel = floor(avgItemLevel);
	avgItemLevelEquipped = floor(avgItemLevelEquipped);
	buf["itemlevel"]=avgItemLevelEquipped
	-- base stats
	buf["strength"]=UnitStat("player",1);
	
	-- agility
	local stat, effectiveStat, posBuff, negBuff = UnitStat("player", 2);
	local agility = GetAttackPowerForStat(2,effectiveStat);
	if posBuff > 0 then color=colorPos elseif posBuff < 0 then color=colorNeg else color="" end	
	buf["agility"] = color..BreakUpLargeNumbers(agility).."|r";
	
	-- stamina
	local stat, effectiveStat, posBuff, negBuff = UnitStat("player", 3);
	if posBuff > 0 then color=colorPos elseif posBuff < 0 then color=colorNeg else color="" end
	buf["stamina"] = color..BreakUpLargeNumbers(effectiveStat).."|r";
	
	-- intellect
	local stat, effectiveStat, posBuff, negBuff = UnitStat("player", 4);
	if posBuff > 0 then color=colorPos elseif posBuff < 0 then color=colorNeg else color="" end
	buf["intellect"] = color..BreakUpLargeNumbers(effectiveStat).."|r";
	

	-- damage
	local speed, offhandSpeed = UnitAttackSpeed("player");
	local minDamage, maxDamage, minOffHandDamage, maxOffHandDamage, physicalBonusPos, physicalBonusNeg, percent = GetAppropriateDamage("player");
	local displayMin = max(floor(minDamage),1);
	local displayMinLarge = BreakUpLargeNumbers(displayMin);
	local displayMax = max(ceil(maxDamage),1);
	local displayMaxLarge = BreakUpLargeNumbers(displayMax);
	minDamage = (minDamage / percent) - physicalBonusPos - physicalBonusNeg;
	maxDamage = (maxDamage / percent) - physicalBonusPos - physicalBonusNeg;
	local baseDamage = (minDamage + maxDamage) * 0.5;
	local fullDamage = (baseDamage + physicalBonusPos + physicalBonusNeg) * percent;
	local totalBonus = (fullDamage - baseDamage);
	local damageTooltip = displayMinLarge.." - "..displayMaxLarge;	
	if ( totalBonus < 0.1 and totalBonus > -0.1 ) then
		totalBonus = 0.0;
	end
	if ( totalBonus == 0 ) then
		if ( ( displayMin < 100 ) and ( displayMax < 100 ) ) then 
			buf["damage"]=displayMinLarge.." - "..displayMaxLarge;	
		else
			buf["damage"]=displayMinLarge.."-"..displayMaxLarge;
		end
	else
		local color;
		if ( totalBonus > 0 ) then
			color = colorPos;
		else
			color = colorNeg;
		end
		if ( ( displayMin < 100 ) and ( displayMax < 100 ) ) then 
			buf["damage"]=color..displayMinLarge.." - "..displayMaxLarge.."|r";
		else
			buf["damage"]=color..displayMinLarge.."-"..displayMaxLarge.."|r";
		end
	end
	
	-- defense
	buf["dodge"]=string.format("%.2f%%", GetDodgeChance())
	buf["parry"]=string.format("%.2f%%", GetParryChance())
	buf["block"]=string.format("%.2f%%", GetBlockChance())
	buf["armor"]=UnitArmor("player")
	buf["defensemastery"]=string.format("%.2f%%", GetMasteryEffect("player"))	
	-- melee
	buf["meleehaste"] = string.format("%.2f%%", GetMeleeHaste("player"))
	local base, posBuff, negBuff = UnitAttackPower("player")
	buf["attackpower"]=base + posBuff + negBuff
	buf["meleecrit"]=string.format("%.2f%%", GetCritChance("player"))
	buf["meleearmorpen"] = string.format("%.2f%%", GetCombatRatingBonus("25"))
	local mainSpeed, offSpeed = UnitAttackSpeed("player");
	local MH = mainSpeed
	local OH = offSpeed
	buf["weaponspeed"]=string.format("%.2f", MH)
	buf["meleemastery"]=string.format("%.2f%%", GetMasteryEffect("player"))
	-- spell
	buf["spellpower"]=GetSpellBonusDamage(2)
	buf["spellcrit"]=string.format("%.2f%%", GetSpellCritChance("2"))
	buf["spellhaste"]=string.format("%.2f%%", UnitSpellHaste("player"))	
	local MP5_1 = (GetManaRegen() * 5)
	buf["manaregen"]=string.format("%.0f", MP5_1)
	buf["spellmastery"]=string.format("%.2f%%", GetMasteryEffect("player"))	
	-- ranged
	buf["haste"]=string.format("%.2f%%", GetRangedHaste("player"))
	buf["rangedarmorpen"]=string.format("%.2f%%", GetCombatRatingBonus("25"))
	local base, posBuff, negBuff = UnitRangedAttackPower("player")
	local Range_AP = base + posBuff + negBuff
	buf["rangedattackpower"]=Range_AP
	buf["crit"]=string.format("%.2f%%", GetRangedCritChance("25"))
	buf["speed"]=string.format("%.2f", UnitRangedDamage("player"))
	buf["mastery"]=string.format("%.2f%%", GetMasteryEffect("player"))	
	-- secondary stats
	buf["versatility"]=string.format("%.2f%%", GetCombatRatingBonus("29"))
	--buf["leech"]=string.format("%.2f%%", GetCombatRatingBonus("17"))	
	-- Basic Stats
	local race, raceEn = UnitRace("Player");
	local level = UnitLevel("Player");
	local class, classFileName = UnitClass("Player")
	local color = RAID_CLASS_COLORS[classFileName]
	local coloredname="|c"..color.colorStr..name.."|r";
	local _,guildRankName,guildRankIndex = GetGuildInfo("Player")
	-- pvp
	local honorkills, dishonorills = GetPVPLifetimeStats()	
	-- player info	
	buf["class"]=class
	buf["nameclasscolor"]=coloredname
	buf["race"]=race
	buf["level"]=level
	buf["guildrank"]=guildRankName or " "
	buf["guildindex"]=guildRankIndex or " "
	if user then
		if user["mod"]==true then buf["mod"]=true else buf["mod"]=false end
	end
	-- save it
	if (not DuckLib:CompareTables(buf,user)) then			-- if new stats are different from the stored ones
		_debug("New stats are different from the stored ones, so sync.")
		dbPath["//MBPR/Users/"..name]=buf						-- save it and sync
	end
end

-- PM code COMMENT
-- Referenced from MessageBoard.xml
-- List senders - conversations with unread messages first
function MessageBoardXML:ViewPMs()
	mt:Run("ViewPMs",self.ViewPMsMT,self);
end

function MessageBoardXML:ViewPMsMT()
	MBWhere="viewpms"
	MessageBoard_MainNewTopic:Disable()
	MessageBoard_MainNewPoll:Disable()
	MessageBoard_MainMsgReply:Disable()
	for i = 1,9 do _G["MessageBoard_MainNavBarButton"..i].SelectedBG:Hide()	end
	MessageBoard_Back:SetScript('OnClick', function() MessageBoardXML:ViewPMs(); end)
	MessageBoard_Back:SetText("Private Messages");
	MessageBoard_Back:SetSize(145, 30);
	MessageBoard_Main_MessageScroller:Hide(); MessageBoard_Main_MessageScroller_Browser_Frame:Hide();MessageBoard_CurrentTopic:Hide();
    local page=[[<wn><meta><page background="#00000000"/><font color="#FFFFFFFF" size="14"/><elements><button inheritSys="MB_Topic_Buttons2"/></elements></meta><body><img height=5></img>]];
	local icon=Icon.NORMAL_new;				-- unread messages
	local want=true;						-- want unread messages
	for seq=1,2 do
		for k,v in pairs(MessageBoard_LS.PM) do							-- walk all senders
			local unread=nil;											-- no unread yet
			for i in ipairs(v) do										-- walk all conversation with this sender
				if (not MessageBoard:GetReadFlag("fakePMcategory",k,i)) then unread=true; break; end	-- look for unread messages
			end
			if (unread==want) then										-- got what we want
				page=page..[[<img align="left" pad="0" height="20" width="20">]]..icon..[[</img><button height="30" category="]]..k..[[" id="pms" pad="0" width="85%" glue="left">]]..k..[[</button>]];
				page=page..[[<button inheritSys="MessageBoard_PostButtonsTemplate" sender="]]..k..[[" id="deleteconversation" pad="10" width="50" glue="left">Delete</button>]];
			end
			mt:Yield();
		end
		icon=Icon.NORMAL;					-- now we want read messages
		want=nil;							-- look for read messages
	end
    page=page..[[</body></wn>]];
    MessageBoard_Main_MessageScroller:Show();MessageBoard_Main_MessageScroller_Browser_Frame:Show();
    DuckLib.Render:Render(MessageBoard_Main_MessageScroller_Browser_Frame,page);
end

-- Referenced from rendered xml by "MessageBoardXML:ViewPMs()"
function MessageBoard:ViewPM(sender,dummy)
	mt:Run("ViewPM",self.ViewPMMT,self,sender,dummy);
end

function MessageBoard:ViewPMMT(sender,dummy)
	MBWhere="viewpm"
	MessageBoard_Back:SetScript('OnClick', function() MessageBoardXML:ViewPMs(); end)
	MessageBoard.Selected.Topic = nil
	MBMorePostsBelow:Hide();
	local page=[[<wn><meta><page background="#00000000"/><font size="14"/><elements><button inheritSys="MessageBoard_PostButtonsTemplate"/></elements></meta><body>]];
	page=page..[[<h2 glue="left" pad="10">|cFFC0C0C0]]..sender..[[|r</h2>]];
	page=page..[[<img glue="left" pad="10" width="25" height="25">]]..Icon.NORMAL..[[</img>]];
	page=page..[[<img align="center" pad="10" height="12" width="100%">Interface\FriendsFrame\UI-FriendsFrame-OnlineDivider</img>]];
	local pm=MessageBoard_LS.PM[sender]; if (not pm) then return; end	-- grab conversation, abort if it does not exist
	local msg;
	for i,str in ipairs(pm) do
		msg=unpacktable(str,true);	-- tableize (it's a word ... now)
		local avatar=dbPath["//MBPR/Avatar/"..msg.Author] or [[Interface\Icons\INV_Misc_QuestionMark]];
		if (msg.Text:match("(%a+://%S+)")) then msg.Text=msg.Text:gsub("(%a+://%S+)","|cFF4682B4|Hhttp:%1|h[%1]|h|r"); end
		if (msg.Text:match("%f[%S](www%.[-%w_%%]+%.%S+)")) then msg.Text=msg.Text:gsub("%f[%S](www%.[-%w_%%]+%.%S+)","|cFF4682B4|Hhttp:%1|h[%1]|h|r"); end
		page=page..[[<set pm=yes sender="]]..sender..[[" msg="]]..i..[["/>]];
		if (msg.colorR) then page=page..[[<set r="]]..msg.colorR..[[" g="]]..msg.colorG..[[" b="]]..msg.colorB..[["/>]]; end
		page=page.."<img click=left width=60 height=60 pad=10 id=user tip=author user='"..msg.Author.."' author='"..msg.Author.."'>";
		page=page..avatar;
		page=page..[[</img><text size="12" pad="10" glue="left">|cFFC0C0C0from: ]];
		page=page..msg.Author.." - "..MessageBoard:FormatTime(msg.Time).."|r</text>";		
		local mcolor="FFF5F5F5";
		if (msg.colorR) then mcolor=string.format("FF%02X%02X%02X",msg.colorR*255,msg.colorG*255,msg.colorB*255); end
		page=page..[[<text pad="25">|C]]..mcolor..msg.Text..[[|r</text>]];
		page=page..[[<img color="0000" pad="40" height="1" width="200"></img>]];
		page=page..[[<button id="reply" pad="40" glue="left" >]]..L["TOPIC_SCROLLER_REPLY"]..[[</button>]];
		page=page..[[<button id="delete" pad="17" glue="left">]]..L["TOPIC_SCROLLER_DELETE"]..[[</button>]];
		page=page..[[<img align="center" pad="40" height="12" width="90%">Interface\FriendsFrame\UI-FriendsFrame-OnlineDivider</img>]];
		MessageBoard:SetReadFlag("fakePMcategory",msg.Author,i,true);		-- seen it
		mt:Yield();
	end
	page=page..[[</body></wn>]];
	MessageBoard_Main_MessageScroller:Show();MessageBoard_Main_MessageScroller_Browser_Frame:Show();
	MessageBoard_CurrentTopicText:SetText(ren:toASCII(sender));
	MessageBoard_CurrentTopic:SetSize(MessageBoard_CurrentTopic:GetTextWidth()+30,30)
	MessageBoard_CurrentTopic:Disable()
	MessageBoard_CurrentTopic:Show();
	MessageBoardXML:Update()
	DuckLib.Render:Render(MessageBoard_Main_MessageScroller_Browser_Frame,page);
end

function MessageBoard:MoveTopic(tocatno,tocatname,topic,fromcatno)
	local to=dbPath["//MBPR/Base/"..tocatno];
	local from=dbPath["//MBPR/Base/"..fromcatno];
	if (not to[topic]) then
		local state=from:DM_Property(topic,1)                       -- Get the state
		to[topic]=from[topic];									    -- Copy the table
		to:DM_Property(topic,1,state) 								-- Reapply the state
		from[topic]=nil;											-- make the topic nil
		from:DM_SetRedundancy(topic,true);					        -- it's now redundant
		MessageBoardXML:Update();
		MessageBoardXML:ViewTopics();        					    -- view remaining topics
		DuckLibMB:Chat("MessageBoard:|cFFFFFFFF "..L["MOVED_TOPIC"].." "..tostring(topic).." "..L["MOVED_FROM"].." "..tostring(fromcatno).." "..L["MOVED_TO"].." "..tostring(tocatno),0,0.9,0.9);
	else
		DuckLibMB:Chat("MessageBoard:|cFFFFFFFF "..L["TOPIC_EXISTS_CAT"].." "..tostring(tocatname),0,0.9,0.9);
	end
end

function MBFunc.SetBan(toon,reason)
    dbPath["//MBPR"]:DM_SetBan(toon,reason);
    for i=1,9 do dbPath["//MBPR"..i]:DM_SetBan(toon,reason); end
	if reason then
		DuckLibMB:Chat("MessageBoard:|cFFFFFFFF "..tostring(toon).." "..L["IS_BANNED"],0,0.9,0.9)
	else
		DuckLibMB:Chat("MessageBoard:|cFFFFFFFF "..tostring(toon).." "..L["IS_NOT_BANNED"],0,0.9,0.9)
	end
	MessageBoard_ViewUserStats(toon)
end

function MBFunc.GetBan(toon)
    return dbPath["//MBPR"]:DM_GetBan(toon);
end
